(function (global, factory) {
	typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('ramda')) :
	typeof define === 'function' && define.amd ? define(['exports', 'ramda'], factory) :
	(factory((global.cssapi = {}),global.R));
}(this, (function (exports,ramda) { 'use strict';

var ramda__default = 'default' in ramda ? ramda['default'] : ramda;

/**
 * A function that returns `undefined`.
 *
 * @func stubUndefined
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.0.0|v1.0.0}
 * @category Function
 * @sig ... -> undefined
 * @return {undefined}
 * @example
 *
 * RA.stubUndefined(); //=> undefined
 * RA.stubUndefined(1, 2, 3); //=> undefined
 */
var stubUndefined = /*#__PURE__*/ramda.always(void 0); // eslint-disable-line no-void

/**
 * Checks if input value is `undefined`.
 *
 * @func isUndefined
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/0.0.1|v0.0.1}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isNotUndefined|isNotUndefined}
 * @example
 *
 * RA.isUndefined(1); //=> false
 * RA.isUndefined(undefined); //=> true
 * RA.isUndefined(null); //=> false
 */
var isUndefined = /*#__PURE__*/ramda.equals( /*#__PURE__*/stubUndefined());

/**
 * Checks if input value is complement `undefined`.
 *
 * @func isNotUndefined
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/0.0.1|v0.0.1}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isUndefined|isUndefined}
 * @example
 *
 * RA.isNotUndefined(1); //=> true
 * RA.isNotUndefined(undefined); //=> false
 * RA.isNotUndefined(null); //=> true
 */
var isNotUndefined = /*#__PURE__*/ramda.complement(isUndefined);

/**
 * Checks if input value is `null`.
 *
 * @func isNull
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/0.1.0|v0.1.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isNotNull|isNotNull}
 * @example
 *
 * RA.isNull(1); //=> false
 * RA.isNull(undefined); //=> false
 * RA.isNull(null); //=> true
 */
var isNull = /*#__PURE__*/ramda.equals(null);

/**
 * Checks if input value is complement of `null`.
 *
 * @func isNotNull
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/0.1.0|v0.1.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isNull|isNull}
 * @example
 *
 * RA.isNotNull(1); //=> true
 * RA.isNotNull(undefined); //=> true
 * RA.isNotNull(null); //=> false
 */
var isNotNull = /*#__PURE__*/ramda.complement(isNull);

/**
 * Checks if input value is complement of `null` or `undefined`.
 *
 * @func isNotNil
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/0.3.0|v0.3.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link http://ramdajs.com/docs/#isNil|R.isNil}
 * @example
 *
 * RA.isNotNil(null); //=> false
 * RA.isNotNil(undefined); //=> false
 * RA.isNotNil(0); //=> true
 * RA.isNotNil([]); //=> true
 */
var isNotNil = /*#__PURE__*/ramda.complement(ramda.isNil);

/**
 * Tests whether or not an object is an array.
 *
 * @private
 * @param {*} val The object to test.
 * @return {Boolean} `true` if `val` is an array, `false` otherwise.
 * @example
 *
 *      _isArray([]); //=> true
 *      _isArray(null); //=> false
 *      _isArray({}); //=> false
 */
var _isArray = Array.isArray || function _isArray(val) {
  return val != null && val.length >= 0 && Object.prototype.toString.call(val) === '[object Array]';
};

/**
 * Checks if input value is `Array`.
 *
 * @func isArray
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/0.3.0|v0.3.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isNotArray|isNotArray}
 * @example
 *
 * RA.isArray([]); //=> true
 * RA.isArray(null); //=> false
 * RA.isArray({}); //=> false
 */
var isArray = Array.isArray || _isArray;

/**
 * Checks if input value is an empty `Array`.
 *
 * @func isEmptyArray
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.4.0|v2.4.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isNotEmptyArray|isNotEmptyArray}
 * @example
 *
 * RA.isEmptyArray([]); // => true
 * RA.isEmptyArray([42]); // => false
 * RA.isEmptyArray({}); // => false
 * RA.isEmptyArray(null); // => false
 * RA.isEmptyArray(undefined); // => false
 * RA.isEmptyArray(42); // => false
 * RA.isEmptyArray('42'); // => false
 */
var isEmptyArray = /*#__PURE__*/ramda.both(isArray, ramda.isEmpty);

/**
 * Checks if input value is complement of `Array`
 *
 * @func isNotArray
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/0.3.0|v0.3.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isArray|isArray}
 * @example
 *
 * RA.isNotArray([]); //=> false
 * RA.isNotArray(null); //=> true
 * RA.isNotArray({}); //=> true
 */
var isNotArray = /*#__PURE__*/ramda.complement(isArray);

/**
 * Returns true if the given value is not its type's empty value; `false` otherwise.
 *
 * @func isNotEmpty
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/0.4.0|v0.4.0}
 * @category Logic
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link http://ramdajs.com/docs/#isEmpty|R.isEmpty}
 * @example
 *
 * RA.isNotEmpty([1, 2, 3]); //=> true
 * RA.isNotEmpty([]); //=> false
 * RA.isNotEmpty(''); //=> false
 * RA.isNotEmpty(null); //=> true
 * RA.isNotEmpty(undefined): //=> true
 * RA.isNotEmpty({}); //=> false
 * RA.isNotEmpty({length: 0}); //=> true
 */
var isNotEmpty = /*#__PURE__*/ramda.complement(ramda.isEmpty);

/**
 * Checks if input value is not an empty `Array`.
 *
 * @func isNonEmptyArray
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.4.0|v2.4.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isEmptyArray|isEmptyArray}
 * @example
 *
 * RA.isNonEmptyArray([42]); // => true
 * RA.isNonEmptyArray([]); // => false
 * RA.isNonEmptyArray({}); // => false
 * RA.isNonEmptyArray(null); // => false
 * RA.isNonEmptyArray(undefined); // => false
 * RA.isNonEmptyArray(42); // => false
 * RA.isNonEmptyArray('42'); // => false
 */
var isNonEmptyArray = /*#__PURE__*/ramda.both(isArray, isNotEmpty);

/**
 * Checks if input value is `Boolean`.
 *
 * @func isBoolean
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/0.3.0|v0.3.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isNotBoolean|isNotBoolean}
 * @example
 *
 * RA.isBoolean(false); //=> true
 * RA.isBoolean(true); //=> true
 * RA.isBoolean(null); //=> false
 */
var isBoolean = /*#__PURE__*/ramda.is(Boolean);

/**
 * Checks if input value is complement of `Boolean`.
 *
 * @func isNotBoolean
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/0.3.0|v0.3.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isBoolean|isBoolean}
 * @example
 *
 * RA.isNotBoolean(false); //=> false
 * RA.isNotBoolean(true); //=> false
 * RA.isNotBoolean(null); //=> true
 */
var isNotBoolean = /*#__PURE__*/ramda.complement(isBoolean);

/**
 * Returns `true` if the given value is its type's empty value, `null` or `undefined`.
 *
 * @func isNilOrEmpty
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/0.4.0|v0.4.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link http://ramdajs.com/docs/#isEmpty|R.isEmpty}, {@link http://ramdajs.com/docs/#isNil|R.isNil}
 * @example
 *
 * RA.isNilOrEmpty([1, 2, 3]); //=> false
 * RA.isNilOrEmpty([]); //=> true
 * RA.isNilOrEmpty(''); //=> true
 * RA.isNilOrEmpty(null); //=> true
 * RA.isNilOrEmpty(undefined): //=> true
 * RA.isNilOrEmpty({}); //=> true
 * RA.isNilOrEmpty({length: 0}); //=> false
 */
var isNilOrEmpty = /*#__PURE__*/ramda.either(ramda.isNil, ramda.isEmpty);

function _isString(x) {
  return Object.prototype.toString.call(x) === '[object String]';
}
var _isString_1 = _isString;

/**
 * Checks if input value is `String`.
 *
 * @func isString
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/0.4.0|v0.4.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isNotString|isNotString}
 * @example
 *
 * RA.isString('abc'); //=> true
 * RA.isString(1); //=> false
 */
var isString = _isString_1;

/**
 * Checks if input value is an empty `String`.
 *
 * @func isEmptyString
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.4.0|v2.4.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isNotEmptyString|isNotEmptyString}
 * @example
 *
 * RA.isEmptyString(''); // => true
 * RA.isEmptyString('42'); // => false
 * RA.isEmptyString(new String('42')); // => false
 * RA.isEmptyString(new String('')); // => false
 * RA.isEmptyString([42]); // => false
 * RA.isEmptyString({}); // => false
 * RA.isEmptyString(null); // => false
 * RA.isEmptyString(undefined); // => false
 * RA.isEmptyString(42); // => false
 */
var isEmptyString = /*#__PURE__*/ramda.equals('');

/**
 * Checks if input value is complement of `String`.
 *
 * @func isNotString
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/0.4.0|v0.4.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isString|isString}
 * @example
 *
 * RA.isNotString('abc'); //=> false
 * RA.isNotString(1); //=> true
 */
var isNotString = /*#__PURE__*/ramda.complement(isString);

var GeneratorFunction = null;
try {
  GeneratorFunction = /*#__PURE__*/new Function('return function* () {}')().constructor; // eslint-disable-line no-new-func
} catch (e) {} // eslint-disable-line no-empty

/* eslint-disable max-len */
/**
 * Checks if input value is `Generator Function`.
 *
 * @func isGeneratorFunction
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/0.5.0|v0.5.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isFunction|isFunction}, {@link RA.isAsyncFunction|isAsyncFunction}, {@link RA.isNotGeneratorFunction|isNotGeneratorFunction}
 * @example
 *
 * RA.isGeneratorFunction(function* test() { }); //=> true
 * RA.isGeneratorFunction(null); //=> false
 * RA.isGeneratorFunction(function test() { }); //=> false
 * RA.isGeneratorFunction(() => {}); //=> false
 */
/* eslint-enable max-len */
var isGeneratorFunction = function isGeneratorFunction(val) {
  var toStringCheck = Object.prototype.toString.call(val) === '[object GeneratorFunction]';
  var legacyConstructorCheck = isNotNull(GeneratorFunction) && val instanceof GeneratorFunction;

  return toStringCheck || legacyConstructorCheck;
};

/* eslint-disable max-len */
/**
 * Checks if input value is `Async Function`.
 *
 * @func isAsyncFunction
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/0.5.0|v0.5.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isFunction|isFunction}, {@link RA.isNotAsyncFunction|isNotAsyncFunction}, {@link RA.isGeneratorFunction|isGeneratorFunction}
 * @example
 *
 * RA.isAsyncFunction(async function test() { }); //=> true
 * RA.isAsyncFunction(null); //=> false
 * RA.isAsyncFunction(function test() { }); //=> false
 * RA.isAsyncFunction(() => {}); //=> false
 */
/* eslint-enable max-len */
var isAsyncFunction = function isAsyncFunction(val) {
  return Object.prototype.toString.call(val) === '[object AsyncFunction]';
};

/* eslint-disable max-len */
/**
 * Checks if input value is `Function`.
 *
 * @func isFunction
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/0.5.0|v0.5.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isNotFunction|isNotFunction}, {@link RA.isAsyncFunction|isNotAsyncFunction}, {@link RA.isGeneratorFunction|isGeneratorFunction}
 * @example
 *
 * RA.isFunction(function test() { }); //=> true
 * RA.isFunction(function* test() { }); //=> true
 * RA.isFunction(async function test() { }); //=> true
 * RA.isFunction(() => {}); //=> true
 * RA.isFunction(null); //=> false
 * RA.isFunction('abc'); //=> false
 */
/* eslint-enable max-len */
var isFunction = /*#__PURE__*/ramda.anyPass([function (val) {
  return Object.prototype.toString.call(val) === '[object Function]';
}, isGeneratorFunction, isAsyncFunction]);

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

var isOfTypeObject = function isOfTypeObject(val) {
  return (typeof val === 'undefined' ? 'undefined' : _typeof(val)) === 'object';
};

/* eslint-disable max-len */
/**
 * Checks if input value is language type of `Object`.
 *
 * @func isObj
 * @aliases isObject
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/0.5.0|v0.5.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isNotObj|isNotObj}, {@link RA.isObjLike|isObjLike}, {@link RA.isPlainObj|isPlainObj}
 * @example
 *
 * RA.isObj({}); //=> true
 * RA.isObj([]); //=> true
 * RA.isObj(() => {}); //=> true
 * RA.isObj(null); //=> false
 * RA.isObj(undefined); //=> false
 */
/* eslint-enable max-len */
var isObj = /*#__PURE__*/ramda.both(isNotNull, /*#__PURE__*/ramda.either(isOfTypeObject, isFunction));

/* eslint-disable max-len */
/**
 * Checks if input value is complement of language type of `Object`.
 *
 * @func isNotObj
 * @aliases isNotObject
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/0.5.0|v0.5.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isObj|isObj}, {@link RA.isObjLike|isObjLike}, {@link RA.isPlainObj|isPlainObj}
 * @example
 *
 * RA.isNotObj({}); //=> false
 * RA.isNotObj([]); //=> false
 * RA.isNotObj(() => {}); //=> false
 * RA.isNotObj(null); //=> true
 * RA.isNotObj(undefined); //=> true
 */
/* eslint-enable max-len */
var isNotObj = /*#__PURE__*/ramda.complement(isObj);

/**
 * Checks if input value is not an empty `String`.
 *
 * @func isNonEmptyString
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.4.0|v2.4.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isEmptyString|isEmptyString}
 * @example
 *
 * RA.isNonEmptyString('42'); // => true
 * RA.isNonEmptyString(''); // => false
 * RA.isNonEmptyString(new String('42')); // => false
 * RA.isNonEmptyString(new String('')); // => false
 * RA.isNonEmptyString([42]); // => false
 * RA.isNonEmptyString({}); // => false
 * RA.isNonEmptyString(null); // => false
 * RA.isNonEmptyString(undefined); // => false
 * RA.isNonEmptyString(42); // => false
 */
var isNonEmptyString = /*#__PURE__*/ramda.allPass([isString, isNotObj, isNotEmpty]);

var _typeof$1 = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

/* eslint-disable max-len */
/**
 * Tests whether or not an object is similar to an array.
 *
 * @func isArrayLike
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.9.0|v1.9.0}
 * @licence https://github.com/ramda/ramda/blob/master/LICENSE.txt
 * @category List
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @returns {boolean} `true` if `val` has a numeric length property and extreme indices defined; `false` otherwise.
 * @see {@link RA.isNotArrayLike|isNotArrayLike}

 * @example
 *
 * RA.isArrayLike([]); //=> true
 * RA.isArrayLike(true); //=> false
 * RA.isArrayLike({}); //=> false
 * RA.isArrayLike({length: 10}); //=> false
 * RA.isArrayLike({0: 'zero', 9: 'nine', length: 10}); //=> true
 */
/* eslint-enable max-len */
var isArrayLike = function isArrayLike(val) {
  if (isArray(val)) {
    return true;
  }
  if (!val) {
    return false;
  }
  if (isString(val)) {
    return false;
  }
  if ((typeof val === 'undefined' ? 'undefined' : _typeof$1(val)) !== 'object') {
    return false;
  }
  if (val.nodeType === 1) {
    return !!val.length;
  }
  if (val.length === 0) {
    return true;
  }
  if (val.length > 0) {
    return ramda.has(0, val) && ramda.has(val.length - 1, val);
  }
  return false;
};



/**
 The MIT License (MIT)

 Copyright (c) 2013-2016 Scott Sauyet and Michael Hurley

 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
*/

/**
 * Tests whether or not an object is similar to an array.
 *
 * @func isNotArrayLike
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/0.5.0|v0.5.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isArrayLike|isArrayLike}
 * @example
 *
 * RA.isNotArrayLike([]); //=> false
 * RA.isNotArrayLike(true); //=> true
 * RA.isNotArrayLike({}); //=> true
 * RA.isNotArrayLike({length: 10}); //=> true
 * RA.isNotArrayLike({0: 'zero', 9: 'nine', length: 10}); //=> false
 */
var isNotArrayLike = /*#__PURE__*/ramda.complement(isArrayLike);

/* eslint-disable max-len */
/**
 * Checks if input value is complement of `Generator Function`
 *
 * @func isNotGeneratorFunction
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/0.5.0|v0.5.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isFunction|isFunction}, {@link RA.isAsyncFunction|isAsyncFunction}, {@link RA.isGeneratorFunction|isGeneratorFunction}
 * @example
 *
 * RA.isNotGeneratorFunction(function* test() { }); //=> false
 * RA.isNotGeneratorFunction(null); //=> true
 * RA.isNotGeneratorFunction(function test() { }); //=> true
 * RA.isNotGeneratorFunction(() => {}); //=> true
 */
/* eslint-enable max-len */
var isNotGeneratorFunction = /*#__PURE__*/ramda.complement(isGeneratorFunction);

/* eslint-disable max-len */
/**
 * Checks if input value is complement of `Async Function`
 *
 * @func isNotAsyncFunction
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/0.5.0|v0.5.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isFunction|isFunction}, {@link RA.isAsyncFunction|isAsyncFunction}, {@link RA.isGeneratorFunction|isGeneratorFunction}
 * @example
 *
 * RA.isNotAsyncFunction(async function test() { }); //=> false
 * RA.isNotAsyncFunction(null); //=> true
 * RA.isNotAsyncFunction(function test() { }); //=> true
 * RA.isNotAsyncFunction(() => {}); //=> true
 */
/* eslint-enable max-len */
var isNotAsyncFunction = /*#__PURE__*/ramda.complement(isAsyncFunction);

/* eslint-disable max-len */
/**
 * Checks if input value is complement of `Function`.
 *
 * @func isNotFunction
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/0.5.0|v0.5.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isFunction|isFunction}, {@link RA.isAsyncFunction|isNotAsyncFunction}, {@link RA.isGeneratorFunction|isGeneratorFunction}
 * @example
 *
 * RA.isNotFunction(function test() { }); //=> false
 * RA.isNotFunction(function* test() { }); //=> false
 * RA.isNotFunction(async function test() { }); //=> false
 * RA.isNotFunction(() => {}); //=> false
 * RA.isNotFunction(null); //=> true
 * RA.isNotFunction('abc'); //=> true
 */
/* eslint-enable max-len */
var isNotFunction = /*#__PURE__*/ramda.complement(isFunction);

/* eslint-disable max-len */
/**
 * Checks if value is object-like. A value is object-like if it's not null and has a typeof result of "object".
 *
 * @func isObjLike
 * @aliases isObjectLike
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/0.5.0|v0.5.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isNotObjLike|isNotObjLike}, {@link RA.isObj|isObj}, {@link RA.isPlainObj|isPlainObj}
 * @example
 *
 * RA.isObjLike({}); //=> true
 * RA.isObjLike([]); //=> true
 * RA.isObjLike(() => {}); //=> false
 * RA.isObjLike(null); //=> false
 * RA.isObjLike(undefined); //=> false
 */
/* eslint-enable max-len */
var isObjLike = /*#__PURE__*/ramda.both(isNotNull, isOfTypeObject);

/* eslint-disable max-len */
/**
 * Checks if value is not object-like. A value is object-like if it's not null and has a typeof result of "object".
 *
 * @func isNotObjLike
 * @aliases isNotObjectLike
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/0.5.0|v0.5.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isObjLike|isObjLike}, {@link RA.isObj|isObj}, {@link RA.isPlainObj|isPlainObj}
 * @example
 *
 * RA.isNotObjLike({}); //=> false
 * RA.isNotObjLike([]); //=> false
 * RA.isNotObjLike(() => {}); //=> true
 * RA.isNotObjLike(null); //=> true
 * RA.isNotObjLike(undefined); //=> true
 */
/* eslint-enable max-len */
var isNotObjLike = /*#__PURE__*/ramda.complement(isObjLike);

function _isObject(x) {
  return Object.prototype.toString.call(x) === '[object Object]';
}
var _isObject_1 = _isObject;

var isObjectConstructor = /*#__PURE__*/ramda.pipe(ramda.toString, /*#__PURE__*/ramda.equals( /*#__PURE__*/ramda.toString(Object)));
var hasObjectConstructor = /*#__PURE__*/ramda.pathSatisfies( /*#__PURE__*/ramda.both(isFunction, isObjectConstructor), ['constructor']);

/* eslint-disable max-len */
/**
 * Check to see if an object is a plain object (created using `{}`, `new Object()` or `Object.create(null)`).
 *
 * @func isPlainObj
 * @aliases isPlainObject
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/0.5.0|v0.5.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isNotPlainObj|isNotPlainObj}, {@link RA.isObjLike|isObjLike}, {@link RA.isObj|isObj}
 * @example
 *
 * class Bar {
 *   constructor() {
 *     this.prop = 'value';
 *   }
 * }
 *
 * RA.isPlainObj(new Bar()); //=> false
 * RA.isPlainObj({ prop: 'value' }); //=> true
 * RA.isPlainObj(['a', 'b', 'c']); //=> false
 * RA.isPlainObj(Object.create(null); //=> true
 * RA.isPlainObj(new Object()); //=> true
 */
/* eslint-enable max-len */
var isPlainObj = function isPlainObj(val) {
  if (!isObjLike(val) || !_isObject_1(val)) {
    return false;
  }

  var proto = Object.getPrototypeOf(val);

  if (isNull(proto)) {
    return true;
  }

  return hasObjectConstructor(proto);
};

/* eslint-disable max-len */
/**
 * Check to see if an object is a not plain object (created using `{}`, `new Object()` or `Object.create(null)`).
 *
 * @func isNotPlainObj
 * @aliases isNotPlainObject
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/0.5.0|v0.5.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isPlainObj|isPlainObj}, {@link RA.isObjLike|isObjLike}, {@link RA.isObj|isObj}
 * @example
 *
 * class Bar {
 *   constructor() {
 *     this.prop = 'value';
 *   }
 * }
 *
 * RA.isNotPlainObj(new Bar()); //=> true
 * RA.isNotPlainObj({ prop: 'value' }); //=> false
 * RA.isNotPlainObj(['a', 'b', 'c']); //=> true
 * RA.isNotPlainObj(Object.create(null); //=> false
 * RA.isNotPlainObj(new Object()); //=> false
 */
/* eslint-enable max-len */
var isNotPlainObj = /*#__PURE__*/ramda.complement(isPlainObj);

/* eslint-disable max-len */
/**
 * Checks if value is `Date` object.
 *
 * @func isDate
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/0.6.0|v0.6.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isNotDate|isNotDate}, {@link RA.isValidDate|isValidDate}, {@link RA.isNotValidDate|isNotValidDate}
 * @example
 *
 * RA.isDate(new Date()); //=> true
 * RA.isDate('1997-07-16T19:20+01:00'); //=> false
 */
/* eslint-enable max-len */
var isDate = /*#__PURE__*/ramda.is(Date);

/**
 * Checks if value is complement of `Date` object.
 *
 * @func isNotDate
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/0.6.0|v0.6.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isDate|isDate}
 * @example
 *
 * RA.isNotDate(new Date()); //=> false
 * RA.isNotDate('1997-07-16T19:20+01:00'); //=> true
 */
var isNotDate = /*#__PURE__*/ramda.complement(isDate);

function _isNumber(x) {
  return Object.prototype.toString.call(x) === '[object Number]';
}
var _isNumber_1 = _isNumber;

/**
 * Checks if value is a `Number` primitive or object.
 *
 * @func isNumber
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/0.6.0|v0.6.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isNotNumber|isNotNumber}
 * @example
 *
 * RA.isNumber(5); // => true
 * RA.isNumber(Number.MAX_VALUE); // => true
 * RA.isNumber(-Infinity); // => true
 * RA.isNumber(NaN); // => true
 * RA.isNumber('5'); // => false
 */
var isNumber = _isNumber_1;

// eslint-disable-next-line no-restricted-globals
var isNaNPolyfill = /*#__PURE__*/ramda.both(isNumber, isNaN);

/**
 * Checks whether the passed value is `NaN` and its type is `Number`.
 * It is a more robust version of the original, global isNaN().
 *
 *
 * @func isNaN
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/0.6.0|v0.6.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isNotNaN|isNotNaN}
 * @example
 *
 * RA.isNaN(NaN); // => true
 * RA.isNaN(Number.NaN); // => true
 * RA.isNaN(0 / 0); // => true
 *
 * // e.g. these would have been true with global isNaN().
 * RA.isNaN('NaN'); // => false
 * RA.isNaN(undefined); // => false
 * RA.isNaN({}); // => false
 * RA.isNaN('blabla'); // => false
 *
 * RA.isNaN(true); // => false
 * RA.isNaN(null); // => false
 * RA.isNaN(37); // => false
 * RA.isNaN('37'); // => false
 * RA.isNaN('37.37'); // => false
 * RA.isNaN(''); // => false
 * RA.isNaN(' '); // => false
 */
var _isNaN = Number.isNaN || isNaNPolyfill;

/**
 * Checks whether the passed value is complement of `NaN` and its type is not `Number`.
 *
 * @func isNotNaN
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/0.6.0|v0.6.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isNaN|isNaN}
 * @example
 *
 * RA.isNotNaN(NaN); // => false
 * RA.isNotNaN(Number.NaN); // => false
 * RA.isNotNaN(0 / 0); // => false
 *
 * RA.isNotNaN('NaN'); // => true
 * RA.isNotNaN(undefined); // => true
 * RA.isNotNaN({}); // => true
 * RA.isNotNaN('blabla'); // => true
 *
 * RA.isNotNaN(true); // => true
 * RA.isNotNaN(null); // => true
 * RA.isNotNaN(37); // => true
 * RA.isNotNaN('37'); // => true
 * RA.isNotNaN('37.37'); // => true
 * RA.isNotNaN(''); // => true
 * RA.isNotNaN(' '); // => true
 */
var isNotNaN = /*#__PURE__*/ramda.complement(_isNaN);

/* eslint-disable max-len */
/**
 * Checks if value is valid `Date` object.
 *
 * @func isValidDate
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.8.0|v1.8.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isDate|isDate}, {@link RA.isNotDate|isNotDate}, {@link RA.isNotValidDate|isNotValidDate}
 * @example
 *
 * RA.isValidDate(new Date()); //=> true
 * RA.isValidDate(new Date('a')); //=> false
 */
/* eslint-enable max-len */
var isValidDate = /*#__PURE__*/ramda.both(isDate, /*#__PURE__*/ramda.pipe( /*#__PURE__*/ramda.invoker(0, 'getTime'), isNotNaN));

/**
 * Checks if value is complement of valid `Date` object.
 *
 * @func isNotValidDate
 * @aliases isInvalidDate
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.8.0|v1.8.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isValidDate|isValidDate}, {@link RA.isDate|isDate}, {@link RA.isNotDate|isNotDate}
 * @example
 *
 * RA.isNotValidDate(new Date()); //=> false
 * RA.isNotValidDate(new Date('a')); //=> true
 */
var isNotValidDate = /*#__PURE__*/ramda.complement(isValidDate);

/**
 * Checks if value is a complement of `Number` primitive or object.
 *
 * @func isNotNumber
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/0.6.0|v0.6.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isNumber|isNumber}
 * @example
 *
 * RA.isNotNumber(5); // => false
 * RA.isNotNumber(Number.MAX_VALUE); // => false
 * RA.isNotNumber(-Infinity); // => false
 * RA.isNotNumber('5'); // => true
 */
var isNotNumber = /*#__PURE__*/ramda.complement(isNumber);

/**
 * Checks if value is a positive `Number` primitive or object. Zero is not considered positive.
 *
 * @func isPositive
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.15.0|v1.15.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isNegative|isNegative}
 * @example
 *
 * RA.isPositive(1); // => true
 * RA.isPositive(Number.MAX_VALUE); // => true
 * RA.isPositive(-Infinity); // => false
 * RA.isPositive(NaN); // => false
 * RA.isPositive('5'); // => false
 */
var isPositive = /*#__PURE__*/ramda.both(isNumber, /*#__PURE__*/ramda.lt(0));

/**
 * Checks if value is a negative `Number` primitive or object. Zero is not considered neither
 * positive or negative.
 *
 * @func isNegative
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.15.0|v1.15.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isPositive|isPositive}
 * @example
 *
 * RA.isNegative(-1); // => true
 * RA.isNegative(Number.MIN_VALUE); // => false
 * RA.isNegative(+Infinity); // => false
 * RA.isNegative(NaN); // => false
 * RA.isNegative('5'); // => false
 */
var isNegative = /*#__PURE__*/ramda.both(isNumber, /*#__PURE__*/ramda.gt(0));

/**
 * Checks if value is a non-positive `Number` primitive or object. This includes all negative
 * numbers and zero.
 *
 * @func isNonPositive
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.6.0|v2.6.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isNegative|isNegative}, {@link RA.isNonNegative|isNonNegative}
 * @example
 *
 * RA.isNonPositive(0); // => true
 * RA.isNonPositive(-1); // => true
 * RA.isNonPositive(-Infinity); // => true
 * RA.isNonPositive(Number.MIN_VALUE); // => true
 *
 * RA.isNonPositive(Infinity); // => false
 * RA.isNonPositive(Number.MAX_VALUE); // => false
 * RA.isNonPositive(NaN); // => false
 */
var isNonPositive = /*#__PURE__*/ramda.both(isNumber, /*#__PURE__*/ramda.flip(ramda.lte)(0));

/**
 * Checks if value is a non-negative `Number` primitive or object. This includes all positive
 * numbers and zero.
 *
 * @func isNonNegative
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.6.0|v2.6.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isPositive|isPositive}, {@link RA.isNonPositive|isNonPositive}
 * @example
 *
 * RA.isNonNegative(0); // => true
 * RA.isNonNegative(1); // => true
 * RA.isNonNegative(Infinity); // => true
 * RA.isNonNegative(Number.MAX_VALUE); // => true
 *
 * RA.isNonNegative(-Infinity); // => false
 * RA.isNonNegative(Number.MIN_VALUE); // => false
 * RA.isNonNegative(NaN); // => false
 */
var isNonNegative = /*#__PURE__*/ramda.both(isNumber, /*#__PURE__*/ramda.flip(ramda.gte)(0));

// eslint-disable-next-line no-restricted-globals
var isFinitePolyfill = /*#__PURE__*/ramda.both(isNumber, isFinite);

/**
 * Checks whether the passed value is a finite `Number`.
 *
 * @func isFinite
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/0.7.0|v0.7.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isNotFinite|isNotFinite}
 * @example
 *
 * RA.isFinite(Infinity); //=> false
 * RA.isFinite(NaN); //=> false
 * RA.isFinite(-Infinity); //=> false
 *
 * RA.isFinite(0); // true
 * RA.isFinite(2e64); // true
 *
 * RA.isFinite('0');  // => false
 *                    // would've been true with global isFinite('0')
 * RA.isFinite(null); // => false
 *                    // would've been true with global isFinite(null)
 */
var _isFinite = Number.isFinite || isFinitePolyfill;

/**
 * Checks whether the passed value is complement of finite `Number`.
 *
 *
 * @func isNotFinite
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/0.7.0|v0.7.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isFinite|isFinite}
 * @example
 *
 * RA.isNotFinite(Infinity); //=> true
 * RA.isNotFinite(NaN); //=> true
 * RA.isNotFinite(-Infinity); //=> true
 *
 * RA.isNotFinite(0); // false
 * RA.isNotFinite(2e64); // false
 *
 * RA.isNotFinite('0');  // => true
 * RA.isNotFinite(null); // => true
 */
var isNotFinite = /*#__PURE__*/ramda.complement(_isFinite);

var isIntegerPolyfill = /*#__PURE__*/ramda.both(_isFinite, /*#__PURE__*/ramda.converge(ramda.equals, [Math.floor, ramda.identity]));

/**
 * Checks whether the passed value is an `integer`.
 *
 * @func isInteger
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/0.7.0|v0.7.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isNotInteger|isNotInteger}
 * @example
 *
 * RA.isInteger(0); //=> true
 * RA.isInteger(1); //=> true
 * RA.isInteger(-100000); //=> true
 *
 * RA.isInteger(0.1);       //=> false
 * RA.isInteger(Math.PI);   //=> false
 *
 * RA.isInteger(NaN);       //=> false
 * RA.isInteger(Infinity);  //=> false
 * RA.isInteger(-Infinity); //=> false
 * RA.isInteger('10');      //=> false
 * RA.isInteger(true);      //=> false
 * RA.isInteger(false);     //=> false
 * RA.isInteger([1]);       //=> false
 */
var isInteger = Number.isInteger || isIntegerPolyfill;

/**
 * Checks whether the passed value is complement of an `integer`.
 *
 *
 * @func isNotInteger
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/0.7.0|v0.7.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isInteger|isInteger}
 * @example
 *
 * RA.isNotInteger(0); //=> false
 * RA.isNotInteger(1); //=> false
 * RA.isNotInteger(-100000); //=> false
 *
 * RA.isNotInteger(0.1);       //=> true
 * RA.isNotInteger(Math.PI);   //=> true
 *
 * RA.isNotInteger(NaN);       //=> true
 * RA.isNotInteger(Infinity);  //=> true
 * RA.isNotInteger(-Infinity); //=> true
 * RA.isNotInteger('10');      //=> true
 * RA.isNotInteger(true);      //=> true
 * RA.isNotInteger(false);     //=> true
 * RA.isNotInteger([1]);       //=> true
 */
var isNotInteger = /*#__PURE__*/ramda.complement(isInteger);

/**
 * Checks whether the passed value is a `float`.
 *
 * @func isFloat
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.14.0|v1.14.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isNotFloat|isNotFloat}
 * @example
 *
 * RA.isFloat(0); //=> false
 * RA.isFloat(1); //=> false
 * RA.isFloat(-100000); //=> false
 *
 * RA.isFloat(0.1);       //=> true
 * RA.isFloat(Math.PI);   //=> true
 *
 * RA.isFloat(NaN);       //=> false
 * RA.isFloat(Infinity);  //=> false
 * RA.isFloat(-Infinity); //=> false
 * RA.isFloat('10');      //=> false
 * RA.isFloat(true);      //=> false
 * RA.isFloat(false);     //=> false
 * RA.isFloat([1]);       //=> false
 */
var isFloat = /*#__PURE__*/ramda.both(_isFinite, /*#__PURE__*/ramda.complement(isInteger));

/**
 * Checks whether the passed value is complement of a `float`.
 *
 * @func isNotFloat
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.14.0|v1.14.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isFloat|isFloat}
 * @example
 *
 * RA.isNotFloat(0); //=> true
 * RA.isNotFloat(1); //=> true
 * RA.isNotFloat(-100000); //=> true
 *
 * RA.isNotFloat(0.1);       //=> false
 * RA.isNotFloat(Math.PI);   //=> false
 *
 * RA.isNotFloat(NaN);       //=> true
 * RA.isNotFloat(Infinity);  //=> true
 * RA.isNotFloat(-Infinity); //=> true
 * RA.isNotFloat('10');      //=> true
 * RA.isNotFloat(true);      //=> true
 * RA.isNotFloat(false);     //=> true
 * RA.isNotFloat([1]);       //=> true
 */
var isNotFloat = /*#__PURE__*/ramda.complement(isFloat);

/**
 * Checks if value is a valid `Number`. A valid `Number` is a number that is not `NaN`, `Infinity`
 * or `-Infinity`.
 *
 * @func isValidNumber
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.2.0|v2.2.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isNotValidNumber|isNotValidNumber}
 * @example
 *
 * RA.isValidNumber(1); //=> true
 * RA.isValidNumber(''); //=> false
 * RA.isValidNumber(NaN); //=> false
 * RA.isValidNumber(Infinity); //=> false
 * RA.isValidNumber(-Infinity); //=> false
 */
var isValidNumber = /*#__PURE__*/ramda.either(isInteger, isFloat);

/**
 * Checks if value is not a valid `Number`. A valid `Number` is a number that is not `NaN`,
 * `Infinity` or `-Infinity`.
 *
 * @func isNotValidNumber
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.2.0|v2.2.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isValidNumber|isValidNumber}
 * @example
 *
 * RA.isNotValidNumber(1); //=> false
 * RA.isNotValidNumber(''); //=> true
 * RA.isNotValidNumber(NaN); //=> true
 * RA.isNotValidNumber(Infinity); //=> true
 * RA.isNotValidNumber(-Infinity); //=> true
 */
var isNotValidNumber = /*#__PURE__*/ramda.complement(isValidNumber);

/**
 * Checks if value is odd integer number.
 * An odd number is an integer which is not a multiple DIVISIBLE of two.
 *
 * @func isOdd
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.18.0|v1.18.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isEven|isEven}
 * @example
 *
 * RA.isOdd(1); // => true
 * RA.isOdd(-Infinity); // => false
 * RA.isOdd(4); // => false
 * RA.isOdd(3); // => true
 */
var isOdd = /*#__PURE__*/ramda.both(isInteger, /*#__PURE__*/ramda.pipe( /*#__PURE__*/ramda.flip(ramda.modulo)(2), /*#__PURE__*/ramda.complement(ramda.equals)(0)));

/**
 * Checks if value is even integer number.
 * An even number is an integer which is "evenly divisible" by two.
 * Zero is an even number because zero divided by two equals zero,
 * which despite not being a natural number, is an integer.
 * Even numbers are either positive or negative.
 *
 * @func isEven
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.18.0|v1.18.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isOdd|isOdd}
 * @example
 *
 * RA.isEven(0); // => true
 * RA.isEven(1); // => false
 * RA.isEven(-Infinity); // => false
 * RA.isEven(4); // => true
 * RA.isEven(3); // => false
 */
var isEven = /*#__PURE__*/ramda.both(isInteger, /*#__PURE__*/ramda.complement(isOdd));

/**
 * Checks if input value is a pair.
 *
 * @func isPair
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.19.0|v1.19.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link http://ramdajs.com/docs/#pair|R.pair}, {@link RA.isNotPair|isNotPair}
 * @example
 *
 * RA.isPair([]); // => false
 * RA.isPair([0]); // => false
 * RA.isPair([0, 1]); // => true
 * RA.isPair([0, 1, 2]); // => false
 * RA.isPair({ 0: 0, 1: 1 }); // => false
 * RA.isPair({ foo: 0, bar: 0 }); // => false
 */
var isPair = /*#__PURE__*/ramda.both(isArray, /*#__PURE__*/ramda.pipe(ramda.length, /*#__PURE__*/ramda.equals(2)));

/**
 * Checks if input value is complement of a pair.
 *
 * @func isNotPair
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.19.0|v1.19.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link http://ramdajs.com/docs/#pair|R.pair}, {@link RA.isPair|isPair}
 * @example
 *
 * RA.isNotPair([]); // => true
 * RA.isNotPair([0]); // => true
 * RA.isNotPair([0, 1]); // => false
 * RA.isNotPair([0, 1, 2]); // => true
 * RA.isNotPair({0: 0, 1: 1}); // => true
 * RA.isNotPair({foo: 0, bar: 0}); // => true
 */
var isNotPair = /*#__PURE__*/ramda.complement(isPair);

/**
 * Checks if input value is a `thenable`.
 * `thenable` is an object or function that defines a `then` method.
 *
 * @func isThenable
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.1.0|v2.1.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isPromise|isPromise}
 * @example
 *
 * RA.isThenable(null); // => false
 * RA.isThenable(undefined); // => false
 * RA.isThenable([]); // => false
 * RA.isThenable(Promise.resolve()); // => true
 * RA.isThenable(Promise.reject()); // => true
 * RA.isThenable({ then: () => 1 }); // => true
 */
var isThenable = /*#__PURE__*/ramda.pathSatisfies(isFunction, ['then']);

/**
 * Checks if input value is a native `Promise`.
 * The Promise object represents the eventual completion (or failure)
 * of an asynchronous operation, and its resulting value.
 *
 * @func isPromise
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.1.0|v2.1.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link https://promisesaplus.com/|Promises/A+}, {@link RA.isThenable|isThenable}
 * @example
 *
 * RA.isPromise(null); // => false
 * RA.isPromise(undefined); // => false
 * RA.isPromise([]); // => false
 * RA.isPromise(Promise.resolve()); // => true
 * RA.isPromise(Promise.reject()); // => true
 * RA.isPromise({ then: () => 1 }); // => false
 */
var isPromise = /*#__PURE__*/ramda.both(isObj, /*#__PURE__*/ramda.pipe(ramda.toString, /*#__PURE__*/ramda.equals('[object Promise]')));

/**
 * Checks if input value is the Boolean primitive `true`. Will return false for Boolean objects
 * created using the `Boolean` function as a constructor.
 *
 * @func isTrue
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.6.0|v2.6.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isFalse|isFalse}, {@link RA.isTruthy|isTruthy}, {@link RA.isFalsy|isFalsy}
 * @example
 *
 * RA.isTrue(true); // => true
 * RA.isTrue(Boolean(true)); // => true
 * RA.isTrue(false); // => false
 * RA.isTrue(1); // => false
 * RA.isTrue('true'); // => false
 * RA.isTrue(new Boolean(true)); // => false
 */

var isTrue = /*#__PURE__*/ramda.equals(true);

/**
 * Checks if input value is the Boolean primitive `false`. Will return false for all values created
 * using the `Boolean` function as a constructor.
 *
 * @func isFalse
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.6.0|v2.6.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isTrue|isTrue}, {@link RA.isTruthy|isTruthy}, {@link RA.isFalsy|isFalsy}
 * @example
 *
 * RA.isFalse(false); // => true
 * RA.isFalse(Boolean(false)); // => true
 * RA.isFalse(true); // => false
 * RA.isFalse(0); // => false
 * RA.isFalse(''); // => false
 * RA.isFalse(null); // => false
 * RA.isFalse(undefined); // => false
 * RA.isFalse(NaN); // => false
 * RA.isFalse([]); // => false
 * RA.isFalse(new Boolean(false)); // => false
 */

/*#__PURE__*/ramda.equals(false);

/**
 * In JavaScript, a `truthy` value is a value that is considered true
 * when evaluated in a Boolean context. All values are truthy unless
 * they are defined as falsy (i.e., except for `false`, `0`, `""`, `null`, `undefined`, and `NaN`).
 *
 * @func isTruthy
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.2.0|v2.2.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link https://developer.mozilla.org/en-US/docs/Glossary/Truthy|truthy}, {@link RA.isFalsy|isFalsy}
 * @example
 *
 * RA.isTruthy({}); // => true
 * RA.isTruthy([]); // => true
 * RA.isTruthy(42); // => true
 * RA.isTruthy(3.14); // => true
 * RA.isTruthy('foo'); // => true
 * RA.isTruthy(new Date()); // => true
 * RA.isTruthy(Infinity); // => true
 */
var isTruthy = Boolean;

/**
 * A falsy value is a value that translates to false when evaluated in a Boolean context.
 * Falsy values are `false`, `0`, `""`, `null`, `undefined`, and `NaN`.
 *
 * @func isFalsy
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.2.0|v2.2..0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link https://developer.mozilla.org/en-US/docs/Glossary/Falsy|falsy}, {@link RA.isTruthy|isTruthy}
 * @example
 *
 * RA.isFalsy(false); // => true
 * RA.isFalsy(0); // => true
 * RA.isFalsy(''); // => true
 * RA.isFalsy(null); // => true
 * RA.isFalsy(undefined); // => true
 * RA.isFalsy(NaN); // => true
 */
var isFalsy = /*#__PURE__*/ramda.complement(isTruthy);

/**
 * Checks if value is `RegExp` object.
 *
 * @func isRegExp
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.5.0|v2.5.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isNotRegExp|isNotRegExp}
 * @example
 *
 * RA.isRegExp(new RegExp()); //=> true
 * RA.isRegExp(/(?:)/); //=> true
 * RA.isRegExp(1); //=> false
 */
var isRegExp = /*#__PURE__*/ramda.is(RegExp);

/**
 * Checks if value is complement of `RegExp` object.
 *
 * @func isNotRegExp
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.5.0|v2.5.0}
 * @category Type
 * @sig * -> Boolean
 * @param {*} val The value to test
 * @return {boolean}
 * @see {@link RA.isRegExp|isRegExp}
 * @example
 *
 * RA.isNotRegExp(1); //=> true
 * RA.isNotRegExp(/(?:)/); //=> false
 * RA.isNotRegExp(new RegExp()); //=> false
 */
var isNotRegExp = /*#__PURE__*/ramda.complement(isRegExp);

/**
 * A function that returns `null`.
 *
 * @func stubNull
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.6.0|v1.6.0}
 * @category Function
 * @sig ... -> null
 * @return {null}
 * @example
 *
 * RA.stubNull(); //=> null
 * RA.stubNull(1, 2, 3); //=> null
 */
var stubNull = /*#__PURE__*/ramda.always(null);

/**
 * This function returns a new empty object.
 *
 * @func stubObj
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.1.0|v2.1.0}
 * @category Function
 * @sig ... -> Object
 * @aliases stubObject
 * @return {Object} Returns the new empty object.
 * @example
 *
 * RA.stubObj(); //=> {}
 * RA.stubObj(1, 2, 3); //=> {}
 */

/**
 * A function that returns empty string.
 *
 * @func stubString
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.1.0|v2.1.0}
 * @category Function
 * @sig ... -> String
 * @return {string} The empty string
 * @example
 *
 * RA.stubString(); //=> ''
 * RA.stubString(1, 2, 3); //=> ''
 */
var stubString = /*#__PURE__*/ramda.always('');

/**
 * A function that returns new empty array on every call.
 *
 * @func stubArray
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.1.0|v2.1.0}
 * @category Function
 * @sig ... -> Array
 * @return {Array} New empty array
 * @example
 *
 * RA.stubArray(); //=> []
 * RA.stubArray(1, 2, 3); //=> []
 */

/**
 * A function that performs no operations.
 *
 * @func noop
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.0.0|v1.0.0}
 * @category Function
 * @sig ... -> undefined
 * @return {undefined}
 * @example
 *
 * RA.noop(); //=> undefined
 * RA.noop(1, 2, 3); //=> undefined
 */
var noop = /*#__PURE__*/ramda.always( /*#__PURE__*/stubUndefined());

var mapping = /*#__PURE__*/Object.freeze({
  equals: 'fantasy-land/equals',
  lte: 'fantasy-land/lte',
  compose: 'fantasy-land/compose',
  id: 'fantasy-land/id',
  concat: 'fantasy-land/concat',
  empty: 'fantasy-land/empty',
  map: 'fantasy-land/map',
  contramap: 'fantasy-land/contramap',
  ap: 'fantasy-land/ap',
  of: 'fantasy-land/of',
  alt: 'fantasy-land/alt',
  zero: 'fantasy-land/zero',
  reduce: 'fantasy-land/reduce',
  traverse: 'fantasy-land/traverse',
  chain: 'fantasy-land/chain',
  chainRec: 'fantasy-land/chainRec',
  extend: 'fantasy-land/extend',
  extract: 'fantasy-land/extract',
  bimap: 'fantasy-land/bimap',
  promap: 'fantasy-land/promap'
});

var isFunctor = /*#__PURE__*/ramda.either( /*#__PURE__*/ramda.pathSatisfies(isFunction, ['map']), /*#__PURE__*/ramda.pathSatisfies(isFunction, [mapping.map]));
var isApply = /*#__PURE__*/ramda.both(isFunctor, /*#__PURE__*/ramda.either( /*#__PURE__*/ramda.pathSatisfies(isFunction, ['ap']), /*#__PURE__*/ramda.pathSatisfies(isFunction, [mapping.ap])));

var ap$1 = /*#__PURE__*/ramda.curryN(2, function (applyF, applyX) {
  // return original ramda `ap` if not Apply spec
  if (!isApply(applyF) || !isApply(applyX)) {
    return ramda.ap(applyF, applyX);
  }

  try {
    // new version of `ap` starting from ramda version > 0.23.0
    return applyF.ap(applyX);
  } catch (e) {
    // old version of `ap` till ramda version <= 0.23.0
    return applyX.ap(applyF);
  }
});

/**
 * "lifts" a function to be the specified arity, so that it may "map over" objects that satisfy
 * the fantasy land Apply spec of algebraic structures.
 *
 * Lifting is specific for {@link https://github.com/scalaz/scalaz|scalaz} and {@link http://www.functionaljava.org/|functional java} implementations.
 * Old version of fantasy land spec were not compatible with this approach,
 * but as of fantasy land 1.0.0 Apply spec also adopted this approach.
 *
 * This function acts as interop for ramda <= 0.23.0 and {@link https://monet.github.io/monet.js/|monet.js}.
 *
 * More info {@link https://github.com/fantasyland/fantasy-land/issues/50|here}.
 *
 * @func liftFN
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.2.0|v1.2.0}
 * @category Function
 * @sig Apply a => Number -> (a... -> a) -> (a... -> a)
 * @param {Number} arity The arity of the lifter function
 * @param {Function} fn The function to lift into higher context
 * @return {Function} The lifted function
 * @see {@link http://ramdajs.com/docs/#lift|R.lift}, {@link http://ramdajs.com/docs/#ap|R.ap}
 * @example
 *
 * const { Maybe } = require('monet');
 *
 * const add3 = (a, b, c) => a + b + c;
 * const madd3 = RA.liftFN(3, add3);
 *
 * madd3(Maybe.Some(10), Maybe.Some(15), Maybe.Some(17)); //=> Maybe.Some(42)
 * madd3(Maybe.Some(10), Maybe.Nothing(), Maybe.Some(17)); //=> Maybe.Nothing()
 */
var liftFN = /*#__PURE__*/ramda.curry(function (arity, fn) {
  var lifted = ramda.curryN(arity, fn);
  return ramda.curryN(arity, function () {
    for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }

    var accumulator = ramda.map(lifted, ramda.head(args));
    var apps = ramda.slice(1, Infinity, args);
    return ramda.reduce(ap$1, accumulator, apps);
  });
});

/* eslint-disable max-len */
/**
 * The catamorphism is a way of folding a type into a value.
 *
 * **Either**
 *
 * If the either is right than the right function will be executed with
 * the `right` value and the value of the function returned. Otherwise the left function
 * will be called with the `left` value.
 *
 * **Maybe**
 *
 * If the maybe is Some than the right function will be executed with the `some` value and the value of the function
 * returned. Otherwise the left function with be called without an argument.
 *
 *
 *
 * @func cata
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.4.0|v1.4.0}
 * @category Function
 * @sig (a -> b) -> (a -> c) -> Cata a -> b | c
 * @param {Function} leftFn The left function that consumes the left value
 * @param {Function} rightFn The right function that consumes the right value
 * @param {Cata} catamorphicObj Either, Maybe or any other type with catamorphic capabilities (`cata` or `either` method)
 * @return {*}
 * @see {@link https://monet.github.io/monet.js/#cata|cata explained}
 * @example
 *
 * // Either
 * const eitherR = Either.Right(1);
 * const eitherL = Either.Left(2);
 *
 * RA.cata(identity, identity, eitherR); //=> 1
 * RA.cata(identity, identity, eitherL); //=> 2
 *
 * // Maybe
 * const maybeSome = Maybe.Some(1);
 * const maybeNothing = Maybe.Nothing();
 *
 * RA.cata(identity, identity, maybeSome); //=> 1
 * RA.cata(identity, identity, maybeNothing); //=> undefined
 */
/* eslint-enable */
var catamorphism = /*#__PURE__*/ramda.curry(function (leftFn, rightFn, catamorphicObj) {
  if (isFunction(catamorphicObj.cata)) {
    return catamorphicObj.cata(leftFn, rightFn);
  }
  return catamorphicObj.either(leftFn, rightFn);
});

/**
 * Weaves a configuration into function returning the runnable monad like `Reader` or `Free`.
 * This allows us to pre-bind the configuration in advance and use the weaved function
 * without need to explicitly pass the configuration on every call.
 *
 * @func weave
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.7.0|v1.7.0}
 * @category Function
 * @sig (*... -> *) -> * -> (*... -> *)
 * @param {Function} fn The function to weave
 * @param {*} config The configuration to weave into fn
 * @return {Function} Auto-curried weaved function
 * @example
 *
 * const { Reader: reader } = require('monet');
 *
 * const log = value => reader(
 *   config => config.log(value)
 * );
 *
 * // no weaving
 * log('test').run(console); //=> prints 'test'
 *
 * // weaving
 * const wlog = RA.weave(log, console);
 * wlog('test'); //=> prints 'test'
 */
var weave = /*#__PURE__*/ramda.curryN(2, function (fn, config) {
  return ramda.curryN(fn.length, function () {
    return fn.apply(undefined, arguments).run(config);
  });
});

/**
 * Weaves a configuration into function returning the runnable monad like `Reader` or `Free`.
 * This allows us to pre-bind the configuration in advance and use the weaved function
 * without need to explicitly pass the configuration on every call.
 *
 * @func weaveLazy
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.10.0|v1.10.0}
 * @category Function
 * @sig (*... -> *) -> (* -> *) -> (*... -> *)
 * @param {Function} fn The function to weave
 * @param {Function} configAccessor The function that returns the configuration object
 * @return {Function} Auto-curried weaved function
 * @example
 *
 * const { Reader: reader } = require('monet');
 *
 * const log = value => reader(
 *   config => config.log(value)
 * );
 *
 * const consoleAccessor = R.always(console);
 *
 * // no weaving
 * log('test').run(console); //=> prints 'test'
 *
 * // weaving
 * const wlog = RA.weaveLazy(log, consoleAccessor);
 * wlog('test'); //=> prints 'test'
 */
var weaveLazy = /*#__PURE__*/ramda.curryN(2, function (fn, configAccessor) {
  return ramda.curryN(fn.length, function () {
    return fn.apply(undefined, arguments).run(configAccessor());
  });
});

/**
 * Returns a curried equivalent of the provided function, with the specified arity.
 * This function is like curryN, except that the provided arguments order is reversed.
 *
 * @func curryRightN
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.12.0|v1.12.0}
 * @category Function
 * @sig Number -> (* -> a) -> (* -> a)
 * @param {Number} length The arity for the returned function
 * @param {Function} fn The function to curry
 * @return {Function}  A new, curried function
 * @see {@link http://ramdajs.com/docs/#curryN|R.curryN}, {@link RA.curryRight|curryRight}
 * @example
 *
 * const concatStrings = (a, b, c) => a + b + c;
 * const concatStringsCurried = RA.curryRightN(3, concatStrings);
 *
 * concatStringCurried('a')('b')('c'); // => 'cba'
 */
var curryRightN = /*#__PURE__*/ramda.curryN(2, function (arity, fn) {
  return ramda.curryN(arity, function wrapper() {
    for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }

    return fn.apply(this, ramda.reverse(args));
  });
});

/**
 * Returns a curried equivalent of the provided function.
 * This function is like curry, except that the provided arguments order is reversed.
 *
 * @func curryRight
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.12.0|v1.12.0}
 * @category Function
 * @sig (* -> a) -> (* -> a)
 * @param {Function} fn The function to curry
 * @return {Function}  A new, curried function
 * @see {@link http://ramdajs.com/docs/#curry|R.curry}, {@link RA.curryRightN|curryRightN}
 * @example
 *
 * const concatStrings = (a, b, c) => a + b + c;
 * const concatStringsCurried = RA.curryRight(concatStrings);
 *
 * concatStringCurried('a')('b')('c'); // => 'cba'
 */
var curryRight = /*#__PURE__*/ramda.converge(curryRightN, [ramda.length, ramda.identity]);

/**
 * Composable shortcut for `Promise.all`.
 *
 * The `allP` method returns a single Promise that resolves when all of the promises
 * in the iterable argument have resolved or when the iterable argument contains no promises.
 * It rejects with the reason of the first promise that rejects.
 *
 * @func allP
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.3.0|v2.3.0}
 * @category Function
 * @sig [Promise a] -> [a]
 * @param {Iterable.<*>} iterable An iterable object such as an Array or String
 * @return {Promise} An already resolved Promise if the iterable passed is empty. An asynchronously resolved Promise if the iterable passed contains no promises. Note, Google Chrome 58 returns an already resolved promise in this case. A pending Promise in all other cases. This returned promise is then resolved/rejected asynchronously (as soon as the stack is empty) when all the promises in the given iterable have resolved, or if any of the promises reject. See the example about "Asynchronicity or synchronicity of allP" below.
 * @see {@link RA.resolveP|resolveP}, {@link RA.rejectP|rejectP}
 * @example
 *
 * RA.allP([1, 2]); //=> Promise([1, 2])
 * RA.allP([1, Promise.resolve(2)]); //=> Promise([1, 2])
 * RA.allP([Promise.resolve(1), Promise.resolve(2)]); //=> Promise([1, 2])
 * RA.allP([1, Promise.reject(2)]); //=> Promise(2)
 */
var allP = /*#__PURE__*/ramda.bind(Promise.all, Promise);

/* eslint-disable max-len */
/**
 * Composable shortcut for `Promise.resolve`.
 *
 * Returns a Promise object that is resolved with the given value.
 * If the value is a thenable (i.e. has a "then" method), the returned promise will
 * "follow" that thenable, adopting its eventual state.
 *
 * @func resolveP
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.16.0|v1.16.0}
 * @category Function
 * @sig a -> Promise a
 * @param {*} [value=undefined] Argument to be resolved by this Promise. Can also be a Promise or a thenable to resolve
 * @return {Promise} A Promise that is resolved with the given value, or the promise passed as value, if the value was a promise object
 * @see {@link RA.rejectP|rejectP}
 * @example
 *
 * RA.resolveP(); //=> Promise(undefined)
 * RA.resolveP('a'); //=> Promise('a')
 * RA.resolveP([1, 2, 3]); //=> Promise([1, 2, 3])
 */
/* eslint-enable max-len */
var resolveP = /*#__PURE__*/ramda.bind(Promise.resolve, Promise);

/**
 * Composable shortcut for `Promise.reject`.
 *
 * Returns a Promise object that is rejected with the given reason.
 *
 * @func rejectP
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.16.0|v1.16.0}
 * @category Function
 * @sig a -> Promise a
 * @param {*} [reason=undefined] Reason why this Promise rejected
 * @return {Promise} A Promise that is rejected with the given reason
 * @see {@link RA.resolveP|resolveP}
 * @example
 *
 * RA.rejectP(); //=> Promise(undefined)
 * RA.rejectP('a'); //=> Promise('a')
 * RA.rejectP([1, 2, 3]); //=> Promise([1, 2, 3])
 */
var rejectP = /*#__PURE__*/ramda.bind(Promise.reject, Promise);

/**
 * Composable shortcut for `Promise.then`.
 * The thenP function returns a Promise. It takes two arguments: a callback function for the success of the Promise
 * and the promise instance itself.
 *
 * @func thenP
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.8.0|v2.8.0}
 * @category Function
 * @sig (a -> Promise b | b) -> Promise b
 * @param {Function} onFulfilled A Function called if the Promise is fulfilled. This function has one argument, the fulfillment value
 * @param {Promise} promise Any Promise or Thenable object
 * @return {Promise} A Promise in the pending status

 * @see {@link RA.resolveP|resolveP}, {@link RA.rejectP|rejectP}, {@link RA.allP|allP}
 * @example
 *
 * const promise = Promise.resolve(1);
 * const add1 = v => v + 1;
 *
 * RA.thenP(add1, promise); // => Promise(2)
 */
var thenP = /*#__PURE__*/ramda.invoker(1, 'then');

/**
 * Y-combinator
 *
 * The Y combinator is an interesting function which only works with functional languages,
 * showing how recursion can still be done even without any variable or function declarations,
 * only functions and parameters
 *
 * @func Y
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.3.0|v2.3.0}
 * @category Function
 * @sig (a, ... -> b -> b) -> (a, ... -> b)
 * @param {Function} le Recursive function maker
 * @return {Function}
 * @see {@link http://kestas.kuliukas.com/YCombinatorExplained/|Y combinator explained}
 * @example
 *
 * const makeFact = givenFact => (n) => {
 *   if (n < 2) { return 1 }
 *   return n * givenFact(n - 1);
 * };
 *
 * const factorial = RA.Y(makeFact);
 *
 * factorial(5); //=> 120
 */

/* eslint-disable max-len */
/**
 * Runs the given list of functions in order with the supplied object, then returns the object.
 * Also known as the normal order sequencing combinator.
 *
 * Acts as a transducer if a transformer is given as second parameter.
 *
 * @func seq
 * @aliases sequencing
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.3.0|v2.3.0}
 * @category Function
 * @sig [(a -> *), (a -> *), ...] -> a -> a
 * @param {Array} fns The list of functions to call in order with `x` whose return values will be thrown away
 * @param {*} x
 * @return {*} `x`
 * @see {@link http://ramdajs.com/docs/#tap|R.tap}, {@link http://www.cs.rpi.edu/academics/courses/spring11/proglang/handouts/lambda-calculus-chapter.pdf|sequencing combinator explained}
 * @example
 *
 * RA.seq([console.info, console.log])('foo'); //=> prints 'foo' via info then log
 *
 * // usage in composition
 * R.pipe(
 *   R.concat('prefix '),
 *   RA.seq([
 *     console.info, //=> prints 'prefix test'
 *     console.log //=> prints 'prefix test'
 *   ]),
 *   R.toUpper
 * )('test'); //=> 'PREFIX TEST'
 */
/* eslint-enable max-len */

var seq = /*#__PURE__*/ramda.curry(function (fns, x) {
  return ramda.tap(function (tx) {
    return ramda.map(function (fn) {
      return fn(tx);
    })(fns);
  })(x);
});

/**
 * Can be used as a way to compose multiple invokers together to form polymorphic functions,
 * or functions that exhibit different behaviors based on their argument(s).
 * Consumes dispatching functions and keep trying to invoke each in turn, until a non-nil value is returned.
 *
 * Accepts a list of dispatching functions and returns a new function.
 * When invoked, this new function is applied to some arguments,
 * each dispatching function is applied to those same arguments until one of the
 * dispatching functions returns a non-nil value.
 *
 * @func dispatch
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.6.0|v2.6.0}
 * @category Function
 * @sig [((a, b, ...) -> x1), ((a, b, ...) -> x2), ...] -> x1 | x2 | ...
 * @param {!Array} functions A list of functions
 * @return {*|undefined} Returns the first not-nil value, or undefined if either an empty list is provided or none of the dispatching functions returns a non-nil value
 * @see {@link RA.isNotNil}
 * @example
 *
 * // returns first non-nil value
 * const stubNil = () => null;
 * const stubUndefined = () => undefined;
 * const addOne = v => v + 1;
 * const addTwo = v => v + 2;
 *
 * RA.dispatch([stubNil, stubUndefined, addOne, addTwo])(1); //=> 2
 *
 * // acts as a switch
 * const fnSwitch = RA.dispatch([
 *   R.ifElse(RA.isString, s => `${s}-join`, RA.stubUndefined),
 *   R.ifElse(RA.isNumber, n => n + 1, RA.stubUndefined),
 *   R.ifElse(RA.isDate, R.T, RA.stubUndefined),
 * ]);
 * fnSwitch(1); //=> 2
 */
var byArity = /*#__PURE__*/ramda.comparator(function (a, b) {
  return a.length > b.length;
});

var getMaxArity = /*#__PURE__*/ramda.pipe( /*#__PURE__*/ramda.sort(byArity), ramda.head, /*#__PURE__*/ramda.prop('length'));

var iteratorFn = /*#__PURE__*/ramda.curry(function (args, accumulator, fn) {
  var result = fn.apply(undefined, args);

  return isNotNil(result) ? ramda.reduced(result) : accumulator;
});

var dispatch = function dispatch(functions) {
  var arity = getMaxArity(functions);

  return ramda.curryN(arity, function () {
    for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }

    return ramda.reduce(iteratorFn(args), undefined, functions);
  });
};

/*#__PURE__*/ramda.ifElse(isNonEmptyArray, dispatch, stubUndefined);

/**
 * {@link http://ramdajs.com/docs/#map|R.map} function that more closely resembles Array.prototype.map.
 * It takes two new parameters to its callback function: the current index, and the entire list.
 *
 * `mapIndexed` implementation is simple : `
 * const mapIndexed = R.addIndex(R.map);
 * `
 * @func mapIndexed
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.5.0|v2.5.0}
 * @category List
 * @typedef Idx = Number
 * @sig Functor f => ((a, Idx, f a) => b) => f a -> f b
 * @param {Function} fn The function to be called on every element of the input `list`
 * @param {Array} list The list to be iterated over
 * @return {Array} The new list
 * @see {@link http://ramdajs.com/docs/#addIndex|R.addIndex}, {@link http://ramdajs.com/docs/#map|R.map}
 * @example
 *
 * RA.mapIndexed((val, idx, list) => idx + '-' + val, ['f', 'o', 'o', 'b', 'a', 'r']);
 * //=> ['0-f', '1-o', '2-o', '3-b', '4-a', '5-r']
 */
var mapIndexed = /*#__PURE__*/ramda.addIndex(ramda.map);

/**
 * {@link http://ramdajs.com/docs/#reduce|R.reduce} function that more closely resembles Array.prototype.reduce.
 * It takes two new parameters to its callback function: the current index, and the entire list.
 *
 * `reduceIndexed` implementation is simple : `
 * const reduceIndexed = R.addIndex(R.reduce);
 * `
 * @func reduceIndexed
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.5.0|v2.5.0}
 * @category List
 * @typedef Idx = Number
 * @sig ((a, b, Idx, [b]) => a) -> a -> [b] -> a
 * @param {Function} fn The iterator function. Receives four values,
 * the accumulator, the current element from the array, index and the entire list
 * @param {*} acc The accumulator value
 * @param {Array} list The list to iterate over
 * @return {*} The final, accumulated value
 * @see {@link http://ramdajs.com/docs/#addIndex|R.addIndex}, {@link http://ramdajs.com/docs/#reduce|R.reduce}
 * @example
 *
 * const initialList = ['f', 'o', 'o', 'b', 'a', 'r'];
 *
 * reduceIndexed((acc, val, idx, list) => acc + '-' + val + idx, '', initialList);
 * //=> "-f0-o1-o2-b3-a4-r5"
 */
var reduceIndexed = /*#__PURE__*/ramda.addIndex(ramda.reduce);

// helpers
var filterIndexed = /*#__PURE__*/ramda.addIndex(ramda.filter);
var containsIndex = /*#__PURE__*/ramda.curry(function (indexes, val, index) {
  return ramda.contains(index, indexes);
});

/**
 * Picks values from list by indexes.
 *
 * @func pickIndexes
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.1.0|v1.1.0}
 * @category List
 * @sig  [Number] -> [a] -> [a]
 * @param {Array} indexes The indexes to pick
 * @param {Array} list The list to pick values from
 * @return {Array} New array containing only values at `indexes`
 * @see {@link http://ramdajs.com/docs/#pick|R.pick}, {@link RA.omitIndexes|omitIndexes}
 * @example
 *
 * RA.pickIndexes([0, 2], ['a', 'b', 'c']); //=> ['a', 'c']
 */
var pickIndexes = /*#__PURE__*/ramda.curry(function (indexes, list) {
  return filterIndexed(containsIndex(indexes), list);
});

/**
 * Creates a list from arguments.
 *
 * @func list
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.1.0|v1.1.0}
 * @category List
 * @sig  a... -> [a...]
 * @param {...*} items The items of the feature list
 * @return {Array} New list created from items
 * @see {@link https://github.com/ramda/ramda/wiki/Cookbook#create-a-list-function|Ramda Cookbook}
 * @example
 *
 * RA.list('a', 'b', 'c'); //=> ['a', 'b', 'c']
 */
var list = /*#__PURE__*/ramda.unapply(ramda.identity);

/**
 * Returns a singleton array containing the value provided.
 * If value is already an array, it is returned as is.
 *
 * @func ensureArray
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.6.0|v2.6.0}
 * @category List
 * @sig a | [a] -> [a]
 * @param {*|Array} val the value ensure as Array
 * @return {Array}
 * @see {@link http://ramdajs.com/docs/#of|R.of}
 * @example
 *
 * RA.ensureArray(42); //=> [42]
 * RA.ensureArray([42]); //=> [42]
 */
var ensureArray = /*#__PURE__*/ramda.when(isNotArray, ramda.of);

var leftIdentitySemigroup = { concat: ramda.identity };

/**
 * Returns the result of concatenating the given lists or strings.
 * Note: RA.concatAll expects all elements to be of the same type. It will throw an error if you concat an Array with a non-Array value.
 * Dispatches to the concat method of the preceding element, if present. Can also concatenate multiple elements of a [fantasy-land compatible semigroup](https://github.com/fantasyland/fantasy-land#semigroup).
 * Returns undefined if empty array was passed.
 *
 * @func concatAll
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.6.0|v2.6.0}
 * @category List
 * @sig [[a]] -> [a] | Undefined
 * @sig [String] -> String | Undefined
 * @sig Semigroup s => Foldable s f => f -> s | Undefined
 * @param {Array.<Array|string>} list List containing elements that will be concatenated
 * @return {Array|string|undefined} Concatenated elements
 * @see {@link http://ramdajs.com/docs/#concat|R.concat}, {@link RA.concatRight|concatRight}, {@link http://ramdajs.com/docs/#unnest|R.unnest}, {@link http://ramdajs.com/docs/#join|R.join}
 * @example
 *
 * concatAll([[1], [2], [3]]); //=> [1, 2, 3]
 * concatAll(['1', '2', '3']); //=> '123'
 * concatAll([]); //=> undefined;
 */
var concatAll = /*#__PURE__*/ramda.pipe( /*#__PURE__*/ramda.reduce(ramda.concat, leftIdentitySemigroup), /*#__PURE__*/ramda.when( /*#__PURE__*/ramda.identical(leftIdentitySemigroup), stubUndefined));

/**
 * Returns the result of concatenating the given lists or strings.
 *
 * Note: R.concat expects both arguments to be of the same type, unlike
 * the native Array.prototype.concat method.
 * It will throw an error if you concat an Array with a non-Array value.
 * Dispatches to the concat method of the second argument, if present.
 *
 * @func concatRight
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.11.0|v1.11.0}
 * @category List
 * @sig [a] -> [a] -> [a]
 * @sig String -> String -> String
 * @param {Array|String} firstList The first list
 * @param {Array|String} secondList The second list
 * @return {Array|String} A list consisting of the elements of `secondList`
 * followed by the elements of `firstList`.
 * @see {@link http://ramdajs.com/docs/#concat|R.concat}
 * @example
 *
 * RA.concatRight('ABC', 'DEF'); //=> 'DEFABC'
 * RA.concatRight([4, 5, 6], [1, 2, 3]); //=> [1, 2, 3, 4, 5, 6]
 * RA.concatRight([], []); //=> []
 */
var concatRight = /*#__PURE__*/ramda.flip(ramda.concat);

/* eslint-disable max-len */
/**
 * Given an `Iterable`(arrays are `Iterable`), or a promise of an `Iterable`,
 * which produces promises (or a mix of promises and values),
 * iterate over all the values in the `Iterable` into an array and
 * reduce the array to a value using the given iterator function.
 *
 * If the iterator function returns a promise, then the result of the promise is awaited,
 * before continuing with next iteration. If any promise in the array is rejected or a promise
 * returned by the iterator function is rejected, the result is rejected as well.
 *
 * If `initialValue` is `undefined` (or a promise that resolves to `undefined`) and
 * the `Iterable` contains only 1 item, the callback will not be called and
 * the `Iterable's` single item is returned. If the `Iterable` is empty, the callback
 * will not be called and `initialValue` is returned (which may be undefined).
 *
 * This function is basically equivalent to {@link http://bluebirdjs.com/docs/api/promise.reduce.html|bluebird.reduce}.
 *
 * @func reduceP
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.13.0|v1.13.0}
 * @category List
 * @typedef MaybePromise = Promise.<*> | *
 * @sig ((Promise a, MaybePromise b) -> Promise a) -> MaybePromise a -> MaybePromise [MaybePromise b] -> Promise a
 * @param {Function} fn The iterator function. Receives two values, the accumulator and the current element from the list
 * @param {*|Promise.<*>} acc The accumulator value
 * @param {Array.<*>|Promise.<Array<*|Promise.<*>>>} list The list to iterate over
 * @return {Promise} The final, accumulated value
 * @see {@link http://ramdajs.com/docs/#reduce|R.reduce}, {@link RA.reduceRightP|reduceRightP}, {@link http://bluebirdjs.com/docs/api/promise.reduce.html|bluebird.reduce}
 * @example
 *
 * RA.reduceP(
 *   (total, fileName) => fs
 *     .readFileAsync(fileName, 'utf8')
 *     .then(contents => total + parseInt(contents, 10)),
 *   0,
 *   ['file1.txt', 'file2.txt', 'file3.txt']
 * ); // => Promise(10)
 *
 * RA.reduceP(
 *   (total, fileName) => fs
 *     .readFileAsync(fileName, 'utf8')
 *     .then(contents => total + parseInt(contents, 10)),
 *   Promise.resolve(0),
 *   ['file1.txt', 'file2.txt', 'file3.txt']
 * ); // => Promise(10)
 *
 * RA.reduceP(
 *   (total, fileName) => fs
 *     .readFileAsync(fileName, 'utf8')
 *     .then(contents => total + parseInt(contents, 10)),
 *   0,
 *   [Promise.resolve('file1.txt'), 'file2.txt', 'file3.txt']
 * ); // => Promise(10)
 *
 * RA.reduceP(
 *   (total, fileName) => fs
 *     .readFileAsync(fileName, 'utf8')
 *     .then(contents => total + parseInt(contents, 10)),
 *   0,
 *   Promise.resolve([Promise.resolve('file1.txt'), 'file2.txt', 'file3.txt'])
 * ); // => Promise(10)
 *
 */
/* esline-enable max-len */
var reduceP = /*#__PURE__*/ramda.curryN(3, function (fn, acc, list) {
  return resolveP(list).then(function (iterable) {
    var listLength = ramda.length(iterable);

    if (listLength === 0) {
      return acc;
    }

    var reducer = ramda.reduce(function (accP, currentValueP) {
      return accP.then(function (previousValue) {
        return allP([previousValue, currentValueP]);
      }).then(function (_ref) {
        var previousValue = _ref[0],
            currentValue = _ref[1];

        if (isUndefined(previousValue) && listLength === 1) {
          return currentValue;
        }

        return fn(previousValue, currentValue);
      });
    });

    return reducer(resolveP(acc), iterable);
  });
});

// in older ramda versions the order of the arguments is flipped
var flipArgs = /*#__PURE__*/ramda.pipe(ramda.reduceRight(ramda.concat, ''), ramda.equals('ba'))(['a', 'b']);

/* eslint-disable max-len */
/**
 * Given an `Iterable`(arrays are `Iterable`), or a promise of an `Iterable`,
 * which produces promises (or a mix of promises and values),
 * iterate over all the values in the `Iterable` into an array and
 * reduce the array to a value using the given iterator function.
 *
 * Similar to {@link RA.reduceP|reduceP} except moves through the input list from the right to the left.
 * The iterator function receives two values: (value, acc),
 * while the arguments' order of reduceP's iterator function is (acc, value).
 *
 * @func reduceRightP
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.13.0|v1.13.0}
 * @category List
 * @typedef MaybePromise = Promise.<*> | *
 * @sig ((MaybePromise b, Promise a) -> Promise a) -> MaybePromise a -> MaybePromise [MaybePromise b] -> Promise a
 * @param {Function} fn The iterator function. Receives two values, the current element from the list and the accumulator
 * @param {*|Promise.<*>} acc The accumulator value
 * @param {Array.<*>|Promise.<Array<*|Promise.<*>>>} list The list to iterate over
 * @return {Promise} The final, accumulated value
 * @see {@link RA.reduceP|reduceP}, {@link http://bluebirdjs.com/docs/api/promise.reduce.html|bluebird.reduce}
 * @example
 *
 * RA.reduceRightP(
 *   (fileName, total) => fs
 *     .readFileAsync(fileName, 'utf8')
 *     .then(contents => total + parseInt(contents, 10)),
 *   0,
 *   ['file1.txt', 'file2.txt', 'file3.txt']
 * ); // => Promise(10)
 *
 * RA.reduceRightP(
 *   (fileName, total) => fs
 *     .readFileAsync(fileName, 'utf8')
 *     .then(contents => total + parseInt(contents, 10)),
 *   Promise.resolve(0),
 *   ['file1.txt', 'file2.txt', 'file3.txt']
 * ); // => Promise(10)
 *
 * RA.reduceRightP(
 *   (fileName, total) => fs
 *     .readFileAsync(fileName, 'utf8')
 *     .then(contents => total + parseInt(contents, 10)),
 *   0,
 *   [Promise.resolve('file1.txt'), 'file2.txt', 'file3.txt']
 * ); // => Promise(10)
 *
 * RA.reduceRightP(
 *   (fileName, total) => fs
 *     .readFileAsync(fileName, 'utf8')
 *     .then(contents => total + parseInt(contents, 10)),
 *   0,
 *   Promise.resolve([Promise.resolve('file1.txt'), 'file2.txt', 'file3.txt'])
 * ); // => Promise(10)
 *
 */
/* esline-enable max-len */
var reduceRightP = /*#__PURE__*/ramda.curryN(3, function (fn, acc, list) {
  return resolveP(list).then(function (iterable) {
    var listLength = ramda.length(iterable);

    if (listLength === 0) {
      return acc;
    }

    var reducer = ramda.reduceRight(function (arg1, arg2) {
      var accP = void 0;
      var currentValueP = void 0;

      if (flipArgs) {
        accP = arg1;
        currentValueP = arg2;
      } else {
        accP = arg2;
        currentValueP = arg1;
      }

      return accP.then(function (previousValue) {
        return allP([previousValue, currentValueP]);
      }).then(function (_ref) {
        var previousValue = _ref[0],
            currentValue = _ref[1];

        if (isUndefined(previousValue) && listLength === 1) {
          return currentValue;
        }

        return fn(currentValue, previousValue);
      });
    });

    return reducer(resolveP(acc), iterable);
  });
});

/**
 * Returns the elements of the given list or string (or object with a slice method)
 * from fromIndex (inclusive).
 * Dispatches to the slice method of the second argument, if present.
 *
 * @func sliceFrom
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.16.0|v1.16.0}
 * @category List
 * @sig  Number -> [a] -> [a]
 * @param {number} fromIndex The start index (inclusive)
 * @param {Array|string} list The list or string to slice
 * @return {Array|string} The sliced list or string
 * @see {@link http://ramdajs.com/docs/#slice|R.slice}, {@link RA.sliceTo|sliceTo}
 * @example
 *
 * RA.sliceFrom(1, [1, 2, 3]); //=> [2, 3]
 */
var sliceFrom = /*#__PURE__*/ramda.slice(ramda.__, Infinity);

/**
 * Returns the elements of the given list or string (or object with a slice method)
 * to toIndex (exclusive).
 * Dispatches to the slice method of the second argument, if present.
 *
 * @func sliceTo
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.16.0|v1.16.0}
 * @category List
 * @sig  Number -> [a] -> [a]
 * @param {number} toIndex The end index (exclusive)
 * @param {Array|string} list The list or string to slice
 * @return {Array|string} The sliced list or string
 * @see {@link http://ramdajs.com/docs/#slice|R.slice}, {@link RA.sliceFrom|sliceFrom}
 * @example
 *
 * RA.sliceTo(2, [1, 2, 3]); //=> [1, 2]
 */
var sliceTo = /*#__PURE__*/ramda.slice(0);

// helpers
var rejectIndexed = /*#__PURE__*/ramda.addIndex(ramda.reject);
var containsIndex$1 = /*#__PURE__*/ramda.curry(function (indexes, val, index) {
  return ramda.contains(index, indexes);
});

/**
 * Returns a partial copy of an array omitting the indexes specified.
 *
 * @func omitIndexes
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.19.0|v1.19.0}
 * @category List
 * @sig [Int] -> [a] -> [a]
 * @see {@link http://ramdajs.com/docs/#omit|R.omit}, {@link RA.pickIndexes|pickIndexes}
 * @param {!Array} indexes The array of indexes to omit from the new array
 * @param {!Array} list The array to copy from
 * @return {!Array} The new array with omitted indexes
 * @example
 *
 * RA.omitIndexes([-1, 1, 3], ['a', 'b', 'c', 'd']); //=> ['a', 'c']
 */
var omitIndexes = /*#__PURE__*/ramda.curry(function (indexes, list) {
  return rejectIndexed(containsIndex$1(indexes), list);
});

/**
 * Creates an array with all falsy values removed.
 * The values false, null, 0, "", undefined, and NaN are falsy.
 *
 * @func compact
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.5.0|v2.5.0}
 * @category List
 * @sig Filterable f => f a -> f a
 * @param {Array} list The array to compact
 * @return {Array} Returns the new array of filtered values
 * @see {@link RA.isFalsy|isFalsy}
 * @example
 *
 * RA.compact([0, 1, false, 2, '', 3]); //=> [1, 2, 3]
 */
var compact = /*#__PURE__*/ramda.reject(isFalsy);

/**
 * Returns a new list containing the contents of the given list, followed by the given element.
 * Like {@link http://ramdajs.com/docs/#append|R.append} but with argument order reversed.
 *
 * @func appendFlipped
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.5.0|v2.5.0}
 * @category List
 * @sig [a] -> a -> [a]
 * @param {Array} list The list of elements to add a new item to
 * @param {*} el The element to add to the end of the new list
 * @return {Array} A new list containing the elements of the old list followed by `el`
 * @see {@link http://ramdajs.com/docs/#append|R.append}
 * @example
 *
 * RA.appendFlipped(['write', 'more'], 'tests'); //=> ['write', 'more', 'tests']
 * RA.appendFlipped([], 'tests'); //=> ['tests']
 * RA.appendFlipped(['write', 'more'], ['tests']); //=> ['write', 'more', ['tests']]
 */
var appendFlipped = /*#__PURE__*/ramda.flip(ramda.append);

/**
 * Returns true if the specified value is equal, in R.equals terms,
 * to at least one element of the given list or false otherwise.
 * Given list can be a string.
 *
 * Like {@link http://ramdajs.com/docs/#contains|R.contains} but with argument order reversed.
 *
 * @func contained
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.8.0|v2.8.0}
 * @category List
 * @sig [a] -> a -> Boolean
 * @param {Array|String} list The list to consider
 * @param {*} a The item to compare against
 * @return {boolean} Returns Boolean `true` if an equivalent item is in the list or `false` otherwise
 * @see {@link http://ramdajs.com/docs/#contains|R.contains}
 * @example
 *
 * RA.contained([1, 2, 3], 3); //=> true
 * RA.contained([1, 2, 3], 4); //=> false
 * RA.contained([{ name: 'Fred' }], { name: 'Fred' }); //=> true
 * RA.contained([[42]], [42]); //=> true
 */
var contained = /*#__PURE__*/ramda.flip(ramda.contains);

/**
 * Returns a new list with the item at the position `fromIdx` moved to the position `toIdx`. If the
 * `toIdx` is out of the `list` range, the item will be placed at the last position of the `list`.
 * When negative indices are provided, the behavior of the move is unspecified.
 *
 * @func move
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.8.0|v2.8.0}
 * @category List
 * @sig Number -> Number -> [a] -> [a]
 * @param {Number} fromIdx The position of item to be moved
 * @param {Number} toIdx The position of item after move
 * @param {Array} list The list containing the item to be moved
 * @return {Array}
 * @example
 *
 * const list = ['a', 'b', 'c', 'd', 'e'];
 * RA.move(1, 3, list) //=> ['a', 'c', 'd', 'b', 'e']
 */
var move = /*#__PURE__*/ramda.curry(function (fromIdx, toIdx, list) {
  return ramda.compose(ramda.insert(toIdx, ramda.nth(fromIdx, list)), ramda.remove(fromIdx, 1))(list);
});

var compareLength = /*#__PURE__*/ramda.curry(function (comparator$$1, value, list) {
  return ramda.compose(comparator$$1(value), ramda.length)(list);
});

/**
 * Returns `true` if the supplied list or string has a length greater than `valueLength`.
 *
 * @func lengthGt
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.8.0|v2.8.0}
 * @category List
 * @sig Number -> [*] -> Boolean
 * @param {number} valueLength The length of the list or string
 * @param {Array|string} value The list or string
 * @return {boolean}
 * @see {@link RA.lengthEq|lengthEq}, {@link RA.lengthNotEq|lengthNotEq}, {@link RA.lengthLt|lengthLt}, {@link RA.lengthLte|lengthLte}, {@link RA.lengthGte|lengthGte}, {@link http://ramdajs.com/docs/#gt|gt},  {@link http://ramdajs.com/docs/#length|length}
 * @example
 *
 * RA.lengthGt(3, [1,2,3,4]); //=> true
 * RA.lengthGt(3, [1,2,3]); //=> false
 */
var lengthGt = /*#__PURE__*/compareLength( /*#__PURE__*/ramda.flip(ramda.gt));

/**
 * Returns `true` if the supplied list or string has a length less than `valueLength`.
 *
 * @func lengthLt
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.8.0|v2.8.0}
 * @category List
 * @sig Number -> [*] -> Boolean
 * @param {number} valueLength The length of the list or string
 * @param {Array|string} value The list or string
 * @return {boolean}
 * @see {@link RA.lengthEq|lengthEq}, {@link RA.lengthNotEq|lengthNotEq}, {@link RA.lengthGt|lengthGt}, {@link RA.lengthLte|lengthLte}, {@link RA.lengthGte|lengthGte}, {@link http://ramdajs.com/docs/#lt|lt}, {@link http://ramdajs.com/docs/#length|length}
 * @example
 *
 * RA.lengthLt(3, [1,2]); //=> true
 * RA.lengthLt(3, [1,2,3]); //=> false
 */
var lengthLt = /*#__PURE__*/compareLength( /*#__PURE__*/ramda.flip(ramda.lt));

/**
 * Returns `true` if the supplied list or string has a length greater than or equal to
 * `valueLength`.
 *
 * @func lengthGte
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.8.0|v2.8.0}
 * @category List
 * @sig Number -> [*] -> Boolean
 * @param {number} valueLength The length of the list or string
 * @param {Array|string} value The list or string
 * @return {boolean}
 * @see {@link RA.lengthEq|lengthEq}, {@link RA.lengthNotEq|lengthNotEq}, {@link RA.lengthLt|lengthLt}, {@link RA.lengthGt|lengthGt}, {@link RA.lengthLte|lengthLte}, {@link http://ramdajs.com/docs/#gte|gte}, {@link http://ramdajs.com/docs/#length|length}
 * @example
 *
 * RA.lengthGte(3, [1,2,3,4]); //=> true
 * RA.lengthGte(3, [1,2,3]); //=> true
 * RA.lengthGte(3, [1,2,3]); //=> false
 */
var lengthGte = /*#__PURE__*/compareLength( /*#__PURE__*/ramda.flip(ramda.gte));

/**
 * Returns `true` if the supplied list or string has a length less than or equal to `valueLength`.
 *
 * @func lengthLte
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.8.0|v2.8.0}
 * @category List
 * @sig Number -> [*] -> Boolean
 * @param {number} valueLength The length of the list or string
 * @param {Array|string} value The list or string
 * @return {boolean}
 * @see {@link RA.lengthEq|lengthEq}, {@link RA.lengthNotEq|lengthNotEq}, {@link RA.lengthLt|lengthLt}, {@link RA.lengthGt|lengthGt}, {@link RA.lengthGte|lengthGte}, {@link http://ramdajs.com/docs/#lte|lte}, {@link http://ramdajs.com/docs/#length|length}
 * @example
 *
 * RA.lengthLte(3, [1,2]); //=> true
 * RA.lengthLte(3, [1,2,3]); //=> true
 * RA.lengthLte(3, [1,2,3,4]); //=> false
 */
var lengthLte = /*#__PURE__*/compareLength( /*#__PURE__*/ramda.flip(ramda.lte));

/**
 * Returns `true` if the supplied list or string has a length equal to `valueLength`.
 *
 * @func lengthEq
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.8.0|v2.8.0}
 * @category List
 * @sig Number -> [*] -> Boolean
 * @param {number} valueLength The length of the list or string
 * @param {Array|string} value The list or string
 * @return {boolean}
 * @see {@link RA.lengthNotEq|lengthNotEq}, {@link RA.lengthLt|lengthLt}, {@link RA.lengthGt|lengthGt}, {@link RA.lengthLte|lengthLte}, {@link RA.lengthGte|lengthGte},, {@link http://ramdajs.com/docs/#equals|equals}, {@link http://ramdajs.com/docs/#length|length}
 * @example
 *
 * RA.lengthEq(3, [1,2,3]); //=> true
 * RA.lengthEq(3, [1,2,3,4]); //=> false
 */
var lengthEq = /*#__PURE__*/compareLength(ramda.equals);

/**
 * Returns `true` if the supplied list or string has a length not equal to `valueLength`.
 *
 * @func lengthNotEq
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.8.0|v2.8.0}
 * @category List
 * @sig Number -> [*] -> Boolean
 * @param {number} valueLength The length of the list or string
 * @param {Array|string} value The list or string
 * @return {boolean}
 * @see {@link RA.lengthEq|lengthEq}, {@link RA.lengthLt|lengthLt}, {@link RA.lengthGt|lengthGt}, {@link RA.lengthLte|lengthLte}, {@link RA.lengthGte|lengthGte}, {@link http://ramdajs.com/docs/#equals|equals}, {@link http://ramdajs.com/docs/#length|length}
 * @example
 *
 * RA.lengthNotEq(3, [1,2,3,4]); //=> true
 * RA.lengthNotEq(3, [1,2,3]); //=> false
 */
var lengthNotEq = /*#__PURE__*/compareLength( /*#__PURE__*/ramda.complement(ramda.equals));

// Original idea for this function was conceived by https://github.com/jackmellis
// in https://github.com/char0n/ramda-adjunct/pull/513.

/**
 * Returns true if all items in the list are equivalent using `R.equals` for equality comparisons.
 *
 * @func allEqual
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.9.0|v2.9.0}
 * @category List
 * @sig [a] -> Boolean
 * @param {Array} list The list of values
 * @return {boolean}
 * @see {@link https://ramdajs.com/docs/#equals|equals}
 * @example
 *
 * RA.allEqual([ 1, 2, 3, 4 ]); //=> false
 * RA.allEqual([ 1, 1, 1, 1 ]); //=> true
 * RA.allEqual([]); //=> true
 *
 */
var allEqual = /*#__PURE__*/ramda.pipe(ramda.uniq, /*#__PURE__*/lengthLte(1));

/**
 * Acts as multiple path: arrays of paths in, array of values out. Preserves order.
 *
 * @func paths
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.2.0|v1.2.0}
 * @category List
 * @sig  [[k]] -> {k: v} - [v]
 * @param {Array} ps The property paths to fetch
 * @param {Object} obj The object to query
 * @return {Array} The corresponding values or partially applied function
 * @see {@link https://github.com/ramda/ramda/wiki/Cookbook#derivative-of-rprops-for-deep-fields|Ramda Cookbook}, {@link http://ramdajs.com/docs/#props|R.props}
 * @example
 *
 * const obj = {
 *   a: { b: { c: 1 } },
 *   x: 2,
 * };
 *
 * RA.paths([['a', 'b', 'c'], ['x']], obj); //=> [1, 2]
 */
var paths = /*#__PURE__*/ramda.curry(function (ps, obj) {
  return ramda.ap([ramda.path(ramda.__, obj)], ps);
});

/**
 * Creates a new object with the own properties of the provided object, but the
 * keys renamed according to logic of renaming function.
 *
 * Keep in mind that in the case of keys conflict is behaviour undefined and
 * the result may vary between various JS engines!
 *
 * @func renameKeysWith
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.5.0|v1.5.0}
 * @category Object
 * @sig (a -> b) -> {a: *} -> {b: *}
 * @param {Function} fn Function that renames the keys
 * @param {!Object} obj Provided object
 * @return {!Object} New object with renamed keys
 * @see {@link https://github.com/ramda/ramda/wiki/Cookbook#rename-keys-of-an-object-by-a-function|Ramda Cookbook}, {@link RA.renameKeys|renameKeys}
 * @example
 *
 * RA.renameKeysWith(R.concat('a'), { A: 1, B: 2, C: 3 }) //=> { aA: 1, aB: 2, aC: 3 }
 */
var renameKeysWith = /*#__PURE__*/ramda.curry(function (fn, obj) {
  return ramda.pipe(ramda.toPairs, ramda.map(ramda.adjust(fn, 0)), ramda.fromPairs)(obj);
});

var valueOrKey = function valueOrKey(keysMap) {
  return function (key) {
    if (ramda.has(key, keysMap)) {
      return keysMap[key];
    }
    return key;
  };
};

/**
 * Creates a new object with the own properties of the provided object, but the
 * keys renamed according to the keysMap object as `{oldKey: newKey}`.
 * When some key is not found in the keysMap, then it's passed as-is.
 *
 * Keep in mind that in the case of keys conflict is behaviour undefined and
 * the result may vary between various JS engines!
 *
 * @func renameKeys
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.5.0|v1.5.0}
 * @category Object
 * @sig {a: b} -> {a: *} -> {b: *}
 * @param {!Object} keysMap
 * @param {!Object} obj
 * @return {!Object} New object with renamed keys
 * @see {@link https://github.com/ramda/ramda/wiki/Cookbook#rename-keys-of-an-object|Ramda Cookbook}, {@link RA.renameKeysWith|renameKeysWith}
 * @example
 *
 * const input = { firstName: 'Elisia', age: 22, type: 'human' };
 *
 * RA.renameKeys({ firstName: 'name', type: 'kind', foo: 'bar' })(input);
 * //=> { name: 'Elisia', age: 22, kind: 'human' }
 */
var renameKeys = /*#__PURE__*/ramda.curry(function (keysMap, obj) {
  return renameKeysWith(valueOrKey(keysMap), obj);
});

/**
 * Create a new object with the own properties of the second object merged with
 * the own properties of the first object. If a key exists in both objects,
 * the value from the first object will be used. *
 * Putting it simply: it sets properties only if they don't exist.
 *
 * @func mergeRight
 * @aliases resetToDefault
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.6.0|v1.6.0}
 * @category Object
 * @sig {k: v} -> {k: v} -> {k: v}
 * @param {Object} r Destination
 * @param {Object} l Source
 * @return {Object}
 * @see {@link http://ramdajs.com/docs/#merge|R.merge}, {@link https://github.com/ramda/ramda/wiki/Cookbook#set-properties-only-if-they-dont-exist|Ramda Cookbook}
 * @example
 *
 * RA.mergeRight({ 'age': 40 }, { 'name': 'fred', 'age': 10 });
 * //=> { 'name': 'fred', 'age': 40 }
 */
var mergeRight = /*#__PURE__*/ramda.flip(ramda.merge);

/**
 * Functional equivalent of merging object properties with object spread operator.
 *
 * @func mergeProps
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.17.0|v1.17.0}
 * @category Object
 * @sig [k] -> {k: {a}} -> {a}
 * @see {@link RA.mergePaths|mergePaths}
 * @param {!Array} ps The property names to merge
 * @param {!Object} obj The object to query
 * @return {!Object} The object composed of merged properties of obj
 * @example
 *
 * const obj = {
 *   foo: { fooInner: 1 },
 *   bar: { barInner: 2 }
 * };
 *
 * { ...obj.foo, ...obj.bar }; //=> { fooInner: 1, barInner: 2 }
 * RA.mergeProps(['foo', 'bar'], obj); //=> { fooInner: 1, barInner: 2 }
 */
var mergeProps = /*#__PURE__*/ramda.curryN(2, /*#__PURE__*/ramda.pipe(ramda.props, ramda.mergeAll));

/**
 * Merge objects under corresponding paths.
 *
 * @func mergePaths
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.17.0|v1.17.0}
 * @category Object
 * @sig [[k]] -> {k: {a}} -> {a}
 * @see {@link RA.mergeProps|mergeProps}
 * @param {!Array} paths The property paths to merge
 * @param {!Object} obj The object to query
 * @return {!Object} The object composed of merged property paths of obj
 * @example
 *
 * const obj = {
 *   foo: { fooInner: { fooInner2: 1 } },
 *   bar: { barInner: 2 }
 * };
 *
 * { ...obj.foo.fooInner, ...obj.bar }; //=>  { fooInner2: 1, barInner: 2 }
 * RA.mergePaths([['foo', 'fooInner'], ['bar']], obj); //=> { fooInner2: 1, barInner: 2 }
 */
var mergePaths = /*#__PURE__*/ramda.curryN(2, /*#__PURE__*/ramda.pipe(paths, ramda.mergeAll));

/**
 * Create a new object with the own properties of the object under the `path`
 * merged with the own properties of the provided `source`.
 * If a key exists in both objects, the value from the `source` object will be used.
 *
 * @func mergePath
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.18.0|v1.18.0}
 * @category Object
 * @sig [k] -> {a} -> {k: {a}} -> {k: {a}}
 * @see {@link RA.mergeProp|mergeProp}
 * @param {!Array} path The property path of the destination object
 * @param {!Object} source The source object
 * @param {!Object} obj The object that has destination object under corresponding property path
 * @return {!Object} The new version of object
 * @example
 *
 * RA.mergePath(
 *  ['outer', 'inner'],
 *  { foo: 3, bar: 4 },
 *  { outer: { inner: { foo: 2 } } }
 * ); //=> { outer: { inner: { foo: 3, bar: 4 } }
 */
var mergePath = /*#__PURE__*/ramda.curry(function (path$$1, source, obj) {
  return ramda.over(ramda.lensPath(path$$1), mergeRight(source), obj);
});

/**
 * Create a new object with the own properties of the object under the `p`
 * merged with the own properties of the provided `source`.
 * If a key exists in both objects, the value from the `source` object will be used.
 *
 * @func mergeProp
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.18.0|v1.18.0}
 * @category Object
 * @sig [k] -> {a} -> {k: {a}} -> {k: {a}}
 * @see {@link RA.mergePath|mergePath}
 * @param {!Array} p The property of the destination object
 * @param {!Object} source The source object
 * @param {!Object} obj The object that has destination object under corresponding property
 * @return {!Object} The new version of object
 * @example
 *
 * RA.mergeProp(
 *  'outer',
 *  { foo: 3, bar: 4 },
 *  { outer: { foo: 2 } }
 * ); //=> { outer: { foo: 3, bar: 4 } };
 */
var mergeProp = /*#__PURE__*/ramda.curry(function (p, subj, obj) {
  return mergePath(ramda.of(p), subj, obj);
});

/* eslint-disable max-len */
/**
 * Returns a partial copy of an object containing only the keys
 * that don't satisfy the supplied predicate.
 *
 * @func omitBy
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.6.0|v2.6.0}
 * @category Object
 * @sig ((v, k) -> Boolean) -> {k: v} -> {k: v}
 * @param {!Function} pred A predicate to determine whether or not a key should be included on the output object
 * @param {!Object} obj The object to copy from
 * @return {!Object} A new object only with properties that don't satisfy `pred`
 *
 * @example
 *
 * const isLowerCase = (val, key) => key.toLowerCase() === key;
 * RA.omitBy(isLowerCase, {a: 1, b: 2, A: 3, B: 4}); //=> {A: 3, B: 4}
 */
/* eslint-enable max-len */
var omitBy = /*#__PURE__*/ramda.useWith(ramda.pickBy, [ramda.complement, ramda.identity]);

/**
 * Returns a "view" of the given data structure, determined by the given lens.
 * The lens's focus determines which portion of the data structure is visible.
 * Returns the defaultValue if "view" is null, undefined or NaN; otherwise the "view" is returned.
 *
 * @func viewOr
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.13.0|1.13.0}
 * @category Object
 * @typedef Lens s b = Functor f => (b -> f b) -> s -> f s
 * @sig a -> Lens s b -> s -> b | a
 * @see {@link http://ramdajs.com/docs/#view|R.view}
 * @param {*} defaultValue The default value
 * @param {Function} lens Van Laarhoven lens
 * @param {*} data The data structure
 * @returns {*} "view" or defaultValue
 *
 * @example
 *
 * RA.viewOr('N/A', R.lensProp('x'), {}); // => 'N/A'
 * RA.viewOr('N/A', R.lensProp('x'), { x: 1 }); // => 1
 * RA.viewOr('some', R.lensProp('y'), { y: null }); // => 'some'
 * RA.viewOr('some', R.lensProp('y'), { y: false }); // => false
 */

var viewOr = /*#__PURE__*/ramda.curryN(3, function (defaultValue, lens, data) {
  return ramda.defaultTo(defaultValue, ramda.view(lens, data));
});

/**
 * Returns whether or not an object has an own property with the specified name at a given path.
 *
 * @func hasPath
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.14.0|v1.14.0}
 * @category Object
 * @typedef Idx = String | Int
 * @sig [Idx] -> {a} -> Boolean
 * @param {Array.<string|number>} path The path of the nested property
 * @param {Object} obj The object to test
 * @return {boolean}
 * @see {@link http://ramdajs.com/docs/#has|R.has}
 * @example
 *
 * RA.hasPath(['a', 'b'], { a: { b: 1 } }); //=> true
 * RA.hasPath(['a', 'b', 'c'], { a: { b: 1 } }); //=> false
 * RA.hasPath(['a', 'b'], { a: { } }); //=> false
 * RA.hasPath([0], [1, 2]); //=> true
 */
var hasPath = /*#__PURE__*/ramda.curryN(2, function (objPath, obj) {
  var prop$$1 = ramda.head(objPath);

  // termination conditions
  if (ramda.length(objPath) === 0 || !isObj(obj)) {
    return false;
  } else if (ramda.length(objPath) === 1) {
    return ramda.has(prop$$1, obj);
  }

  return hasPath(ramda.tail(objPath), ramda.path([prop$$1], obj)); // base case
});

/**
 * Spreads object under property path onto provided object.
 * It's like {@link RA.flattenPath|flattenPath}, but removes object under the property path.
 *
 * @func spreadPath
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.19.0|v1.19.0}
 * @category Object
 * @typedef Idx = String | Int
 * @sig [Idx] -> {k: v} -> {k: v}
 * @param {!Array.<string|number>} path The property path to spread
 * @param {!Object} obj The provided object
 * @return {!Object} The result of the spread
 * @see {@link RA.spreadProp|spreadProp}, {@link RA.flattenPath|flattenPath}
 * @example
 *
 * RA.spreadPath(
 *   ['b1', 'b2'],
 *   { a: 1, b1: { b2: { c: 3, d: 4 } } }
 * ); // => { a: 1, c: 3, d: 4, b1: {} };
 */
var spreadPath = /*#__PURE__*/ramda.curryN(2, /*#__PURE__*/ramda.converge(ramda.merge, [ramda.dissocPath, /*#__PURE__*/ramda.pathOr({})]));

/**
 * Spreads object under property onto provided object.
 * It's like {@link RA.flattenProp|flattenProp}, but removes object under the property.
 *
 * @func spreadProp
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.19.0|v1.19.0}
 * @category Object
 * @typedef Idx = String | Int
 * @sig Idx -> {k: v} -> {k: v}
 * @param {!string|number} prop The property to spread
 * @param {!Object} obj The provided object
 * @return {!Object} The result of the spread
 * @see {@link RA.spreadPath|spreadPath}, {@link RA.flattenProp|flattenProp}
 * @example
 *
 * RA.spreadProp('b', { a: 1, b: { c: 3, d: 4 } }); // => { a: 1, c: 3, d: 4 };
 */
var spreadProp = /*#__PURE__*/ramda.curry(function (prop$$1, obj) {
  return spreadPath(ramda.of(prop$$1), obj);
});

/**
 * Flattens a property path so that its fields are spread out into the provided object.
 * It's like {@link RA.spreadPath|spreadPath}, but preserves object under the property path.
 *
 * @func flattenPath
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.19.0|v1.19.0}
 * @category Object
 * @typedef Idx = String | Int
 * @sig [Idx] -> {k: v} -> {k: v}
 * @param {!Array.<string|number>} path The property path to flatten
 * @param {!Object} obj The provided object
 * @return {!Object} The flattened object
 * @see {@link RA.flattenProp|flattenProp}, {@link RA.spreadPath|spreadPath}
 * @example
 *
 * RA.flattenPath(
 *   ['b1', 'b2'],
 *   { a: 1, b1: { b2: { c: 3, d: 4 } } }
 * ); // => { a: 1, c: 3, d: 4, b1: { b2: { c: 3, d: 4 } } };
 */
var flattenPath = /*#__PURE__*/ramda.curry(function (path$$1, obj) {
  return ramda.merge(obj, ramda.pathOr({}, path$$1, obj));
});

/**
 * Flattens a property so that its fields are spread out into the provided object.
 * It's like {@link RA.spreadProp|spreadProp}, but preserves object under the property path.
 *
 * @func flattenProp
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.19.0|v1.19.0}
 * @category Object
 * @typedef Idx = String | Int
 * @sig [Idx] -> {k: v} -> {k: v}
 * @param {!string|number} prop The property to flatten
 * @param {!Object} obj The provided object
 * @return {!Object} The flattened object
 * @see {@link RA.flattenPath|flattenPath}, {@link RA.spreadProp|spreadProp}
 * @example
 *
 * RA.flattenProp(
 *   'b',
 *   { a: 1, b: { c: 3, d: 4 } }
 * ); // => { a: 1, c: 3, d: 4, b: { c: 3, d: 4 } };
 */
var flattenProp = /*#__PURE__*/ramda.curry(function (prop$$1, obj) {
  return flattenPath(ramda.of(prop$$1), obj);
});

/**
 * Returns `true` if data structure focused by the given lens equals provided value.
 *
 * @func lensEq
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.13.0|1.13.0}
 * @category Relation
 * @typedef Lens s a = Functor f => (a -> f a) -> s -> f s
 * @sig  Lens s a -> b -> s -> Boolean
 * @see {@link RA.lensNotEq|lensNotEq}
 * @param {function} lens Van Laarhoven lens
 * @param {*} value The value to compare the focused data structure with
 * @param {*} data The data structure
 * @return {boolean} `true` if the focused data structure equals value, `false` otherwise
 *
 * @example
 *
 * RA.lensEq(R.lensIndex(0), 1, [0, 1, 2]); // => false
 * RA.lensEq(R.lensIndex(1), 1, [0, 1, 2]); // => true
 * RA.lensEq(R.lensPath(['a', 'b']), 'foo', { a: { b: 'foo' } }) // => true
 */
var lensEq = /*#__PURE__*/ramda.curryN(3, function (lens, val, data) {
  return ramda.pipe(ramda.view(lens), ramda.equals(val))(data);
});

/**
 * Returns `true` if data structure focused by the given lens doesn't equal provided value.
 *
 * @func lensNotEq
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.13.0|1.13.0}
 * @category Relation
 * @typedef Lens s a = Functor f => (a -> f a) -> s -> f s
 * @sig Lens s a -> b -> s -> Boolean
 * @see {@link RA.lensEq|lensEq}
 * @param {function} lens Van Laarhoven lens
 * @param {*} value The value to compare the focused data structure with
 * @param {*} data The data structure
 * @return {boolean} `false` if the focused data structure equals value, `true` otherwise
 *
 * @example
 *
 * RA.lensNotEq(R.lensIndex(0), 1, [0, 1, 2]); // => true
 * RA.lensNotEq(R.lensIndex(1), 1, [0, 1, 2]); // => false
 * RA.lensNotEq(R.lensPath(['a', 'b']), 'foo', { a: { b: 'foo' } }) // => false
 */
var lensNotEq = /*#__PURE__*/ramda.complement(lensEq);

/**
 * Returns `true` if data structure focused by the given lens satisfies the predicate.
 * Note that the predicate is expected to return boolean value and will be evaluated
 * as `false` unless the predicate returns `true`.
 *
 * @func lensSatisfies
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.13.0|1.13.0}
 * @category Relation
 * @typedef Lens s a = Functor f => (a -> f a) -> s -> f s
 * @sig  Boolean b => (a -> b) -> Lens s a -> s -> b
 * @see {@link RA.lensNotSatisfy|lensNotSatisfy}
 * @param {Function} predicate The predicate function
 * @param {Function} lens Van Laarhoven lens
 * @param {*} data The data structure
 * @return {boolean} `true` if the focused data structure satisfies the predicate, `false` otherwise
 *
 * @example
 *
 * RA.lensSatisfies(RA.isTrue, R.lensIndex(0), [false, true, 1]); // => false
 * RA.lensSatisfies(RA.isTrue, R.lensIndex(1), [false, true, 1]); // => true
 * RA.lensSatisfies(RA.isTrue, R.lensIndex(2), [false, true, 1]); // => false
 * RA.lensSatisfies(R.identity, R.lensProp('x'), { x: 1 }); // => false
 */
var lensSatisfies = /*#__PURE__*/ramda.curryN(3, function (predicate, lens, data) {
  return ramda.pipe(ramda.view(lens), predicate, isTrue)(data);
});

/**
 * Returns `true` if data structure focused by the given lens doesn't satisfy the predicate.
 * Note that the predicate is expected to return boolean value.
 *
 * @func lensNotSatisfy
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.13.0|1.13.0}
 * @category Relation
 * @typedef Lens s a = Functor f => (a -> f a) -> s -> f s
 * @sig  Boolean b => (a -> b) -> Lens s a -> s -> b
 * @see {@link RA.lensSatisfies|lensSatisfies}
 * @param {Function} predicate The predicate function
 * @param {Function} lens Van Laarhoven lens
 * @param {*} data The data structure
 * @return {boolean} `false` if the focused data structure satisfies the predicate, `true` otherwise
 *
 * @example
 *
 * RA.lensNotSatisfy(RA.isTrue, R.lensIndex(0), [false, true, 1]); // => true
 * RA.lensNotSatisfy(RA.isTrue, R.lensIndex(1), [false, true, 1]); // => false
 * RA.lensNotSatisfy(RA.isTrue, R.lensIndex(2), [false, true, 1]); // => true
 * RA.lensNotSatisfy(R.identity, R.lensProp('x'), { x: 1 }); // => true
 */
var lensNotSatisfy = /*#__PURE__*/ramda.complement(lensSatisfies);

// type :: Monad a => a -> String
var type = /*#__PURE__*/ramda.either( /*#__PURE__*/ramda.path(['@@type']), /*#__PURE__*/ramda.path(['constructor', '@@type']));

// typeEquals :: Monad a => String -> a -> Boolean
var typeEquals = /*#__PURE__*/ramda.curry(function (typeIdent, monad) {
  return type(monad) === typeIdent;
});

// isSameType :: (Monad a, Monad b) => a -> b -> Boolean
var isSameType = /*#__PURE__*/ramda.curryN(2, /*#__PURE__*/ramda.useWith(ramda.equals, [type, type]));

// isNotSameType :: (Monad a, Monad b) => a -> b -> Boolean
var isNotSameType = /*#__PURE__*/ramda.complement(isSameType);

// aliases :: Prototype -> NewPrototypePairs
//     Prototype = Object
//     NewPrototypePairs = Array
var aliases = /*#__PURE__*/ramda.pipe(ramda.toPairs, /*#__PURE__*/ramda.map( /*#__PURE__*/ramda.over( /*#__PURE__*/ramda.lensIndex(0), /*#__PURE__*/ramda.replace('fantasy-land/', ''))));

var _functorTrait;
var _applyTrait;
var _setoidTrait;
var _semigroupTrait;
var _chainTrait;
var _ordTrait;

var functorTrait = (_functorTrait = {}, _functorTrait[mapping.map] = function (fn) {
  return this.constructor[mapping.of](fn(this.value));
}, _functorTrait);

var applyTrait = (_applyTrait = {}, _applyTrait[mapping.ap] = function (applyWithFn) {
  var _this = this;

  return applyWithFn.map(function (fn) {
    return fn(_this.value);
  });
}, _applyTrait);

var setoidTrait = (_setoidTrait = {}, _setoidTrait[mapping.equals] = function (setoid) {
  return isSameType(this, setoid) && ramda.equals(this.value, setoid.value);
}, _setoidTrait);

var semigroupTrait = (_semigroupTrait = {}, _semigroupTrait[mapping.concat] = function (semigroup) {
  var concatenatedValue = this.value;

  if (isString(this.value) || isNumber(this.value)) {
    concatenatedValue = this.value + semigroup.value;
  } else if (ramda.pathSatisfies(isFunction, ['value', mapping.concat], this)) {
    concatenatedValue = this.value[mapping.concat](semigroup.value);
  } else if (ramda.pathSatisfies(isFunction, ['value', 'concat'], this)) {
    concatenatedValue = this.value.concat(semigroup.value);
  }

  return this.constructor[mapping.of](concatenatedValue);
}, _semigroupTrait);

var chainTrait = (_chainTrait = {}, _chainTrait[mapping.chain] = function (fn) {
  var newChain = fn(this.value);

  return isSameType(this, newChain) ? newChain : this;
}, _chainTrait);

var ordTrait = (_ordTrait = {}, _ordTrait[mapping.lte] = function (ord) {
  return isSameType(this, ord) && (this.value < ord.value || this[mapping.equals](ord));
}, _ordTrait);

var _createClass = /*#__PURE__*/ function () { function defineProperties(target, props$$1) { for (var i = 0; i < props$$1.length; i++) { var descriptor = props$$1[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

// we do this here for jsdocs generate properly
var of$1 = mapping.of;
var ap$3 = mapping.ap;
var map$1 = mapping.map;
var equals$1 = mapping.equals;
var concat$1 = mapping.concat;
var chain = mapping.chain;
var lte$1 = mapping.lte;
var empty$1 = mapping.empty;
var contramap = mapping.contramap;

/**
 * The simplest {@link https://github.com/fantasyland/fantasy-land|fantasy-land}
 * compatible monad which attaches no information to values.
 *
 * The Identity type is a very simple type that has no interesting side effects and
 * is effectively just a container of some value. So why does it exist ?
 * The Identity type is often used as the base monad of a monad
 * transformer when no other behaviour is required.
 *
 * @memberOf RA
 * @implements
 * {@link https://github.com/fantasyland/fantasy-land#apply|Apply},
 * {@link https://github.com/fantasyland/fantasy-land#applicative|Applicative},
 * {@link https://github.com/fantasyland/fantasy-land#functor|Functor},
 * {@link https://github.com/fantasyland/fantasy-land#setoid|Setoid},
 * {@link https://github.com/fantasyland/fantasy-land#semigroup|Semigroup},
 * {@link https://github.com/fantasyland/fantasy-land#chain|Chain},
 * {@link https://github.com/fantasyland/fantasy-land#monad|Monad},
 * {@link https://github.com/fantasyland/fantasy-land#ord|Ord},
 * {@link https://github.com/fantasyland/fantasy-land#monoid|Monoid*},
 * {@link https://github.com/fantasyland/fantasy-land#contravariant|Contravariant}
 * @since {@link https://char0n.github.io/ramda-adjunct/1.8.0|v1.8.0}
 */

var Identity = /*#__PURE__*/function () {
  /**
   * Fantasy land {@link https://github.com/fantasyland/fantasy-land#applicative|Applicative} specification.
   *
   * @static
   * @sig of :: Applicative f => a -> f a
   * @param {*} value
   * @returns {RA.Identity}
   * @example
   *
   * const a = Identity.of(1); //=> Identity(1)
   */
  Identity[of$1] = function (value) {
    return new Identity(value);
  };

  /**
   * @static
   */


  _createClass(Identity, null, [{
    key: '@@type',
    get: function get() {
      return 'RA/Identity';
    }

    /**
     * Private constructor. Use {@link RA.Identity.of|Identity.of} instead.
     *
     * @private
     * @param {*} value
     * @return {RA.Identity}
     */

  }]);

  function Identity(value) {
    _classCallCheck(this, Identity);

    this.value = value;
  }

  /**
   * Catamorphism for a value.
   * @returns {*}
   * @example
   *
   * const a = Identity.of(1);
   * a.get(); //=> 1
   */


  Identity.prototype.get = function get() {
    return this.value;
  };

  /**
   * Fantasy land {@link https://github.com/fantasyland/fantasy-land#apply|Apply} specification.
   *
   * @sig ap :: Apply f => f a ~> f (a -> b) -> f b
   * @param {RA.Identity} applyWithFn
   * @return {RA.Identity}
   * @example
   *
   * const a = Identity.of(1);
   * const b = Identity.of(1).map(a => b => a + b);
   *
   * a.ap(b); //=> Identity(2)
   */


  Identity.prototype[ap$3] = function (applyWithFn) {
    return applyTrait[ap$3].call(this, applyWithFn);
  };

  /**
   * Fantasy land {@link https://github.com/fantasyland/fantasy-land#functor|Functor} specification.
   *
   * @sig map :: Functor f => f a ~> (a -> b) -> f b
   * @param {Function} fn
   * @return {RA.Identity}
   * @example
   *
   * const a = Identity.of(1);
   * a.map(a => a + 1); //=> Identity(2)
   */


  Identity.prototype[map$1] = function (fn) {
    return functorTrait[map$1].call(this, fn);
  };

  /**
   * Fantasy land {@link https://github.com/fantasyland/fantasy-land#setoid|Setoid} specification.
   *
   * @sig equals :: Setoid a => a ~> a -> Boolean
   * @param {RA.Identity} setoid
   * @return {boolean}
   * @example
   *
   * const a = Identity.of(1);
   * const b = Identity.of(1);
   * const c = Identity.of(2);
   *
   * a.equals(b); //=> true
   * a.equals(c); //=> false
   */


  Identity.prototype[equals$1] = function (setoid) {
    return setoidTrait[equals$1].call(this, setoid);
  };

  /**
   * Fantasy land {@link https://github.com/fantasyland/fantasy-land#semigroup|Semigroup} specification.
   *
   * @sig concat :: Semigroup a => a ~> a -> a
   * @param {RA.Identity} semigroup
   * @return {RA.Identity}
   * @example
   *
   * const a = Identity.of(1);
   * const b = Identity.of(1);
   * a.concat(b); //=> 2
   *
   * const c = Identity.of('c');
   * const d = Identity.of('d');
   * c.concat(d); //=> 'cd'
   *
   * const e = Identity.of(['e']);
   * const f = Identity.of(['f']);
   * e.concat(f); //=> ['e', 'f']
   */


  Identity.prototype[concat$1] = function (semigroup) {
    return semigroupTrait[concat$1].call(this, semigroup);
  };

  /**
   * Fantasy land {@link https://github.com/fantasyland/fantasy-land#chain|Chain} specification.
   *
   * @sig chain :: Chain m => m a ~> (a -> m b) -> m b
   * @param {Function} fn Function returning the value of the same {@link https://github.com/fantasyland/fantasy-land#semigroup|Chain}
   * @return {RA.Identity}
   * @example
   *
   * const a = Identity.of(1);
   * const fn = val => Identity.of(val + 1);
   *
   * a.chain(fn).chain(fn); //=> Identity(3)
   */


  Identity.prototype[chain] = function (fn) {
    return chainTrait[chain].call(this, fn);
  };

  /**
   * Fantasy land {@link https://github.com/fantasyland/fantasy-land#ord|Ord} specification.
   *
   * @sig lte :: Ord a => a ~> a -> Boolean
   * @param {RA.Identity} ord
   * @return {boolean}
   * @example
   *
   * const a = Identity.of(1);
   * const b = Identity.of(1);
   * const c = Identity.of(2);
   *
   * a.lte(b); //=> true
   * a.lte(c); //=> true
   * c.lte(a); //=> false
   */


  Identity.prototype[lte$1] = function (ord) {
    return ordTrait[lte$1].call(this, ord);
  };

  /**
   * Fantasy land {@link https://github.com/fantasyland/fantasy-land#monoid|Monoid*} specification.
   * Partial implementation of Monoid specification. `empty` method on instance only, returning
   * identity value of the wrapped type. Using `R.empty` under the hood.
   *
   *
   * @sig empty :: Monoid m => () -> m
   * @return {RA.Identity}
   * @example
   *
   * const a = Identity.of('test');
   * const i = a.empty();
   *
   * a.concat(i); //=> Identity('string');
   * i.concat(a); //=> Identity('string');
   */


  Identity.prototype[empty$1] = function () {
    return this.constructor.of(ramda.empty(this.value));
  };

  /**
   * Fantasy land {@link https://github.com/fantasyland/fantasy-land#contravariant|Contravariant} specification.
   *
   * @sig contramap :: Contravariant f => f a ~> (b -> a) -> f b
   * @param {Function} fn
   * @return {RA.Identity}
   * @example
   *
   * const identity = a => a;
   * const add1 = a => a + 1;
   * const divide2 = a => a / 2;
   *
   * Identity.of(divide2).contramap(add1).get()(3); //=> 2
   * Identity.of(identity).contramap(divide2).contramap(add1).get()(3); //=> 2
   * Identity.of(identity).contramap(a => divide2(add1(a))).get()(3); //=> 2
   */


  Identity.prototype[contramap] = function (fn) {
    var _this = this;

    return this.constructor.of(function (value) {
      return _this.value(fn(value));
    });
  };

  return Identity;
}();

aliases(Identity).forEach(function (_ref) {
  var alias = _ref[0],
      fn = _ref[1];

  Identity[alias] = fn;
});
aliases(Identity.prototype).forEach(function (_ref2) {
  var alias = _ref2[0],
      fn = _ref2[1];

  Identity.prototype[alias] = fn;
});

// This implementation was highly inspired by the implementations
// in ramda-lens library.
//
// https://github.com/ramda/ramda-lens

// isomorphic :: ((a -> b), (b -> a)) -> Isomorphism
//     Isomorphism = x -> y
var isomorphic = function isomorphic(to, from) {
  var isomorphism = function isomorphism(x) {
    return to(x);
  };
  isomorphism.from = from;
  return isomorphism;
};

// isomorphisms :: ((a -> b), (b -> a)) -> (a -> b)
var isomorphisms = function isomorphisms(to, from) {
  return isomorphic(ramda.curry(function (toFunctorFn, target) {
    return ramda.map(from, toFunctorFn(to(target)));
  }), ramda.curry(function (toFunctorFn, target) {
    return ramda.map(to, toFunctorFn(from(target)));
  }));
};

// from :: Isomorphism -> a -> b
var from = /*#__PURE__*/ramda.curry(function (isomorphism, x) {
  return isomorphism.from(x);
});

/**
 * Defines an isomorphism that will work like a lens. It takes two functions.
 * The function that converts and the function that recovers.
 *
 * @func lensIso
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/1.19.0|1.19.0}
 * @category Relation
 * @typedef Lens s a = Functor f => (a -> f a) -> s -> f s
 * @sig (s -> a) -> (a -> s) -> Lens s a
 * @param {!function} to The function that converts
 * @param {!function} from The function that recovers
 * @return {!function} The isomorphic lens
 * @see {@link http://ramdajs.com/docs/#lens|R.lens}
 *
 * @example
 *
 * const lensJSON = RA.lensIso(JSON.parse, JSON.stringify);
 *
 * R.over(lensJSON, assoc('b', 2), '{"a":1}'); //=> '{"a":1,"b":2}'
 * R.over(RA.lensIso.from(lensJSON), R.replace('}', ',"b":2}'), { a: 1 }); // => { a: 1, b: 2 }
 */
var lensIso = /*#__PURE__*/ramda.curry(isomorphisms);
lensIso.from = from;

/**
 * Returns true if the specified object property is not equal,
 * in R.equals terms, to the given value; false otherwise.
 *
 * @func propNotEq
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.3.0|v2.3.0}
 * @category Relation
 * @sig  String -> a -> Object -> Boolean
 * @param {String} name The property to pick
 * @param {a} val The value to compare to
 * @param {Object} object The object, that presumably contains value under the property
 * @return {boolean} Comparison result
 * @see {@link http://ramdajs.com/docs/#propEq|R.propEq}
 * @example
 *
 * const abby = { name: 'Abby', age: 7, hair: 'blond' };
 * const fred = { name: 'Fred', age: 12, hair: 'brown' };
 * const rusty = { name: 'Rusty', age: 10, hair: 'brown' };
 * const alois = { name: 'Alois', age: 15, disposition: 'surly' };
 * const kids = [abby, fred, rusty, alois];
 * const hasNotBrownHair = RA.propNotEq('hair', 'brown');
 *
 * R.filter(hasNotBrownHair, kids); //=> [abby, alois]
 */
var propNotEq = /*#__PURE__*/ramda.complement(ramda.propEq);

/* eslint-disable max-len */
/**
 * Determines whether a nested path on an object doesn't have a specific value,
 * in R.equals terms. Most likely used to filter a list.
 *
 * @func pathNotEq
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.4.0|v2.4.0}
 * @category Relation
 * @sig [Idx] => a => {a} => Boolean
 * @sig Idx = String | Int
 * @param {Array} path The path of the nested property to use
 * @param {a} val The value to compare the nested property with
 * @param {Object} object The object to check the nested property in
 * @return {boolean} Returns Boolean `false` if the value equals the nested object property, `true` otherwise
 * @see {@link http://ramdajs.com/docs/#pathEq|R.pathEq}
 * @example
 *
 * const user1 = { address: { zipCode: 90210 } };
 * const user2 = { address: { zipCode: 55555 } };
 * const user3 = { name: 'Bob' };
 * const users = [ user1, user2, user3 ];
 * const isFamous = R.pathNotEq(['address', 'zipCode'], 90210);
 * R.filter(isFamous, users); //=> [ user2, user3 ]
 */
/* eslint-enable max-len */
var pathNotEq = /*#__PURE__*/ramda.complement(ramda.pathEq);

var inRangeImp = /*#__PURE__*/ramda.ifElse(ramda.gte, function () {
  throw new Error('low must not be greater than high in inRange(low, high, value)');
}, /*#__PURE__*/ramda.useWith(ramda.both, [ramda.lte, ramda.gt]));

/**
 * Checks if `value` is between `low` and up to but not including `high`.
 *
 * @func inRange
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.7.0|v2.7.0}
 * @category Relation
 * @sig Number -> Number -> Number -> Boolean
 * @param {Number} low Start of the range
 * @param {Number} The end of the range
 * @param {Number} value The value to test
 * @return {boolean}
 * @throws {Error} When `low` is greater than or equal to `high`
 * @example
 *
 * RA.inRange(0, 5, 3); //=> true
 * RA.inRange(0, 5, 0); //=> true
 * RA.inRange(0, 5, 4); //=> true
 * RA.inRange(0, 5, 5); //=> false
 * RA.inRange(0, 5, -1); //=> false
 */
/*#__PURE__*/ramda.curry(function (low, high, value) {
  return inRangeImp(low, high)(value);
});

/**
 * Returns the second argument if predicate function returns `true`,
 * otherwise the third argument is returned.
 *
 * @func defaultWhen
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.2.0|v2.2.0}
 * @category Logic
 * @sig  (a -> Boolean) -> b -> a -> a | b
 * @param {!function} predicate The predicate function
 * @param {*} defaultVal The default value
 * @param {*} val `val` will be returned instead of `defaultVal` if predicate returns false
 * @return {*} The `val` if predicate returns `false`, otherwise the default value
 * @see {@link http://ramdajs.com/docs/#defaultTo|R.defaultTo}
 * @example
 *
 * RA.defaultWhen(RA.isNull, 1, null); // => 1
 * RA.defaultWhen(RA.isNull, 1, 2); // => 2
 */
var defaultWhen = /*#__PURE__*/ramda.curry(function (predicate, defaultVal, val) {
  return predicate(val) ? defaultVal : val;
});

/* eslint-disable max-len */
/**
 * A function which calls the two provided functions and returns the complement of `&&`ing the
 * results.
 * It returns true if the first function is false-y and the complement of the second function
 * otherwise. Note that this is short-circuited, meaning that the second function will not be
 * invoked if the first returns a false-y value. In short it will return true unless both predicates
 * return true.
 *
 * In addition to functions, `RA.notBoth` also accepts any fantasy-land compatible
 * applicative functor.
 *
 * @func notBoth
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.3.0|v2.3.0}
 * @category Logic
 * @sig (*... -> Boolean) -> (*... -> Boolean) -> (*... -> Boolean)
 * @param {Function} f A predicate
 * @param {Function} g Another predicate
 * @return {Function} Returns a function that applies its arguments to `f` and `g` and returns the complement of `&&`ing their outputs together.
 * @see {@link http://ramdajs.com/docs/#both|R.both}
 * @example
 *
 * const gt10 = R.gt(R.__, 10)
 * const even = (x) => x % 2 === 0;
 * const f = RA.notBoth(gt10, even);
 *
 * f(12); //=> false
 * f(8); //=> true
 * f(11); //=> true
 * f(9); //=> true
 */
/* eslint-enable max-len */
var notBoth = /*#__PURE__*/ramda.compose(ramda.complement, ramda.both);

/* eslint-disable max-len */
/**
 * A function which calls the two provided functions and returns the complement of `||`ing the
 * results.
 * It returns false if the first function is truth-y and the complement of the second function
 * otherwise. Note that this is short-circuited, meaning that the second function will not be
 * invoked if the first returns a truth-y value. In short it will return true if neither predicate
 * returns true.
 *
 * In addition to functions, `RA.neither` also accepts any fantasy-land compatible
 * applicative functor.
 *
 * @func neither
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.3.0|v2.3.0}
 * @category Logic
 * @sig (*... -> Boolean) -> (*... -> Boolean) -> (*... -> Boolean)
 * @param {Function} f A predicate
 * @param {Function} g Another predicate
 * @return {Function} Returns a function that applies its arguments to `f` and `g` and returns the complement of `||`ing their outputs together.
 * @see {@link http://ramdajs.com/docs/#either|R.either}, {@link http://ramdajs.com/docs/#or|R.or}
 * @example
 *
 * const gt10 = R.gt(R.__, 10)
 * const even = (x) => x % 2 === 0;
 * const f = RA.neither(gt10, even);
 *
 * f(12); //=> false
 * f(8); //=> false
 * f(11); //=> false
 * f(9); //=> true
 */
/* eslint-enable max-len */
var neither = /*#__PURE__*/ramda.compose(ramda.complement, ramda.either);

/**
 * Takes a list of predicates and returns a predicate that returns true for a given list of
 * arguments if one or more of the provided predicates is not satisfied by those arguments. It is
 * the complement of Ramda's allPass.
 *
 * The function returned is a curried function whose arity matches that of the
 * highest-arity predicate.
 *
 * @func notAllPass
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.4.0|v2.4.0}
 * @category Logic
 * @sig [(*... -> Boolean)] -> (*... -> Boolean)
 * @param {Array} predicates An array of predicates to check
 * @return {Function} The combined predicate
 * @see {@link http://ramdajs.com/docs/#allPass|R.allPass}
 * @example
 *
 * const gt10 = R.gt(R.__, 10)
 * const even = (x) => x % 2 === 0;
 * const f = RA.notAllPass([gt10, even]);
 *
 * f(12); //=> false
 * f(8); //=> true
 * f(11); //=> true
 * f(9); //=> true
 */
var notAllPass = /*#__PURE__*/ramda.compose(ramda.complement, ramda.allPass);

/**
 * Takes a list of predicates and returns a predicate that returns true for a given list of
 * arguments if none of the provided predicates are satisfied by those arguments. It is the
 * complement of Ramda's anyPass.
 *
 * The function returned is a curried function whose arity matches that of the
 * highest-arity predicate.
 *
 * @func nonePass
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.5.0|v2.5.0}
 * @category Logic
 * @sig [(*... -> Boolean)] -> (*... -> Boolean)
 * @param {Array} predicates An array of predicates to check
 * @return {Function} The combined predicate
 * @see {@link http://ramdajs.com/docs/#anyPass|R.anyPass}
 * @example
 *
 * const gt10 = R.gt(R.__, 10)
 * const even = (x) => x % 2 === 0;
 * const f = RA.nonePass([gt10, even]);
 *
 * f(12); //=> false
 * f(8); //=> false
 * f(11); //=> false
 * f(9); //=> true
 */
var nonePass = /*#__PURE__*/ramda.compose(ramda.complement, ramda.anyPass);

/**
 * Takes a combining predicate and a list of functions and returns a function which will map the
 * arguments it receives to the list of functions and returns the result of passing the values
 * returned from each function to the combining predicate. A combining predicate is a function that
 * combines a list of Boolean values into a single Boolean value, such as `R.any` or `R.all`. It
 * will test each value using `RA.isTruthy`, meaning the functions don't necessarily have to be
 * predicates.
 *
 * The function returned is curried to the number of functions supplied, and if called with more
 * arguments than functions, any remaining arguments are passed in to the combining predicate
 * untouched.
 *
 * @func argsPass
 * @memberOf RA
 * @since {@link https://char0n.github.io/ramda-adjunct/2.7.0|v2.7.0}
 * @category Logic
 * @sig ((* -> Boolean) -> [*] -> Boolean) -> [(* -> Boolean), ...] -> (*...) -> Boolean
 * @param {Function} combiningPredicate The predicate used to combine the values returned from the
 * list of functions
 * @param {Array} functions List of functions
 * @return {boolean} Returns the combined result of mapping arguments to functions
 * @example
 *
 * RA.argsPass(R.all, [RA.isArray, RA.isBoolean, RA.isString])([], false, 'abc') //=> true
 * RA.argsPass(R.all, [RA.isArray, RA.isBoolean, RA.isString])([], false, 1) //=> false
 * RA.argsPass(R.any, [RA.isArray, RA.isBoolean, RA.isString])({}, 1, 'abc') //=> true
 * RA.argsPass(R.any, [RA.isArray, RA.isBoolean, RA.isString])({}, 1, false) //=> false
 * RA.argsPass(R.none, [RA.isArray, RA.isBoolean, RA.isString])({}, 1, false) //=> true
 * RA.argsPass(R.none, [RA.isArray, RA.isBoolean, RA.isString])({}, 1, 'abc') //=> false
 */
/*#__PURE__*/ramda.curry(function (combiningPredicate, predicates) {
  return ramda.useWith(ramda.compose(combiningPredicate(isTruthy), list), predicates);
});

/**
 * @namespace RA
 */

// Type

function unwrapExports (x) {
	return x && x.__esModule && Object.prototype.hasOwnProperty.call(x, 'default') ? x['default'] : x;
}

function createCommonjsModule(fn, module) {
	return module = { exports: {} }, fn(module, module.exports), module.exports;
}

var breakpointMapping = createCommonjsModule(function (module, exports) {
Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.createBreakpointMapping = exports.assocQuery = exports.assocValue = exports.lQuery = exports.lName = exports.propQuery = exports.propValue = exports.propName = undefined;



var propName = exports.propName = (0, ramda__default.prop)('name');
var propValue = exports.propValue = (0, ramda__default.prop)('value');
var propQuery = exports.propQuery = (0, ramda__default.prop)('query');
var lName = exports.lName = (0, ramda__default.lensProp)('name');
var lQuery = exports.lQuery = (0, ramda__default.lensProp)('query');
var assocValue = exports.assocValue = (0, ramda__default.assoc)('value');
var assocQuery = exports.assocQuery = (0, ramda__default.assoc)('query');

var createBreakpointMapping = exports.createBreakpointMapping = (0, ramda__default.curry)(function (name, value, query) {
  return {
    name: name,
    query: query,
    value: value
  };
});
});

unwrapExports(breakpointMapping);
var breakpointMapping_1 = breakpointMapping.createBreakpointMapping;
var breakpointMapping_2 = breakpointMapping.assocQuery;
var breakpointMapping_3 = breakpointMapping.assocValue;
var breakpointMapping_4 = breakpointMapping.lQuery;
var breakpointMapping_5 = breakpointMapping.lName;
var breakpointMapping_6 = breakpointMapping.propQuery;
var breakpointMapping_7 = breakpointMapping.propValue;
var breakpointMapping_8 = breakpointMapping.propName;

var RE_RHYTHM_UNITS = /^-?\d*\.?\d*ru$/;

var RE_START_OF_LINE = /^/gm;

var RE_WHITESPACE = /\s+/;

var RE_PERCENT_NUMBER = /^\d+%$/;

var RE_CAPITAL_LETTERS = /([A-Z])/g;

var RE_MEDIA_QUERY_STRING = /^@media /;

var RE_RADIAL_GRADIENT = /^radial-gradient\(.+\)$/;

var RE_LINEAR_GRADIENT = /^linear-gradient\(.+\)$/;

var RE_URL = /^url\(.+\)$/;

var RE_UNNESTED_COMMA = /,(?![^()]*(?:\([^()]*\))?\))/g;

var RE_ARGUMENTS_OF_CSS_FUNCTION = /^(?:[a-zA-Z\-_\d]*)\((.*)\)$/g;

var RE_CSS_FUNCTION_NAME = /^([a-zA-Z-]*)\(.*\)$/;

var RE_UNNESTED_WHITESPACE = /\s(?![^()]*\))/g;



// Note we exclude values starting `data:` as they might be images
var RE_TOKEN = /^(?!(?:data|http|https):)([a-zA-Z\-_\d]*):(.*)$/;

var RE_CSS_FUNCTION = /^[a-zA-Z-]*\(.*\)$/;

var RE_TRANSFORM_TRANSLATE_FUNCTION = /^translate(?:x|y|z|3d)?\(.*\)$/;

var RE_CALC_FUNCTION = /^calc\(.*\)$/;

var RE_CALC_VALUES = /(-*[0-9a-z.%]+)/g;

var RE_MODIFIED_MQ = /^(?![<@][^<\s]*<)(?!^<default)(?:[@<>]?[a-zA-Z\d]+(?:(?:[+-])\d*\.?\d*(?:em)?)?)(?:<[a-zA-Z\d]+(?:(?:\+|-)\d*\.?\d*(?:em)?)?)?$/;

var isArray$2 = Array.isArray || function (obj) {
  return Object.prototype.toString.call(obj) === '[object Array]';
};

var isDate$2 = function (obj) {
  return Object.prototype.toString.call(obj) === '[object Date]';
};

var isRegex = function (obj) {
  return Object.prototype.toString.call(obj) === '[object RegExp]';
};

var has$1 = Object.prototype.hasOwnProperty;
var objectKeys = Object.keys || function (obj) {
  var keys$$1 = [];
  for (var key in obj) {
    if (has$1.call(obj, key)) {
      keys$$1.push(key);
    }
  }
  return keys$$1;
};

function dashCase(str) {
  return str.replace(/[A-Z](?:(?=[^A-Z])|[A-Z]*(?=[A-Z][^A-Z]|$))/g, function (s, i) {
    return (i > 0 ? '-' : '') + s.toLowerCase();
  });
}

function map$2(xs, f) {
  if (xs.map) {
    return xs.map(f);
  }
  var res = [];
  for (var i = 0; i < xs.length; i++) {
    res.push(f(xs[i], i));
  }
  return res;
}

function reduce$1(xs, f, acc) {
  if (xs.reduce) {
    return xs.reduce(f, acc);
  }
  for (var i = 0; i < xs.length; i++) {
    acc = f(acc, xs[i], i);
  }
  return acc;
}

function walk(obj) {
  if (!obj || typeof obj !== 'object') {
    return obj;
  }
  if (isDate$2(obj) || isRegex(obj)) {
    return obj;
  }
  if (isArray$2(obj)) {
    return map$2(obj, walk);
  }
  return reduce$1(objectKeys(obj), function (acc, key) {
    var camel = dashCase(key);
    acc[camel] = walk(obj[key]);
    return acc;
  }, {});
}

var dasherize = function (obj) {
  if (typeof obj === 'string') {
    return dashCase(obj);
  }
  return walk(obj);
};

// eslint-disable-next-line import/prefer-default-export
var condDefault = ramda.curry(function (list, value) {
  var listWithDefault = ramda.append([ramda.T, ramda.identity], list);
  return ramda.cond(listWithDefault)(value);
});

var defaultToObj = ramda.defaultTo({});

var defaultToArray = ramda.defaultTo([]);

var slicedToArray = function () {
  function sliceIterator(arr, i) {
    var _arr = [];
    var _n = true;
    var _d = false;
    var _e = undefined;

    try {
      for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) {
        _arr.push(_s.value);

        if (i && _arr.length === i) break;
      }
    } catch (err) {
      _d = true;
      _e = err;
    } finally {
      try {
        if (!_n && _i["return"]) _i["return"]();
      } finally {
        if (_d) throw _e;
      }
    }

    return _arr;
  }

  return function (arr, i) {
    if (Array.isArray(arr)) {
      return arr;
    } else if (Symbol.iterator in Object(arr)) {
      return sliceIterator(arr, i);
    } else {
      throw new TypeError("Invalid attempt to destructure non-iterable instance");
    }
  };
}();













var toConsumableArray = function (arr) {
  if (Array.isArray(arr)) {
    for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) arr2[i] = arr[i];

    return arr2;
  } else {
    return Array.from(arr);
  }
};

var reduceObjIndexed = ramda.curry(function (f, acc, v) {
  return ramda.compose(ramda.reduce(f, acc), ramda.toPairs)(v);
});

var propFlipped = ramda.flip(ramda.prop);

var filterKeys = ramda.curry(function (predicate, obj) {
  return reduceObjIndexed(function (keys$$1, _ref) {
    var _ref2 = slicedToArray(_ref, 2),
        key = _ref2[0],
        value = _ref2[1];

    return ramda.when(function () {
      return predicate(value);
    }, ramda.append(key))(keys$$1);
  }, [], obj);
});

// eslint-disable-next-line import/prefer-default-export
var DEFAULT_BREAKPOINT_NAME = "default";

var LT_MODIFIER = "<";
var GT_MODIFIER = ">";
var AT_MODIFIER = "@";
var MODIFIERS = [LT_MODIFIER, GT_MODIFIER, AT_MODIFIER];
var POSITIVE_OFFSET = "+";
var NEGATIVE_OFFSET = "-";

// eslint-disable-next-line import/prefer-default-export
var CONFIG_FIELD_NAMES = Object.freeze({
  DATA: "data",
  SCOPES: "scopes",
  RESOLVE: "resolve",
  BREAKPOINTS: "breakpoints",
  API: "api",
  TRANSFORMERS: "transformers",
  ALIASES: "aliases"
});

var LENGTH_UNITS = Object.freeze({
  REM: "rem",
  PX: "px",
  EM: "em"
});

var ANGLE_UNITS = Object.freeze({
  DEG: "deg",
  RAD: "rad",
  GRAD: "grad",
  TURN: "turn"
});

var PERCENT_UNIT = "%";

var DIRECTIONS = Object.freeze({
  TOP: "top",
  RIGHT: "right",
  BOTTOM: "bottom",
  LEFT: "left"
});

var SCOPES = CONFIG_FIELD_NAMES.SCOPES;

// -----------------------------------------------------------------------------
// List
// -----------------------------------------------------------------------------

var isContained = ramda.flip(ramda.contains);

var isLengthGt = ramda.curry(function (l, v) {
  return ramda.compose(ramda.gt(ramda.__, l), ramda.length)(v);
});

var isLengthEqualTo = ramda.curry(function (l, v) {
  return ramda.compose(ramda.equals(l), ramda.length)(v);
});

// -----------------------------------------------------------------------------
// Misc
// -----------------------------------------------------------------------------

var isSingletonArray = ramda.both(isArray, isLengthEqualTo(1));

var isDefaultBreakpoint = ramda.equals(DEFAULT_BREAKPOINT_NAME);

var isValidMqValue = ramda.anyPass([isSingletonArray, isString, isValidNumber]);

// -----------------------------------------------------------------------------
// Numeric
// -----------------------------------------------------------------------------

var isNotZero = ramda.complement(ramda.equals(0));

// -----------------------------------------------------------------------------
// Types
// -----------------------------------------------------------------------------

var isNumberWithUnit = ramda.curry(function (units, value) {
  var possibleUnits = ramda.compose(joinWithPipe, ensureArray)(units);
  var regExp = new RegExp('^-?\\d+(?:.\\d+)?(?:' + possibleUnits + ')$');
  return ramda.test(regExp, value);
});

var isEmString = isNumberWithUnit([LENGTH_UNITS.EM]);

/* eslint-disable-next-line no-restricted-globals */
var isNumberString = ramda.both(isString, ramda.complement(isNaN));

var isNotZeroString = ramda.complement(ramda.equals)('0');

var isPercentString = ramda.pipe(ramda.unless(isString, String), ramda.test(RE_PERCENT_NUMBER));

var isValidNonZeroNumber = ramda.pipe(ramda.unless(isString, String), ramda.both(isNumberString, isNotZeroString));

var isValidPositiveNumber = ramda.both(isValidNumber, isPositive);

var isLength = isNumberWithUnit(ramda.values(LENGTH_UNITS));

var isRhythmUnit = ramda.test(RE_RHYTHM_UNITS);

var isNotStringOrArray = ramda.both(isNotString, isNotArray);

var isUrl = ramda.test(RE_URL);

var isLinearGradient = ramda.test(RE_LINEAR_GRADIENT);

var isRadialGradient = ramda.test(RE_RADIAL_GRADIENT);

var isGradient = ramda.either(isLinearGradient, isRadialGradient);

var isToken = ramda.test(RE_TOKEN);

var isCSSFunction = ramda.test(RE_CSS_FUNCTION);
var isTransformTranslateFunction = ramda.test(RE_TRANSFORM_TRANSLATE_FUNCTION);

var isCalcFunction = ramda.test(RE_CALC_FUNCTION);

var isTokenWithName = function isTokenWithName(names) {
  return function (value) {
    var possibleNames = ramda.compose(joinWithPipe, ensureArray)(names);
    var regExp = new RegExp('^(' + possibleNames + '):(.*)$');
    return ramda.test(regExp, value);
  };
};

var isBreakpointProvider = ramda.both(ramda.has('byName'), ramda.has('byIndex'));

var isMediaQueryString = ramda.test(RE_MEDIA_QUERY_STRING);

var isGroups = ramda.test(RE_UNNESTED_COMMA);

var isUnitRemOrEm = isContained([LENGTH_UNITS.EM, LENGTH_UNITS.REM]);

var hasScopes = ramda.has(SCOPES);

var hasNoScopes = ramda.complement(hasScopes);

var hasUnnestedWhitespace = ramda.test(RE_UNNESTED_WHITESPACE);

var isValidModifiedMq = ramda.test(RE_MODIFIED_MQ);

var modifierIsLtModifier = ramda.pipe(ramda.prop('modifier'), ramda.equals(LT_MODIFIER));

var modifierIsGtModifier = ramda.pipe(ramda.prop('modifier'), ramda.equals(GT_MODIFIER));

var modifierIsAtModifier = ramda.pipe(ramda.prop('modifier'), ramda.equals(AT_MODIFIER));

var hasPositiveOffset = ramda.contains(POSITIVE_OFFSET);

var hasNegativeOffset = ramda.contains(NEGATIVE_OFFSET);

// -----------------------------------------------------------------------------
// Chars
// -----------------------------------------------------------------------------

var NEWLINE = '\n';
var DOUBLE_NEWLINE = '' + NEWLINE + NEWLINE;
var SPACE = ' ';
var COMMA = ',';
var COLON = ':';
var HYPHEN = '-';
var FULL_STOP = '.';
var SINGLE_QUOTE = '\'';
var PIPE = '|';
var COMMA_SPACE = '' + COMMA + SPACE;

// -----------------------------------------------------------------------------
// Create
// -----------------------------------------------------------------------------

var toToken = function toToken(v) {
  return new RegExp('#{(' + v + ')}', 'g');
};

// -----------------------------------------------------------------------------
// Print
// -----------------------------------------------------------------------------

var printObj = JSON.stringify;

// -----------------------------------------------------------------------------
// Indent
// -----------------------------------------------------------------------------

var indentLines = ramda.replace(RE_START_OF_LINE, '  ');

// -----------------------------------------------------------------------------
// Join
// -----------------------------------------------------------------------------

var joinWithSpace = ramda.join(SPACE);

var joinWithComma = ramda.join(COMMA);

var joinWithCommaSpace = ramda.join(COMMA_SPACE);

var joinWithHypen = ramda.join(HYPHEN);

var joinWithNoSpace = ramda.join('');

var joinWithDot = ramda.join(FULL_STOP);

var joinWithNewline = ramda.join(NEWLINE);

var joinWithPipe = ramda.join(PIPE);

var joinWithDoubleNewlines = ramda.join(DOUBLE_NEWLINE);

// -----------------------------------------------------------------------------
// Case
// -----------------------------------------------------------------------------

var firstToUpper = ramda.compose(joinWithNoSpace, ramda.over(ramda.lensIndex(0), ramda.toUpper));

var toKebabCase = dasherize;

// -----------------------------------------------------------------------------
// Split
// -----------------------------------------------------------------------------

var splitOnWhitespace = ramda.split(RE_WHITESPACE);

var splitOnUnnestedWhitespace = ramda.split(RE_UNNESTED_WHITESPACE);

var splitOnUnnestedComma = ramda.split(RE_UNNESTED_COMMA);

var splitOnComma = ramda.split(COMMA);

var splitCamelcase = ramda.compose(ramda.split(' '), ramda.replace(RE_CAPITAL_LETTERS, ' $1'));

var splitOnColon = ramda.split(COLON);

// -----------------------------------------------------------------------------
// Wrap
// -----------------------------------------------------------------------------

var wrapWith = function wrapWith(a) {
  var b = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : a;
  return ramda.compose(joinWithNoSpace, ramda.prepend(a), ramda.append(b), String);
};

var wrapWithSingleQuotes = wrapWith(SINGLE_QUOTE);

// -----------------------------------------------------------------------------
// Replace
// -----------------------------------------------------------------------------

var replaceToken = ramda.curry(function (template, tokenName, value) {
  return ramda.replace(toToken(tokenName), value, template);
});

var replaceWithMap = ramda.curry(function (template, valueMap) {
  return reduceObjIndexed(function (acc, _ref) {
    var _ref2 = slicedToArray(_ref, 2),
        key = _ref2[0],
        value = _ref2[1];

    return replaceToken(acc, key, value);
  }, template, valueMap);
});

var replaceWithArray = ramda.curry(function (template, values$$1) {
  return ramda.pipe(reduceIndexed(function (acc, v, idx) {
    return ramda.assoc(ramda.inc(idx), v, acc);
  }, {}), replaceWithMap(template))(values$$1);
});

var replaceTokens = ramda.curry(function (template, value) {
  return condDefault([[isArray, replaceWithArray(template)], [isPlainObj, replaceWithMap(template)]])(value);
});

// -----------------------------------------------------------------------------
// Extract
// -----------------------------------------------------------------------------

var extractFunctionArguments = function extractFunctionArguments(value) {
  var parts = new RegExp(RE_ARGUMENTS_OF_CSS_FUNCTION).exec(value);
  return ramda.nth(1, parts);
};

// -----------------------------------------------------------------------------
// Insert / Append / Prepend
// -----------------------------------------------------------------------------

var appendSubToProp = function appendSubToProp(v) {
  return ramda.compose(joinWithNoSpace, ramda.flatten, ramda.when(isLengthGt(1), ramda.converge(list, [ramda.head, ramda.compose(ramda.map(firstToUpper), ramda.tail)])), compact)(v);
};

var prependSubToProp = ramda.compose(appendSubToProp, ramda.reverse);

var insertSubIntoProp = ramda.compose(ramda.converge(ramda.compose(appendSubToProp, ramda.insert(1)), [ramda.head, ramda.compose(splitCamelcase, joinWithNoSpace, ramda.tail)]), ramda.reverse);

var QUERY_MIN_TEMPLATE = "@media (min-width: #{minWidth})";
var QUERY_MAX_TEMPLATE = "@media (max-width: #{maxWidth})";
var QUERY_MIN_MAX_TEMPLATE = "@media (min-width: #{minWidth}) and (max-width: #{maxWidth})";

var QUERY_TEMPLATE = "#{query} {\n#{value}\n}";

var DECLARATION_TEMPLATE = "#{name}: #{value};";

var CSS_FUNCTION_TEMPLATE = "#{typeOfFunction}(#{value})";

// eslint-disable-next-line import/prefer-default-export
var divideBy = ramda.flip(ramda.divide);

var clampPositive = ramda.clamp(0, Number.POSITIVE_INFINITY);

var elementsOfUnitedNumberString = function elementsOfUnitedNumberString(value) {
  var captures = /^(-?\d+(?:.\d+)?)([a-z|%]+)?$/.exec(value);
  if (!captures || ramda.any(ramda.isNil, [captures, captures[1], captures[2]])) {
    throw new Error('[cssapi] elementsOfUnitedNumber() Supplied value was not a number with a unit: \'' + value + '\'');
  }
  return [Number(captures[1]), captures[2]];
};

var numericPartOfUnitedNumberString = ramda.compose(ramda.view(ramda.lensIndex(0)), elementsOfUnitedNumberString);

var unitPartOfUnitedNumberString = ramda.compose(ramda.view(ramda.lensIndex(1)), elementsOfUnitedNumberString);

var nameOfNamedValue = ramda.pipe(splitOnColon, ramda.last);

var addEmValues = ramda.curry(function (emValue1, emValue2) {
  var value1 = ramda.unless(ramda.equals(0), numericPartOfUnitedNumberString)(emValue1);
  var value2 = ramda.unless(ramda.equals(0), numericPartOfUnitedNumberString)(emValue2);
  return value1 + value2 + 'em';
});

var PX = LENGTH_UNITS.PX;
var REM = LENGTH_UNITS.REM;
var EM = LENGTH_UNITS.EM;


var pxValueToRemOrEmValue = function pxValueToRemOrEmValue(value, baseFontSize) {
  return ramda.divide(value, baseFontSize);
};

var pxValueToRemOrEmString = function pxValueToRemOrEmString(unit, baseFontSize, value) {
  return joinWithNoSpace([ramda.flip(pxValueToRemOrEmValue)(baseFontSize)(value), unit]);
};

var pxValueToPxString = function pxValueToPxString(value) {
  return joinWithNoSpace([value, PX]);
};

var percentageStringToRatio = ramda.compose(divideBy(100), numericPartOfUnitedNumberString);

var ratioToPercentString = ramda.compose(concatRight(PERCENT_UNIT), String, ramda.multiply(100));

var unitlessNumberToDistance = function unitlessNumberToDistance(unit, baseFontSize) {
  return function (value) {
    return ramda.cond([[ramda.equals(PX), function () {
      return pxValueToPxString(value);
    }], [ramda.equals(REM), function () {
      return pxValueToRemOrEmString(REM, baseFontSize, value);
    }], [ramda.equals(EM), function () {
      return pxValueToRemOrEmString(EM, baseFontSize, value);
    }]])(unit);
  };
};

var mulitplyUnitlessNumbersToDistance = function mulitplyUnitlessNumbersToDistance(factor, unit, baseFontSize) {
  return ramda.pipe(numericPartOfUnitedNumberString, ramda.multiply(factor), unitlessNumberToDistance(unit, baseFontSize));
};

var remOrEmToPxValue = function remOrEmToPxValue(value, baseFontSize) {
  return ramda.multiply(value, baseFontSize);
};

var unitedDimensionToUnitlessPixelValue = function unitedDimensionToUnitlessPixelValue(value, baseFontSize) {
  var _elementsOfUnitedNumb = elementsOfUnitedNumberString(value),
      _elementsOfUnitedNumb2 = slicedToArray(_elementsOfUnitedNumb, 2),
      number = _elementsOfUnitedNumb2[0],
      unit = _elementsOfUnitedNumb2[1];

  return isUnitRemOrEm(unit) ? remOrEmToPxValue(number, baseFontSize) : number;
};

var adjustNumberWithUnit = ramda.curry(function (f, value) {
  // In case a unitless number is passed in
  if (isNumber(value)) return f(value);
  // Otherwise calculate using the numeric part and reattach the unit

  var _elementsOfUnitedNumb3 = elementsOfUnitedNumberString(value),
      _elementsOfUnitedNumb4 = slicedToArray(_elementsOfUnitedNumb3, 2),
      n = _elementsOfUnitedNumb4[0],
      unit = _elementsOfUnitedNumb4[1];

  return '' + f(n) + unit;
});

var reduceMaxWidthValue = adjustNumberWithUnit(ramda.flip(ramda.subtract)(0.01));

var createQueryMinHeaderFromTemplate = replaceToken(QUERY_MIN_TEMPLATE, 'minWidth');

var createQueryMaxHeaderFromTemplate = ramda.pipe(reduceMaxWidthValue, replaceToken(QUERY_MAX_TEMPLATE, 'maxWidth'));

var createQueryMinMaxHeaderFromTemplate = ramda.curry(function (maxWidth, minWidth) {
  return replaceTokens(QUERY_MIN_MAX_TEMPLATE, {
    minWidth: minWidth,
    maxWidth: reduceMaxWidthValue(maxWidth)
  });
});

var createQueryFromTemplate = replaceTokens(QUERY_TEMPLATE);

var createDeclarationFromTemplate = replaceTokens(DECLARATION_TEMPLATE);

var createCSSFunctionFromTemplate = replaceTokens(CSS_FUNCTION_TEMPLATE);

var transformArgumentParts = function transformArgumentParts(transform) {
  return ramda.map(ramda.pipe(ramda.trim, splitOnWhitespace, ramda.map(ramda.pipe(ensureArray, transform)), joinWithSpace));
};

var transformFunctionArguments = function transformFunctionArguments(transform) {
  return ramda.pipe(splitOnUnnestedComma, transformArgumentParts(transform), joinWithCommaSpace);
};

var prepareForTransform = ramda.pipe(ramda.when(isNotStringOrArray, String), ramda.when(isString, splitOnUnnestedWhitespace));

var transformValue = ramda.curry(function (transformers, value, data, breakpointName) {
  return ramda.reduce(function (currentValue, transformer) {
    return transformer(currentValue, data, breakpointName);
  }, value, ramda.pipe(ensureArray, ramda.flatten)(transformers));
});

var transformValues = function transformValues(transformers, values$$1, data, breakpointName) {
  return ramda.map(function (value) {
    return transformValue(transformers, value, data, breakpointName);
  }, values$$1);
};


var transformFunctionElements = function transformFunctionElements(transform) {
  return function (value) {
    var typeOfFunction = ramda.match(RE_CSS_FUNCTION_NAME, value)[1];
    return ramda.pipe(ramda.trim, extractFunctionArguments, transformFunctionArguments(transform), function (v) {
      return createCSSFunctionFromTemplate({ typeOfFunction: typeOfFunction, value: v });
    })(value);
  };
};

var transformCalcElements = function transformCalcElements(transform) {
  return function (value) {
    return ramda.pipe(ramda.trim, extractFunctionArguments, ramda.replace(RE_CALC_VALUES, function (v) {
      var r = transform(v);
      return r;
    }), function (v) {
      return createCSSFunctionFromTemplate({ typeOfFunction: 'calc', value: v });
    })(value);
  };
};

var transformer = function transformer(predicate, transformers) {
  return function (value, data, breakpointName) {
    return ramda.when(predicate, function () {
      return transformValue(transformers, value, data, breakpointName);
    })(value);
  };
};

var ERROR_PREFIX = "[cssapi]";
var DATA_PREFIX = "(config.data)";
var BREAKPOINTS_PREFIX = "(config.breakpoints)";
var MQ_PREFIX = "api.mq()";
var API_PREFIX = "api()";
var MIXIN_PREFIX = "mixin()";

// -----------------------------------------------------------------------------
// Utils
// -----------------------------------------------------------------------------

var throwLibError = function throwLibError(message) {
  throw new Error(joinWithSpace([ERROR_PREFIX, message]));
};

var throwPrefixedError = function throwPrefixedError(prefix) {
  return ramda.compose(throwLibError, joinWithSpace, appendFlipped([prefix]));
};

var throwWhenUndefined = function throwWhenUndefined(error) {
  return ramda.when(isUndefined, function () {
    throw error;
  });
};

// -----------------------------------------------------------------------------
// Prefixed Errors
// -----------------------------------------------------------------------------

var throwDataError = throwPrefixedError(DATA_PREFIX);
var throwBreakpointError = throwPrefixedError(BREAKPOINTS_PREFIX);
var throwMQError = throwPrefixedError(MQ_PREFIX);
var throwAPIError = throwPrefixedError(API_PREFIX);
var throwMixinError = throwPrefixedError(MIXIN_PREFIX);

// -----------------------------------------------------------------------------
// Messages
// -----------------------------------------------------------------------------

var noBreakpointAtIndexError = function noBreakpointAtIndexError(idx) {
  return 'Couldn\'t resolve breakpoint at index ' + idx;
};

var noBreakpointWithNameError = function noBreakpointWithNameError(name) {
  return 'Couldn\'t resolve breakpoint with name ' + wrapWithSingleQuotes(name);
};

var invalidBreakpointSyntaxError = function invalidBreakpointSyntaxError(name) {
  return 'The syntax you used to describe your breakpoint range was invalid for ' + wrapWithSingleQuotes(name);
};

var invalidBreakpointError = function invalidBreakpointError(message, args) {
  return message + ' with args: ' + printObj(args);
};

var missingDataNodeError = function missingDataNodeError(name) {
  return 'There is no data node defined named ' + wrapWithSingleQuotes(name);
};

var missingDataItemKeyError = function missingDataItemKeyError(nodeName, keyName) {
  return 'No item has been defined for data.' + nodeName + ' named ' + wrapWithSingleQuotes(keyName);
};

var unrecognisedDataPrefixError = function unrecognisedDataPrefixError(prefix, validPrefixes) {
  return 'Unrecognised prefix encountered: ' + wrapWithSingleQuotes(prefix) + '. Available prefixes are: ' + printObj(validPrefixes);
};

var unsupportedBreakpointValuesError = function unsupportedBreakpointValuesError(declarations) {
  return 'When using the mq() helper you must supply only a single decaration value but you supplied: ' + printObj(declarations);
};

var invalidPropertyError = function invalidPropertyError(name) {
  return 'API doesn\'t support a property named ' + wrapWithSingleQuotes(name);
};

var noThemeObjectError = function noThemeObjectError() {
  return 'There was no theme object available on the props object';
};

var noAPIOnThemeError = function noAPIOnThemeError(value) {
  return 'There was no api function defined on the theme object. Value was: ' + printObj(value);
};

var SCOPES$1 = CONFIG_FIELD_NAMES.SCOPES;
var DATA = CONFIG_FIELD_NAMES.DATA;
var RESOLVE = CONFIG_FIELD_NAMES.RESOLVE;


var findResolvingScopes = function findResolvingScopes(breakpointName, data) {
  return ramda.pipe(ramda.prop(SCOPES$1), ramda.filter(lensSatisfies(ramda.contains(breakpointName), ramda.lensProp(RESOLVE))), ramda.map(ramda.prop(DATA)), ramda.reverse, ramda.append(data))(data);
};

var findDataPropOnScopes = function findDataPropOnScopes(data, dataPropName) {
  return function (breakpointName) {
    // Get scopes that resolve to this breakpoint
    // This will include the default in last position
    var resolvingScopes = findResolvingScopes(breakpointName, data);
    // Run through those scopes in reverse order, starting with the last and
    // ending witht the default, looking for the value.
    return ramda.reduce(function (acc, scopeData) {
      var value = ramda.prop(dataPropName, scopeData);
      if (ramda.isNil(value)) {
        // Acc will either be our fully resolved object, or null
        return acc;
      }
      // If it is an object, we need to run through all scopes as some of its
      // properties might have been defined on an earlier scope.
      if (isObj(value)) {
        if (isObj(acc)) {
          // Merge any props that don't already exist on the object
          return ramda.merge(value, acc);
        }
        // Save the object so we can keep looking for its props
        return value;
      }
      // If the value isn't an object then we have found our result.
      return ramda.reduced(value);
    }, null, resolvingScopes);
  };
};

var findDataProp = function findDataProp(breakpointName, dataPropName) {
  return function (data) {
    // Optimise for early exit
    if (isDefaultBreakpoint(breakpointName) || hasNoScopes(data)) {
      return ramda.prop(dataPropName, data);
    }
    return ramda.pipe(findDataPropOnScopes(data, dataPropName), ramda.defaultTo(ramda.prop(dataPropName, data)))(breakpointName);
  };
};

var keyToValueResolver = function keyToValueResolver(dataPropName) {
  return function (value, data) {
    var breakpointName = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : DEFAULT_BREAKPOINT_NAME;
    return ramda.pipe(findDataProp(breakpointName, dataPropName), ramda.when(isUndefined, function () {
      return throwDataError(missingDataNodeError(dataPropName));
    }))(data);
  };
};

var rhythmUnitsToRemsTransformer = transformer(isRhythmUnit, function (value, data, breakpointName) {
  var rhythm = keyToValueResolver('rhythm')(value, data, breakpointName);
  return mulitplyUnitlessNumbersToDistance(rhythm, data.baseFontSize, LENGTH_UNITS.REM)(value);
});

var unitlessNumberToRemsTransformer = transformer(isValidNonZeroNumber, function (value, data) {
  return unitlessNumberToDistance(LENGTH_UNITS.EM, data.baseFontSize)(value);
});

var lengthToRemsTransformer = [unitlessNumberToRemsTransformer, rhythmUnitsToRemsTransformer];

// the base font size will always be 16 for media queries
var rootPxToEmTransformer = transformValue(lengthToRemsTransformer, ramda.__, { baseFontSize: 16 }, null);

var propName = ramda.prop('name');
var propValue = ramda.prop('value');
var propQuery = ramda.prop('query');
var lName = ramda.lensProp('name');
var lQuery = ramda.lensProp('query');
var assocValue = ramda.assoc('value');
var assocQuery = ramda.assoc('query');

var createBreakpointMapping = ramda.curry(function (name, value, query) {
  return {
    name: name,
    query: query,
    value: value
  };
});

var findBreakpointByName = ramda.curry(function (breakpointMap, name) {
  return ramda.pipe(ramda.find(lensSatisfies(ramda.equals(name), ramda.lensIndex(0))), throwWhenUndefined(noBreakpointWithNameError(name)), ramda.last)(breakpointMap);
});

var findBreakpointIndex = function findBreakpointIndex(breakpointMap, name) {
  return ramda.findIndex(lensSatisfies(ramda.equals(name), ramda.lensIndex(0)), breakpointMap);
};

var findBreakpointByIndex = ramda.curry(function (breakpointMap, idx) {
  return ramda.pipe(ramda.flip(ramda.nth)(breakpointMap), throwWhenUndefined(noBreakpointAtIndexError(idx)))(idx);
});

var findNextBreakpointByIndex = ramda.curry(function (breakpointMap, idx) {
  return ramda.pipe(ramda.inc, findBreakpointByIndex(breakpointMap))(idx);
});

var buildMappingsByIndex = function buildMappingsByIndex(breakpointMap, values$$1) {
  return reduceIndexed(function (mappings, value, idx) {
    return ramda.pipe(findBreakpointByIndex(breakpointMap), ramda.insert(1, value), ramda.apply(createBreakpointMapping), appendFlipped(mappings))(idx);
  }, [], values$$1);
};

var isNotLastBreakpoint = ramda.useWith(ramda.gt, [ramda.pipe(ramda.prop('length'), ramda.dec), ramda.identity]);

var propOffset = ramda.prop('offset');

// eslint-disable-next-line import/prefer-default-export
var hasNoModifier = ramda.complement(ramda.has('modifier'));

var isRange = function isRange(value) {
  return ramda.indexOf(LT_MODIFIER, value) > 1;
};

var firstCharIsModifier = function firstCharIsModifier(value) {
  return ramda.any(ramda.equals(value[0]), MODIFIERS);
};



var addDefaultBreakpoint = ramda.prepend([DEFAULT_BREAKPOINT_NAME, 0]);

var extractOffset = function extractOffset(value) {
  if (hasPositiveOffset(value)) {
    return ramda.split(POSITIVE_OFFSET, value);
  }
  if (hasNegativeOffset(value)) {
    var _split = ramda.split(NEGATIVE_OFFSET, value),
        _split2 = slicedToArray(_split, 2),
        name = _split2[0],
        offset = _split2[1];

    return [name, '-' + offset];
  }
  return [value];
};

var createRangeItem = function createRangeItem(value) {
  var modifier = firstCharIsModifier(value) ? ramda.head(value) : null;
  if (modifier) value = ramda.tail(value);

  var _extractOffset = extractOffset(value),
      _extractOffset2 = slicedToArray(_extractOffset, 2),
      name = _extractOffset2[0],
      offset = _extractOffset2[1];

  var result = {
    name: name,
    offset: offset,
    modifier: modifier
  };

  return ramda.reject(ramda.isNil, result);
};

var parseBreakpoint = function parseBreakpoint(value) {
  if (!isValidModifiedMq(value)) {
    throw Error(invalidBreakpointSyntaxError(value));
  }

  // A range will always have a < in a position greater than 0
  if (!isRange(value)) {
    // It's a single item
    return {
      name: value,
      range: [createRangeItem(value)]
    };
  }

  // Otherwise it's a range (containing two items)
  var values$$1 = ramda.split(LT_MODIFIER, value);
  var firstRangeItem = createRangeItem(values$$1[0]);
  var secondRangeItem = createRangeItem(values$$1[1]);
  return { name: value, range: [firstRangeItem, secondRangeItem] };
};

var lengthEq1 = lengthEq(1);



var nthFlipped = ramda.flip(ramda.nth);

var nthOr = ramda.curry(function (defaultValue, n, a) {
  return ramda.defaultTo(defaultValue, ramda.nth(n, a));
});

var appendFlipped$2 = ramda.flip(ramda.append);

var reduceWithKeys = function reduceWithKeys(reducer) {
  return ramda.converge(ramda.reduce(reducer), [ramda.identity, ramda.keys]);
};

var hasIndex = ramda.curry(function (n, l) {
  return lensSatisfies(isNotUndefined, ramda.lensIndex(n))(l);
});

var numKeys = ramda.compose(ramda.length, ramda.keys);

var breakpointValuesToEms = ramda.map(ramda.apply(ramda.useWith(list, [ramda.identity, ramda.unless(isEmString, rootPxToEmTransformer)])));

var queryHeader = function queryHeader(idx, mappedValues) {
  return function (name) {
    var queryValue = propQuery(mappedValues[idx]);
    if (idx < mappedValues.length - 1) {
      var nextQueryValue = propQuery(mappedValues[idx + 1]);
      return isDefaultBreakpoint(name) ? createQueryMaxHeaderFromTemplate(nextQueryValue) : createQueryMinMaxHeaderFromTemplate(nextQueryValue, queryValue);
    }
    return isDefaultBreakpoint(name) ? null : createQueryMinHeaderFromTemplate(queryValue);
  };
};

var applyOffsetToBreakpointValue = function applyOffsetToBreakpointValue(rangeItem) {
  return ramda.cond([[ramda.pipe(propOffset, ramda.isNil), breakpointMapping_7], [ramda.both(ramda.pipe(propName, isDefaultBreakpoint), ramda.pipe(propOffset, hasNegativeOffset)), function () {
    return 0;
  }], [ramda.T, ramda.pipe(propOffset, ramda.unless(isEmString, rootPxToEmTransformer), addEmValues(rangeItem.value))]])(rangeItem);
};

var renderGtQuery = function renderGtQuery(breakpointMap, rangeItem, rangeItemValue) {
  return isDefaultBreakpoint(rangeItem.name) ? null : createQueryMinHeaderFromTemplate(rangeItemValue);
};

var renderLtQuery = function renderLtQuery(breakpointMap, rangeItem, rangeItemValue) {
  return createQueryMaxHeaderFromTemplate(rangeItemValue);
};

var renderAtQuery = function renderAtQuery(breakpointMap, rangeItem, rangeItemValue) {
  // We need to limit ourselves using a max of the next query if it exists.
  // Use our own index to check if there is a breakpoint after us
  var idx = findBreakpointIndex(breakpointMap, rangeItem.name);

  if (isNotLastBreakpoint(breakpointMap, idx)) {
    // It does exist so get its value
    var nextBreakpointValue = ramda.pipe(findNextBreakpointByIndex(breakpointMap), ramda.nth(1))(idx);
    return isDefaultBreakpoint(rangeItem.name) ? createQueryMaxHeaderFromTemplate(nextBreakpointValue) : createQueryMinMaxHeaderFromTemplate(nextBreakpointValue, rangeItemValue);
  }
  // Otherwise we are the last breakpoint so we don't need a max
  return createQueryMinHeaderFromTemplate(rangeItemValue);
};

var renderRangeQuery = function renderRangeQuery(breakpointMap, firstRangeItem, firstItemValue, range) {
  var secondItemValue = ramda.pipe(ramda.nth(1), applyOffsetToBreakpointValue)(range);
  return isDefaultBreakpoint(firstRangeItem.name) && firstItemValue === 0 ? createQueryMaxHeaderFromTemplate(secondItemValue) : createQueryMinMaxHeaderFromTemplate(secondItemValue, firstItemValue);
};

var renderSingleQuery = ramda.cond([[ramda.either(hasNoModifier, modifierIsGtModifier), function () {
  return renderGtQuery;
}], [modifierIsLtModifier, function () {
  return renderLtQuery;
}], [ramda.T, function () {
  return renderAtQuery;
}]]);

var createQueryHeaderFromRange = function createQueryHeaderFromRange(breakpointMap) {
  return function (range) {
    var firstRangeItem = ramda.head(range);
    var firstItemValue = applyOffsetToBreakpointValue(firstRangeItem);
    var renderQuery = lengthEq1(range) ? renderSingleQuery(firstRangeItem) : renderRangeQuery;
    return renderQuery(breakpointMap, firstRangeItem, firstItemValue, range);
  };
};

var attachBreakpointValues = function attachBreakpointValues(breakpointMap) {
  return function (name, range) {
    return ramda.map(function (rangeItem) {
      return ramda.pipe(propName, findBreakpointByName(breakpointMap), assocValue(ramda.__, rangeItem))(rangeItem);
    }, range);
  };
};

var queryHeaderForRange = function queryHeaderForRange(breakpointMap) {
  return ramda.pipe(attachBreakpointValues(breakpointMap), createQueryHeaderFromRange(breakpointMap));
};

// -----------------------------------------------------------------------------
// API
// -----------------------------------------------------------------------------

var createApi = function createApi(breakpointMap) {
  // Resolve breakpoints for values declared using an object syntax
  var byName = function byName(values$$1) {
    // Parse the breakpoint descriptors
    var parsedValues = reduceObjIndexed(function (mappings, _ref) {
      var _ref2 = slicedToArray(_ref, 2),
          name = _ref2[0],
          value = _ref2[1];

      return ramda.pipe(parseBreakpoint, assocValue(value), appendFlipped(mappings))(name);
    }, [], values$$1);
    // Create query headers
    return ramda.reduce(function (mappings, _ref3) {
      var name = _ref3.name,
          range = _ref3.range,
          value = _ref3.value;
      return ramda.pipe(queryHeaderForRange(breakpointMap), createBreakpointMapping(name, value), appendFlipped(mappings))(name, range);
    }, [], parsedValues);
  };

  // Resolve breakpoints for values declared using an array syntax
  var byIndex = function byIndex(values$$1) {
    var mappedValues = buildMappingsByIndex(breakpointMap, values$$1);

    // Render the appropriate template
    return mapIndexed(function (value, idx) {
      return ramda.pipe(propName, queryHeader(idx, mappedValues), assocQuery(ramda.__, value))(value);
    }, mappedValues);
  };

  var count = function count() {
    return numKeys(breakpointMap);
  };

  return {
    byIndex: byIndex,
    byName: byName,
    count: count
  };
};

// -----------------------------------------------------------------------------
// Exports
// -----------------------------------------------------------------------------

var defaultBreakpointMapProvider = ramda.pipe(defaultToObj, breakpointValuesToEms, createApi);

// eslint-disable-next-line import/prefer-default-export
var SCOPE = "scope";

var argIsScopeObj = ramda.compose(ramda.has(SCOPE), ramda.head);
var argIsObj = ramda.compose(isPlainObj, ramda.head);

var expandScope = function expandScope(provider) {
  return function (value) {
    return ramda.times(ramda.always(value), provider.count());
  };
};

var resolveBreakpointsImpl = function resolveBreakpointsImpl(provider) {
  return function (v) {
    return ramda.cond([[argIsScopeObj, ramda.pipe(ramda.head, ramda.prop(SCOPE), expandScope(provider), resolveBreakpointsImpl(provider))], [argIsObj, ramda.compose(provider.byName, ramda.head)], [ramda.T, provider.byIndex]])(v);
  };
};

var resolveBreakpoints = function resolveBreakpoints(provider) {
  return function () {
    for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }

    return ramda.tryCatch(resolveBreakpointsImpl(provider), function (message) {
      return throwBreakpointError(invalidBreakpointError(message, args));
    })(args);
  };
};

var foundMatch = ramda.flip(ramda.gte)(0);

var findBatchIndex = function findBatchIndex(batches, _ref) {
  var name = _ref.name,
      query = _ref.query;

  var result = ramda.findIndex(ramda.both(lensEq(lName, name), lensEq(lQuery, query)), batches);
  return result;
};

var createNewBatch = function createNewBatch(breakpointMapping, batches) {
  return ramda.append(breakpointMapping, batches);
};

var addToBatch = function addToBatch(_ref2) {
  var name = _ref2.name,
      query = _ref2.query,
      value = _ref2.value;
  return ramda.pipe(propValue, concatRight(value), createBreakpointMapping(name, ramda.__, query));
};

// eslint-disable-next-line import/prefer-default-export
var batchDeclarations = ramda.reduce(function (batches, breakpointMapping) {
  var matchedIndex = findBatchIndex(batches, breakpointMapping);
  return foundMatch(matchedIndex) ? ramda.over(ramda.lensIndex(matchedIndex), addToBatch(breakpointMapping), batches) : createNewBatch(breakpointMapping, batches);
}, []);

var renderQuery = function renderQuery(query) {
  return function (value) {
    return createQueryFromTemplate({ query: query, value: indentLines(value) });
  };
};

var wrapDeclarationWithQuery = function wrapDeclarationWithQuery(query) {
  return ramda.unless(function () {
    return isNull(query);
  }, renderQuery(query));
};

var writeToString = function writeToString(outputString) {
  return ramda.compose(joinWithDoubleNewlines, compact, appendFlipped([outputString]));
};

var renderStyle = function renderStyle(outputString, _ref) {
  var name = _ref.name,
      query = _ref.query,
      value = _ref.value;
  return ramda.pipe(joinWithNewline, wrapDeclarationWithQuery(query, name), writeToString(outputString))(value);
};

var renderStyles = ramda.reduce(renderStyle, '');

var logDeclarations = function logDeclarations(debugTag) {
  return ramda.tap(function (v) {
    if (isNotUndefined(debugTag)) {
      // eslint-disable-next-line no-console
      console.log(debugTag, v);
    }
    return v;
  });
};

var processDeclaration = function processDeclaration(declarationProcessors) {
  return function (acc, _ref) {
    var _ref2 = slicedToArray(_ref, 2),
        processorName = _ref2[0],
        args = _ref2[1];

    var property = ramda.prop(processorName, declarationProcessors);

    ramda.when(ramda.isNil, function () {
      return throwAPIError(invalidPropertyError(processorName));
    })(property);
    return ramda.pipe(ensureArray, ramda.apply(property), appendFlipped(acc))(args);
  };
};

var processDeclarations = function processDeclarations(declarationProcessors) {
  return ramda.pipe(reduceObjIndexed(processDeclaration(declarationProcessors), []), ramda.unnest);
};

var buildApiFunc = function buildApiFunc(declarationProcessors) {
  return function (value, debugTag) {
    return ramda.pipe(processDeclarations(declarationProcessors), batchDeclarations, logDeclarations(debugTag), renderStyles)(value);
  };
};

var attachBreakpointsToDeclarations = function attachBreakpointsToDeclarations(breakpointName, batch) {
  return reduceObjIndexed(function (acc, _ref3) {
    var _ref4 = slicedToArray(_ref3, 2),
        name = _ref4[0],
        value = _ref4[1];

    return ramda.pipe(ramda.unless(isValidMqValue, function () {
      return throwMQError(unsupportedBreakpointValuesError(value));
    }), ramda.objOf(breakpointName), ramda.assoc(name, ramda.__, acc))(value);
  }, {}, batch);
};

var buildMqFunc = function buildMqFunc(apiFunc) {
  var mqFunc = ramda.curry(function (breakpointName, batch) {
    var newMap = attachBreakpointsToDeclarations(breakpointName, batch);
    return apiFunc(newMap);
  });
  apiFunc.mq = mqFunc;
  return apiFunc;
};

var buildApi = ramda.pipe(buildApiFunc, buildMqFunc);

var renderDeclaration = function renderDeclaration(name, value) {
  return ramda.pipe(ramda.evolve({
    name: toKebabCase,
    value: ramda.compose(joinWithSpace, ensureArray)
  }), createDeclarationFromTemplate)({ name: name, value: value });
};

var transformGroups = function transformGroups(transformers, data, name) {
  return ramda.pipe(splitOnUnnestedComma, ramda.map(ramda.trim), ramda.map(ramda.pipe(splitOnUnnestedWhitespace, ramda.map(transformValue(transformers, ramda.__, data, name)), joinWithSpace)), joinWithCommaSpace);
};

var transform = function transform(transformers, name, data) {
  return ramda.ifElse(ramda.both(isString, isGroups), transformGroups(transformers, data, name), transformValue(transformers, ramda.__, data, name));
};

var render = function render(propName$$1, renderer) {
  return ramda.pipe(ensureArray, ramda.partial(renderer, [propName$$1]), list);
};

var createDeclaration = function createDeclaration(propName$$1, data, _ref, _ref2) {
  var _ref$transformers = _ref.transformers,
      transformers = _ref$transformers === undefined ? ramda.identity : _ref$transformers,
      _ref$renderer = _ref.renderer,
      renderer = _ref$renderer === undefined ? renderDeclaration : _ref$renderer;
  var name = _ref2.name,
      query = _ref2.query,
      value = _ref2.value;
  return ramda.pipe(transform(transformers, name, data), render(propName$$1, renderer), createBreakpointMapping(name, ramda.__, query))(value);
};

var buildDeclarationReducer = function buildDeclarationReducer(propName$$1, data, style) {
  return function (declarations, breakpointMapping) {
    return ramda.pipe(createDeclaration, appendFlipped$2(declarations))(propName$$1, data, style, breakpointMapping);
  };
};

var buildDeclaration = function buildDeclaration(propName$$1, data, style) {
  return ramda.reduce(buildDeclarationReducer(propName$$1, data, style), []);
};

var API = CONFIG_FIELD_NAMES.API;
var SCOPES$3 = CONFIG_FIELD_NAMES.SCOPES;
var DATA$1 = CONFIG_FIELD_NAMES.DATA;
var TRANSFORMERS = CONFIG_FIELD_NAMES.TRANSFORMERS;
var BREAKPOINTS = CONFIG_FIELD_NAMES.BREAKPOINTS;


var lDataScopes = ramda.lensPath([DATA$1, SCOPES$3]);
var lData = ramda.lensProp(DATA$1);
var pScopes = ramda.prop(SCOPES$3);
var lApi = ramda.lensProp(API);
var lTransformers = ramda.lensProp(TRANSFORMERS);
var lBreakpoints = ramda.lensProp(BREAKPOINTS);

var SCOPES$2 = CONFIG_FIELD_NAMES.SCOPES;
var ALIASES = CONFIG_FIELD_NAMES.ALIASES;


var expandData = function expandData(config) {
  var resolveDataAlias = propFlipped(config.data.aliases);

  // ---------------------------------------------------------------------------
  // Data Node Items
  // ---------------------------------------------------------------------------

  var expandToken = function expandToken(sourceData) {
    return function (dataNodeItem) {
      var _splitOnColon = splitOnColon(dataNodeItem),
          _splitOnColon2 = slicedToArray(_splitOnColon, 2),
          prefix = _splitOnColon2[0],
          keyName = _splitOnColon2[1];

      var dataNodeName = ramda.has(prefix, sourceData) ? prefix : resolveDataAlias(prefix);
      if (isUndefined(dataNodeName)) {
        var availableKeys = ramda.without([SCOPES$2, ALIASES])(ramda.concat(ramda.keys(sourceData), ramda.keys(config.data[ALIASES])));
        throwDataError(unrecognisedDataPrefixError(prefix, availableKeys));
      }
      var dataNode = ramda.prop(dataNodeName, sourceData);
      if (isUndefined(dataNode)) throwDataError(missingDataNodeError(dataNodeName));

      var resolvedValue = ramda.prop(keyName, dataNode);
      if (isUndefined(resolvedValue)) throwDataError(missingDataItemKeyError(dataNodeName, keyName));

      // eslint-disable-next-line no-use-before-define
      return expandDataNodeItem(sourceData)(resolvedValue);
    };
  };

  var expandParts = function expandParts(sourceData) {
    return function (dataNodeItem) {
      return ramda.pipe(splitOnUnnestedWhitespace,
      // eslint-disable-next-line no-use-before-define
      expandDataNodeItems(sourceData), joinWithSpace)(dataNodeItem);
    };
  };

  var expandDataNodeItems = function expandDataNodeItems(sourceData) {
    return function (dataNodeItems) {
      return (
        // eslint-disable-next-line no-use-before-define
        ramda.map(expandDataNodeItem(sourceData))(dataNodeItems)
      );
    };
  };

  var expandDataNodeItem = function expandDataNodeItem(sourceData) {
    return function (dataNodeItem) {
      return ramda.cond([[isToken, expandToken(sourceData)], [isCSSFunction, transformFunctionElements(expandDataNodeItems(sourceData))], [hasUnnestedWhitespace, expandParts(sourceData)], [ramda.T, ramda.identity]])(dataNodeItem);
    };
  };

  // ---------------------------------------------------------------------------
  // Data Nodes
  // ---------------------------------------------------------------------------

  var expandDataNode = function expandDataNode(dataNodeName) {
    var expandedRootDataItem = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    return reduceWithKeys(function (dataNodeItem, key) {
      var sourceData = ramda.over(ramda.lensProp(dataNodeName), ramda.mergeDeepLeft(dataNodeItem), expandedRootDataItem);
      return ramda.over(ramda.lensProp(key), ramda.when(isString, expandDataNodeItem(sourceData)))(dataNodeItem);
    });
  };

  var expandDataNodes = function expandDataNodes() {
    var expandedRootData = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
    return reduceWithKeys(function (acc, key) {
      return ramda.pipe(expandDataNode, ramda.over(ramda.lensProp(key), ramda.__, acc))(key, expandedRootData);
    });
  };

  // ---------------------------------------------------------------------------
  // Scopes
  // ---------------------------------------------------------------------------

  var expandScopes = function expandScopes(expandedRootData) {
    return ramda.map(ramda.over(lData, expandDataNodes(expandedRootData)));
  };

  // ---------------------------------------------------------------------------
  // Data
  // ---------------------------------------------------------------------------

  var expandDataImp = function expandDataImp(data) {
    // Expand the default data nodes
    var expandedRootData = ramda.pipe(ramda.without(SCOPES$2), expandDataNodes(data))(data);
    // Expand the data nodes inside scopes
    var expandedScopeData = ramda.pipe(pScopes, ramda.when(isNotUndefined, expandScopes(expandedRootData)))(data);

    return ramda.assoc(SCOPES$2, expandedScopeData)(expandedRootData);
  };

  return ramda.over(lData, expandDataImp, config);
};

var partPositionTransformer = function partPositionTransformer(position) {
  return function (transformers) {
    return function (value, data, breakpointName) {
      return ramda.pipe(prepareForTransform, ramda.adjust(transformValue(transformers, ramda.__, data, breakpointName), position))(value);
    };
  };
};

var transformPartsWith = function transformPartsWith(transformers) {
  return function (value, data, breakpointName) {
    value = prepareForTransform(value);
    if (ramda.both(isString, isGroups)(value)) {
      return ramda.map(ramda.pipe(splitOnUnnestedWhitespace, transformValue(transformers, ramda.__, data, breakpointName), joinWithSpace))(value);
    }

    return transformValues(transformers, value, data, breakpointName);
  };
};

var toAppendedProps = function toAppendedProps(propName, style, affixedValues, toProp) {
  return ramda.reduce(function (acc, direction) {
    return ramda.compose(ramda.assoc(ramda.__, style, acc), toProp, list)(propName, direction);
  }, {}, affixedValues);
};

var expandMainProp = function expandMainProp(propName, style) {
  var wrapper = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : ramda.identity;
  return ramda.assoc(propName, ramda.over(lTransformers, wrapper, style));
};

var expandSubProps = function expandSubProps(toProp, suffixes) {
  var wrapper = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : ramda.identity;
  return function (propName, style) {
    return ramda.mergeDeepRight(toAppendedProps(propName, ramda.over(lTransformers, wrapper, style), suffixes, toProp));
  };
};

var wrapWithTransform = function wrapWithTransform(wrapper) {
  return function (propName, style) {
    return ramda.pipe(ramda.over(lTransformers, wrapper), ramda.objOf(propName))(style);
  };
};

var DIRECTIONS$1 = Object.freeze({
  TOP: 'top',
  RIGHT: 'right',
  BOTTOM: 'bottom',
  LEFT: 'left'
});

var CORNERS = Object.freeze({
  TOP_RIGHT: 'topRight',
  BOTTOM_RIGHT: 'bottomRight',
  BOTTOM_LEFT: 'bottomLeft',
  TOP_LEFT: 'topLeft'
});

var DIRECTIONS_LIST = ramda.values(DIRECTIONS$1);
var CORNERS_LIST = ramda.values(CORNERS);

var DIRECTIONS_LIST_HORIZONTAL = [DIRECTIONS$1.LEFT, DIRECTIONS$1.RIGHT];

var DIRECTIONS_LIST_VERTICAL = [DIRECTIONS$1.TOP, DIRECTIONS$1.BOTTOM];

var MIN_MAX = {
  MIN: 'min',
  MAX: 'max'
};

var MIN_MAX_LIST = ramda.values(MIN_MAX);

var AXES = {
  X: 'x',
  Y: 'y'
};

var AXES_LIST = ramda.values(AXES);

var axesExpander = function axesExpander() {
  var _ref = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {},
      _ref$mainWrapper = _ref.mainWrapper,
      mainWrapper = _ref$mainWrapper === undefined ? ramda.identity : _ref$mainWrapper,
      _ref$subWrapper = _ref.subWrapper,
      subWrapper = _ref$subWrapper === undefined ? ramda.identity : _ref$subWrapper;

  return function (propName, style) {
    return ramda.converge(ramda.pipe, [expandMainProp, expandSubProps(appendSubToProp, AXES_LIST, subWrapper)])(propName, style, mainWrapper)({});
  };
};

var allPartsTransformer = function allPartsTransformer(transformers) {
  return function (value, data, breakpointName) {
    var preparedValue = prepareForTransform(value);
    return transformValues(transformers, preparedValue, data, breakpointName);
  };
};

var cornerExpander = function cornerExpander() {
  var _ref = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {},
      _ref$mainWrapper = _ref.mainWrapper,
      mainWrapper = _ref$mainWrapper === undefined ? allPartsTransformer : _ref$mainWrapper,
      _ref$subWrapper = _ref.subWrapper,
      subWrapper = _ref$subWrapper === undefined ? ramda.identity : _ref$subWrapper,
      _ref$createPropNameSt = _ref.createPropNameStrategy,
      createPropNameStrategy = _ref$createPropNameSt === undefined ? appendSubToProp : _ref$createPropNameSt;

  return function (propName, style) {
    return ramda.converge(ramda.pipe, [expandMainProp, expandSubProps(createPropNameStrategy, CORNERS_LIST, subWrapper)])(propName, style, mainWrapper)({});
  };
};

var directionExpander = function directionExpander() {
  var _ref = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {},
      _ref$mainWrapper = _ref.mainWrapper,
      mainWrapper = _ref$mainWrapper === undefined ? allPartsTransformer : _ref$mainWrapper,
      _ref$subWrapper = _ref.subWrapper,
      subWrapper = _ref$subWrapper === undefined ? ramda.identity : _ref$subWrapper,
      _ref$createPropNameSt = _ref.createPropNameStrategy,
      createPropNameStrategy = _ref$createPropNameSt === undefined ? appendSubToProp : _ref$createPropNameSt;

  return function (propName, style) {
    return ramda.converge(ramda.pipe, [expandMainProp, expandSubProps(createPropNameStrategy, DIRECTIONS_LIST, subWrapper)])(propName, style, mainWrapper)({});
  };
};

var directionsExpander = function directionsExpander() {
  return function (propName, style) {
    return expandSubProps(appendSubToProp, DIRECTIONS_LIST)('', style)({});
  };
};

var minMaxExpander = function minMaxExpander() {
  return function (propName, style) {
    return ramda.converge(ramda.pipe, [expandMainProp, expandSubProps(prependSubToProp, MIN_MAX_LIST)])(propName, style)({});
  };
};

// Expanders take an item described in the config and expand it to multiple
// separate properties. For example 'padding' is expanded to 'padding-top',
// 'padding-right', 'padding-bottom' and 'padding-left', or 'directions' is
// expanded to 'top', 'right', 'bottom' and 'left'.
var EXPANDER_MAP = Object.freeze({
  padding: directionExpander(),
  margin: directionExpander(),
  border: directionExpander({
    mainWrapper: transformPartsWith,
    subWrapper: transformPartsWith
  }),
  borderWidth: directionExpander({
    createPropNameStrategy: insertSubIntoProp
  }),
  borderStyle: directionExpander({
    createPropNameStrategy: insertSubIntoProp
  }),
  borderColor: directionExpander({
    createPropNameStrategy: insertSubIntoProp
  }),
  borderRadius: cornerExpander({
    createPropNameStrategy: insertSubIntoProp
  }),
  width: minMaxExpander(),
  height: minMaxExpander(),
  directions: directionsExpander(),
  overflow: axesExpander(),
  outlineColor: wrapWithTransform(transformPartsWith),
  outline: wrapWithTransform(transformPartsWith),
  flex: wrapWithTransform(partPositionTransformer(2)),
  background: wrapWithTransform(transformPartsWith),
  backgroundImage: wrapWithTransform(transformPartsWith),
  backgroundPosition: axesExpander({
    mainWrapper: transformPartsWith,
    subWrapper: transformPartsWith
  }),
  boxShadow: wrapWithTransform(transformPartsWith),
  transformOrigin: wrapWithTransform(transformPartsWith),
  offset: wrapWithTransform(transformPartsWith),
  offsetV: wrapWithTransform(transformPartsWith),
  offsetH: wrapWithTransform(transformPartsWith),
  paddingH: wrapWithTransform(transformPartsWith),
  paddingV: wrapWithTransform(transformPartsWith),
  marginH: wrapWithTransform(transformPartsWith),
  marginV: wrapWithTransform(transformPartsWith),
  borderH: wrapWithTransform(transformPartsWith),
  borderV: wrapWithTransform(transformPartsWith)
});

var expand = reduceObjIndexed(function (api, _ref) {
  var _ref2 = slicedToArray(_ref, 2),
      propName = _ref2[0],
      expander = _ref2[1];

  return ramda.compose(ramda.mergeDeepRight(api), expander)(propName, ramda.prop(propName, api));
}, ramda.__, EXPANDER_MAP);

var expandStyles = ramda.over(lApi, expand);

var renderDeclarations = ramda.map(ramda.apply(renderDeclaration));

var renderBaseline = function renderBaseline(name, value) {
  return ramda.compose(joinWithNewline, renderDeclarations)([['fontSize', value[0]], ['lineHeight', value[1]]]);
};

var renderDirectionProps = function renderDirectionProps(name, value) {
  var right = value[1] || value[0];
  var bottom = value[2] || value[0];
  var left = value[3] || value[1] || value[0];
  var directionValues = [value[0], right, bottom, left];

  return ramda.compose(joinWithNewline, renderDeclarations, ramda.zip(DIRECTIONS_LIST))(directionValues);
};

var renderDualProps = function renderDualProps(propNames) {
  return function (_, value) {
    var directionValues = [value, value];

    return ramda.compose(joinWithNewline, renderDeclarations, ramda.zip(propNames))(directionValues);
  };
};

var renderDualProps$1 = function renderDualProps(propNames) {
  return function (_, value) {
    var firstProp = value[0];
    var lastProp = value[1] || firstProp;
    var directionValues = [firstProp, lastProp];

    return ramda.compose(joinWithNewline, renderDeclarations, ramda.zip(propNames))(directionValues);
  };
};

var divideBy2 = ramda.flip(ramda.divide)(2);

var wholeLinesForFontSize = function wholeLinesForFontSize(minLeading, fontSize, baselineHeight) {
  var lines = Math.ceil(fontSize / baselineHeight);
  return lines * baselineHeight - fontSize >= minLeading ? lines : lines + 1;
};

// eslint-disable-next-line import/prefer-default-export
var linesForFontsize = function linesForFontsize(minLeading, allowHalfLines, baselineHeight, fontSize) {
  return isTrue(allowHalfLines) ? divideBy2(wholeLinesForFontSize(minLeading, fontSize, divideBy2(baselineHeight))) : wholeLinesForFontSize(minLeading, fontSize, baselineHeight);
};

var keyToObjectValueResolver = function keyToObjectValueResolver(dataPropName) {
  return function (value, data, breakpointName) {
    return ramda.pipe(keyToValueResolver(dataPropName), ramda.prop(value), ramda.when(isUndefined, function () {
      return throwDataError(missingDataItemKeyError(dataPropName, value));
    }))(value, data, breakpointName);
  };
};

var keysToValuesResolver = function keysToValuesResolver(name, keys$$1) {
  return function (data, breakpointName) {
    return ramda.map(function (key) {
      return keyToObjectValueResolver(name)(key, data, breakpointName);
    }, keys$$1);
  };
};

var rhythmUnitsToRemsTransformer$1 = function rhythmUnitsToRemsTransformer(fontSizeToLengthTransformer) {
  return transformer(ramda.T, function (value, data, breakpointName) {
    var _keysToObjectValuesRe = keysToValuesResolver('baseline', ['lineHeight', 'minLeading', 'allowHalfLines'])(data, breakpointName),
        _keysToObjectValuesRe2 = slicedToArray(_keysToObjectValuesRe, 3),
        lineHeight = _keysToObjectValuesRe2[0],
        minLeading = _keysToObjectValuesRe2[1],
        allowHalfLines = _keysToObjectValuesRe2[2];

    var baseFontSize = keyToValueResolver('baseFontSize')(value, data, breakpointName);

    var _unless = ramda.unless(isArray, splitOnUnnestedWhitespace)(value),
        _unless2 = slicedToArray(_unless, 2),
        fontSize = _unless2[0],
        lines = _unless2[1];

    var transformedFontSize = transformValue(fontSizeToLengthTransformer, fontSize, data, breakpointName);
    var fontSizeUnitlessPx = unitedDimensionToUnitlessPixelValue(transformedFontSize, baseFontSize);

    var transformedLines = ramda.pipe(ramda.when(isUndefined, function () {
      return linesForFontsize(minLeading, allowHalfLines, lineHeight, fontSizeUnitlessPx);
    }), ramda.multiply(lineHeight), transformValue(fontSizeToLengthTransformer, ramda.__, data, breakpointName))(lines);

    return [transformedFontSize, transformedLines];
  });
};

var dataLookupTransformer = function dataLookupTransformer(dataNodeName) {
  return function (value, data, breakpointName) {
    var aliases = filterKeys(ramda.equals(dataNodeName), data.aliases);
    return ramda.when(isTokenWithName(ramda.prepend(dataNodeName, aliases)), function () {
      var name = nameOfNamedValue(value);
      return keyToObjectValueResolver(dataNodeName)(name, data, breakpointName);
    })(value);
  };
};

var DATA_NODE_NAMES = ['color', 'gradient', 'font', 'scale', 'boxShadow', 'image', 'border'];

var _reduce = ramda.reduce(function (acc, value) {
  return ramda.pipe(dataLookupTransformer, ramda.assoc(value + 'LookupTransformer', ramda.__, acc))(value);
}, {})(DATA_NODE_NAMES);

var colorLookupTransformer = _reduce.colorLookupTransformer;
var gradientLookupTransformer = _reduce.gradientLookupTransformer;
var fontLookupTransformer = _reduce.fontLookupTransformer;
var scaleLookupTransformer = _reduce.scaleLookupTransformer;
var boxShadowLookupTransformer = _reduce.boxShadowLookupTransformer;
var imageLookupTransformer = _reduce.imageLookupTransformer;
var borderLookupTransformer = _reduce.borderLookupTransformer;

var impl = function impl(transformers) {
  return function (value, data, breakpointName) {
    return transformFunctionElements(transformValue(transformPartsWith(transformers), ramda.__, data, breakpointName))(value);
  };
};

var gradientTransformer = function gradientTransformer(transformers) {
  return function (value, data, breakpointName) {
    return transformer(isGradient, impl(transformers))(value, data, breakpointName);
  };
};

var keysToValuesResolver$1 = function keysToValuesResolver(keys$$1) {
  return function (value, data, breakpointName) {
    return ramda.map(function (key) {
      return keyToValueResolver(key)(value, data, breakpointName);
    }, keys$$1);
  };
};

var rhythmUnitsToRemsTransformer$2 = transformer(isRhythmUnit, function (value, data, breakpointName) {
  var _keysToValuesResolver = keysToValuesResolver$1(['rhythm', 'lengthUnit', 'baseFontSize'])(value, data, breakpointName),
      _keysToValuesResolver2 = slicedToArray(_keysToValuesResolver, 3),
      rhythm = _keysToValuesResolver2[0],
      lengthUnit = _keysToValuesResolver2[1],
      baseFontSize = _keysToValuesResolver2[2];

  return mulitplyUnitlessNumbersToDistance(rhythm, lengthUnit, baseFontSize)(value);
});

var calcTransformer = function calcTransformer(transformers) {
  return transformer(isCalcFunction, function (value, data, breakpointName) {
    var r = transformCalcElements(transformValue(transformers, ramda.__, data, breakpointName))(value);
    return r;
  });
};

var unitlessNumberToLengthTransformer = transformer(isValidNonZeroNumber, function (value, data, breakpointName) {
  var _keysToValueResolver = keysToValuesResolver$1(['lengthUnit', 'baseFontSize'])(value, data, breakpointName),
      _keysToValueResolver2 = slicedToArray(_keysToValueResolver, 2),
      lengthUnit = _keysToValueResolver2[0],
      baseFontSize = _keysToValueResolver2[1];

  return unitlessNumberToDistance(lengthUnit, baseFontSize)(value);
});

var lengthTransformers = [
// eslint-disable-next-line no-use-before-define
calcTransformer(rhythmUnitsToRemsTransformer$2)].concat([unitlessNumberToLengthTransformer, rhythmUnitsToRemsTransformer$2]);

var percentageStringToRatioTransformer = transformer(isPercentString, percentageStringToRatio);

var impl$1 = function impl(transformers) {
  return function (value, data, breakpointName) {
    return transformFunctionElements(transformValue(transformPartsWith(transformers), ramda.__, data, breakpointName))(value);
  };
};

var transformTransformer = function transformTransformer(transformers) {
  return function (value, data, breakpointName) {
    return transformer(isTransformTranslateFunction, impl$1(transformers))(value, data, breakpointName);
  };
};

// -----------------------------------------------------------------------------
// Define API
// -----------------------------------------------------------------------------

var defaultConfig = {
  breakpoints: [],
  data: {
    aliases: {
      c: 'color',
      g: 'gradient',
      s: 'scale',
      d: 'boxShadow',
      b: 'border',
      i: 'image',
      f: 'font'
    },
    lengthUnit: LENGTH_UNITS.REM, // | LENGTH_UNITS.PX | LENGTH_UNITS.EM
    baseFontSize: 16, // Font size of your page's root element
    rhythm: 20, // Unit of rhythm for use in layout
    baseline: {
      lineHeight: 20, // Baseline height
      minLeading: 2, // Minimum remaining leading before line or half-line added
      allowHalfLines: true // Allow half-lines to be used in baseline calc
    },
    color: {},
    scale: {},
    gradient: {},
    boxShadow: {},
    image: {},
    border: {}
  },
  api: {
    // -------------------------------------------------------------------------
    // Box Model
    // -------------------------------------------------------------------------

    padding: {
      transformers: lengthTransformers
    },
    margin: {
      transformers: lengthTransformers
    },
    border: {
      transformers: [borderLookupTransformer, lengthTransformers, colorLookupTransformer]
    },
    borderWidth: {
      transformers: lengthTransformers
    },
    borderColor: {
      transformers: colorLookupTransformer
    },
    borderStyle: {},
    borderSpacing: {
      transformers: lengthTransformers
    },
    borderRadius: {
      transformers: lengthTransformers
    },
    borderImageSource: {
      transformers: [imageLookupTransformer, gradientLookupTransformer, gradientTransformer([colorLookupTransformer].concat(toConsumableArray(lengthTransformers)))]
    },

    // -------------------------------------------------------------------------
    // Outline
    // -------------------------------------------------------------------------

    outline: {
      transformers: [borderLookupTransformer, lengthTransformers, colorLookupTransformer]
    },
    outlineColor: {
      transformers: colorLookupTransformer
    },
    outlineOffset: {
      transformers: lengthTransformers
    },
    outlineStyle: {},
    outlineWidth: {
      transformers: lengthTransformers
    },

    // -------------------------------------------------------------------------
    // Text
    // -------------------------------------------------------------------------

    fontFamily: {
      transformers: fontLookupTransformer
    },
    fontSize: {
      transformers: [scaleLookupTransformer, lengthTransformers]
    },
    fontWeight: {},
    fontVarient: {},
    fontStretch: {},
    fontStyle: {},
    lineHeight: {
      transformers: lengthTransformers
    },
    textAlign: {},
    letterSpacing: {
      transformers: lengthTransformers
    },
    wordWrap: {},
    wordSpacing: {},
    textDecoration: {},
    whiteSpace: {},

    // -------------------------------------------------------------------------
    // List
    // -------------------------------------------------------------------------

    listStyle: {
      transformers: imageLookupTransformer
    },
    listStyleImage: {
      transformers: imageLookupTransformer
    },
    listStylePosition: {},
    listStyleType: {},

    // -------------------------------------------------------------------------
    // Background
    // -------------------------------------------------------------------------

    background: {
      transformers: [colorLookupTransformer, imageLookupTransformer, gradientLookupTransformer, gradientTransformer([colorLookupTransformer].concat(toConsumableArray(lengthTransformers))), imageLookupTransformer]
    },

    backgroundAttachment: {},

    backgroundClip: {},

    backgroundColor: {
      transformers: colorLookupTransformer
    },

    backgroundImage: {
      transformers: [gradientLookupTransformer, imageLookupTransformer, gradientTransformer([colorLookupTransformer].concat(toConsumableArray(lengthTransformers))), imageLookupTransformer]
    },

    backgroundOrigin: {},

    backgroundPosition: {
      transformers: lengthTransformers
    },

    backgroundRepeat: {},

    backgroundSize: {
      transformers: lengthTransformers
    },

    // -------------------------------------------------------------------------
    // Color / Visibility
    // -------------------------------------------------------------------------

    opacity: {
      transformers: percentageStringToRatioTransformer
    },
    color: {
      transformers: colorLookupTransformer
    },
    visibility: {},

    // -------------------------------------------------------------------------
    // Layout
    // -------------------------------------------------------------------------

    display: {},
    float: {},
    clear: {},
    position: {},
    directions: {
      transformers: lengthTransformers
    },
    width: {
      transformers: lengthTransformers
    },
    height: {
      transformers: lengthTransformers
    },

    // -------------------------------------------------------------------------
    // Flexbox
    // -------------------------------------------------------------------------

    flex: {
      transformers: lengthTransformers
    },
    flexDirection: {},
    justifyContent: {},
    alignItems: {},
    alignContent: {},
    alignSelf: {},
    flexBasis: {
      transformers: lengthTransformers
    },
    flexShrink: {}, // Doesn't support <length> values
    flexGrow: {}, // Doesn't support <length> values
    flexWrap: {},
    order: {},

    // -------------------------------------------------------------------------
    // Columns
    // -------------------------------------------------------------------------

    columnCount: {},

    columnWidth: {
      transformers: lengthTransformers
    },

    columnGap: {
      transformers: lengthTransformers
    },

    columnRuleWidth: {
      transformers: lengthTransformers
    },

    // -------------------------------------------------------------------------
    // Tables
    // -------------------------------------------------------------------------

    verticalAlign: {},

    // -------------------------------------------------------------------------
    // Misc
    // -------------------------------------------------------------------------

    zIndex: {},
    zoom: {},
    overflow: {},
    boxShadow: {
      transformers: [boxShadowLookupTransformer, colorLookupTransformer, lengthTransformers]
    },
    cursor: {},
    hyphens: {},

    // -------------------------------------------------------------------------
    // Transforms
    // -------------------------------------------------------------------------

    transform: { transformers: transformTransformer(lengthTransformers) },
    transformBox: {},
    transformOrigin: {
      transformers: lengthTransformers
    },

    // -------------------------------------------------------------------------
    // SVG
    // -------------------------------------------------------------------------

    fill: {
      transformers: [colorLookupTransformer, gradientLookupTransformer, gradientTransformer([colorLookupTransformer].concat(toConsumableArray(lengthTransformers)))]
    },

    stroke: {
      transformers: [colorLookupTransformer, gradientLookupTransformer, gradientTransformer([colorLookupTransformer].concat(toConsumableArray(lengthTransformers)))]
    },

    stopColor: {
      transformers: [colorLookupTransformer]
    },

    // -------------------------------------------------------------------------
    // Helpers
    // -------------------------------------------------------------------------

    paddingH: {
      transformers: lengthTransformers,
      renderer: renderDualProps$1(['paddingLeft', 'paddingRight'])
    },

    paddingV: {
      transformers: lengthTransformers,
      renderer: renderDualProps$1(['paddingTop', 'paddingBottom'])
    },

    marginH: {
      transformers: lengthTransformers,
      renderer: renderDualProps$1(['marginRight', 'marginLeft'])
    },

    marginV: {
      transformers: lengthTransformers,
      renderer: renderDualProps$1(['marginTop', 'marginBottom'])
    },

    borderH: {
      transformers: [borderLookupTransformer, lengthTransformers, colorLookupTransformer],
      renderer: renderDualProps(['borderLeft', 'borderRight'])
    },

    borderV: {
      transformers: [borderLookupTransformer, lengthTransformers, colorLookupTransformer],
      renderer: renderDualProps(['borderTop', 'borderBottom'])
    },

    offset: {
      transformers: lengthTransformers,
      renderer: renderDirectionProps
    },

    offsetV: {
      transformers: lengthTransformers,
      renderer: renderDualProps$1(DIRECTIONS_LIST_VERTICAL)
    },

    offsetH: {
      transformers: lengthTransformers,
      renderer: renderDualProps$1(DIRECTIONS_LIST_HORIZONTAL)
    },

    borderTopRadius: {
      transformers: lengthTransformers,
      renderer: renderDualProps$1(['borderTopLeftRadius', 'borderTopRightRadius'])
    },

    borderRightRadius: {
      transformers: lengthTransformers,
      renderer: renderDualProps$1(['borderTopRightRadius', 'borderBottomRightRadius'])
    },

    borderBottomRadius: {
      transformers: lengthTransformers,
      renderer: renderDualProps$1(['borderBottomRightRadius', 'borderBottomLeftRadius'])
    },

    borderLeftRadius: {
      transformers: lengthTransformers,
      renderer: renderDualProps$1(['borderBottomLeftRadius', 'borderTopLeftRadius'])
    },

    baseline: {
      transformers: rhythmUnitsToRemsTransformer$1([scaleLookupTransformer, lengthTransformers]),
      renderer: renderBaseline
    }
  }
};

var mergeWithDefaultConfig = ramda.mergeDeepRight(defaultConfig);

var ensureDataScopes = ramda.over(lDataScopes, defaultToArray);

var createDefaultBreakpointProvider = ramda.compose(defaultBreakpointMapProvider, addDefaultBreakpoint);

var ensureBreakpointProvider = ramda.over(lBreakpoints, ramda.unless(isBreakpointProvider, createDefaultBreakpointProvider));

var createDeclarationProcessor = function createDeclarationProcessor(breakpointProvider, name, data, style) {
  return ramda.pipe(resolveBreakpoints(breakpointProvider), buildDeclaration(name, data, style));
};

var createDeclarationProcessorReducer = function createDeclarationProcessorReducer(breakpointProvider, data) {
  return function (acc, _ref) {
    var _ref2 = slicedToArray(_ref, 2),
        name = _ref2[0],
        style = _ref2[1];

    return ramda.pipe(createDeclarationProcessor, ramda.assoc(name, ramda.__, acc))(breakpointProvider, name, data, style);
  };
};

var createDeclarationProcessors = function createDeclarationProcessors(_ref3) {
  var breakpoints = _ref3.breakpoints,
      data = _ref3.data,
      api = _ref3.api;
  return reduceObjIndexed(createDeclarationProcessorReducer(breakpoints, data), {}, api);
};

// -----------------------------------------------------------------------------
// Exports
// -----------------------------------------------------------------------------

var api = ramda.pipe(defaultToObj, mergeWithDefaultConfig, ensureDataScopes, expandData, expandStyles, ensureBreakpointProvider, createDeclarationProcessors, buildApi);

// Note: When called as a template-literal, a the argument to a function will be
// wrapped in an array, so we need to extract it using head.
// eslint-disable-next-line import/prefer-default-export
var scope = function scope(strings) {
  for (var _len = arguments.length, values$$1 = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
    values$$1[_key - 1] = arguments[_key];
  }

  return ramda.pipe(ramda.zip, ramda.append(ramda.last(strings)), ramda.flatten, joinWithNoSpace, ramda.objOf(SCOPE))(strings, values$$1);
};

var getApiFromProps = function getApiFromProps(errorFunc) {
  return function (props$$1) {
    return ramda.pipe(ramda.prop('theme'), ramda.unless(isPlainObj, function (value) {
      return errorFunc(noThemeObjectError(value));
    }), ramda.prop('api'), ramda.unless(isFunction, function (value) {
      return errorFunc(noAPIOnThemeError(value));
    }))(props$$1);
  };
};

var api$1 = function api(declarationsObj) {
  return ramda.pipe(getApiFromProps(throwAPIError), ramda.flip(ramda.apply)([declarationsObj]));
};

var mixin = function mixin(f) {
  return function () {
    for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }

    return function (props$$1) {
      return ramda.pipe(getApiFromProps(throwMixinError), ramda.flip(ramda.pair)(props$$1), ramda.apply(f), ramda.flip(ramda.apply)(args))(props$$1);
    };
  };
};

exports['default'] = api;
exports.scope = scope;
exports.s = scope;
exports.api = api$1;
exports.mixin = mixin;

Object.defineProperty(exports, '__esModule', { value: true });

})));
